# -*- coding: utf-8 -*-
"""
Created on Thu Dec 15 12:09:38 2016

@author: akidribm
"""
# Imports necessary sort types and functions from the library
# Imports sortal sort types, functions and rules
from sortal import (transformSet, similarityMap, rule, primitiveSort, crossAlgebraPlane2D, point2D, 
                    coordinate, lineSegment2D, line2D, label, weight, disjunctiveSort, attributeSort, 
                    affineTransform,  vector2D, planeSegment2D)
from sortal import description, primitiveSort, disjunctiveSort, rule, sdlParser, numeric, parseReader

# Imports math functions
from decimal import Decimal
import math

# Imports API functions and variables
from api.api_create import create
from api.sort_types import sort_types
from api.transform import transform

# Imports GUI library
from pylab import *
import matplotlib
from sympy import pi,cos,sin
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.backends.backend_qt4agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.figure import Figure
import  matplotlib.pylab as plt
import matplotlib.pyplot as pltplot
from matplotlib.collections import LineCollection
matplotlib.use ("gTkAgg")
from matplotlib.collections import PatchCollection
import tkinter as tk
from tkinter import ttk, Tk, Label, Button, Frame
from tkinter import *

plt.ion()

if __name__ == "__main__":
    
    LARGE_FONT= ("Verdana", 16)
    NORM_FONT= ("Verdana", 12)
    
    a = 15 # Angle to be used for transform function 
    
    rule.cleanup()
    sdlParser.parse('sdl_files/nonparametric_rotating_squares.sdl')
    formM1 = sdlParser.forms['test0a1']
    r1 = sdlParser.rules['r2_a']
    r2 = sdlParser.rules['r2_b']
    
    class Application(Frame):
        
        print ("<<<< GENERATING APPLICATION WINDOW >>>>")
        fig_main = Figure(figsize = (8, 8), dpi = 150) # Figure contains 'Current Drawing'
        fig_rules = Figure(figsize = (6.5, 6.5), dpi = 100) # Figure contains 'Rules'
        
        # Sets main figure to initial shape formM1
        formMain = formM1.duplicate()
        
        blackPlanes = [] # List contains all planes with weight '255' 
        whitePlanes = [] # List contains all planes with weight '0'
        count = 1 # Keeps track of which rule is to be applied
        
        def __init__(self, master = None):
            Frame.__init__(self, master)
            self.grid()
            self.master.title("SGI Test Case: Rotating 2D Square Planes")
    
            # Generates tkinter window grid with 6 rows and 6 columns
            for r in range(6):
                self.master.rowconfigure(r, weight=1)    
            for c in range(6):
                self.master.columnconfigure(c, weight=1)

            # Holds 'Current Drawing' figure (fig_main)
            self.Frame1 = Frame(master, bg="gray")
            self.Frame1.grid(row = 0, column = 0, rowspan = 6, columnspan = 5, sticky = W+E+N+S) 
            self.Frame1_label = Label (master = self.Frame1, text = "Current Drawing, Iteration {}".format(self.count), background = "gray", font = LARGE_FONT, anchor = "center")
            self.Frame1_label.pack()
            self.Frame2_label = Label(master = self.Frame1, text = "Matches are highlighted in orange when the cursor hovers over the plot. \nClick directly on the match plane to apply the relevant rule.", background = 'orange', font = NORM_FONT, anchor = "center")
            self.Frame2_label.pack(fill = BOTH, expand = 1)
            
            # Holds 'Rules' figure (fig_rules)
            # LHS and RHS figures are side by side for each rule
            self.Frame3 = Frame(master, bg='gray')
            self.Frame3.grid(row = 0, column = 5, rowspan = 6, columnspan = 1, sticky = W+E+N+S)
            self.Frame3_label = Label (master = self.Frame3, text ="Rules", background = 'gray', font = LARGE_FONT, anchor = "center")
            self.Frame3_label.pack()
            
            self.menu() # Initializes menu
            self.set_figure() # Embeds matplotlib figures 'fig_main' and 'fig_rules' in tkinter window
            self.call_axes() # Initalizes subplot in 'fig_main' and adjusts plot settings
            self.initialization() # Plots initial form onto 'fig_main', 'a_main'
            self.connect() # Activates mouseover, mouseoff and mouseclick functions
            self.rules_viz() # Plots rules 1 and 2 in RHS of tkinter window
            
        """
        =====================
        API: RULE APPLICATION
        =====================
        
        Generate(count) takes the current self.count value as input and applies r1 (even count value) or r2 (odd count value). 
        This generates the matches as well as the new planes in the form and changes the form's contents.
        The r1.detect/r2.detect method takes a form as input and detects matches in the form. The r1.apply/r2.apply method 
        combines both detect and perform methods (transforming the match from LHS to RHS).
        """
        def generate(self, count):
            if count%2 == 0:
                r1.apply(self.formMain)  
                r2.detect(self.formMain) # For purposes of showing match in GUI
            else:
                r2.apply(self.formMain) 
            
        """
        ========================
        GRAPHICAL USER INTERFACE
        ========================
        """
        # Sets up and formats main figure plot inside self.fig_main
        def call_axes(self):
            self.a_main = self.fig_main.add_subplot(111)
            self.a_main.axes.get_xaxis().set_visible(False)
            self.a_main.axes.get_yaxis().set_visible(False) 
            self.a_main.margins(y=0.2)
            self.a_main.margins(x=0.2)
            self.a_main.axes.set_aspect('equal')
            self.a_main.set_xlim (-1.2,1.2)
            self.a_main.set_ylim (-1.2,1.2)
            
        # Embeds main figure (self.fig_main) and rules figure (self.fig_rules) in tkinter window
        def set_figure(self):
            self.canvas_main = FigureCanvasTkAgg(self.fig_main, self.Frame1)
            self.canvas_main.get_tk_widget().pack(fill = BOTH, expand = 1) 
            
            self.canvas_rules = FigureCanvasTkAgg(self.fig_rules, self.Frame3)
            self.canvas_rules.get_tk_widget().pack(fill = BOTH, expand = 1) 

        # Plots initial figure (black square with two points and two labels)
        def initialization (self):  
            # Sorts planes into black and white planes
            self.rearrange_planes()
            
            # Extracts data coordinates from first black plane (initial shape)
            pts = self.data_prep(self.rearranged_list[-1])
            
            # Adds plane as Polygon patch to plot
            patch = plt.Polygon(pts, facecolor = 'black')
            self.a_main.add_patch(patch)
            
            # Assigns patch as area to check for during mouseover events
            self.over_patch = patch
            
            # Adds labelled points to plot
            self.draw_point_label()
        
        # Plots rules 1 and 2 in Frame 3
        def rules_viz(self):
            ind = 1
            extra_label = ''
            for i in range (4):
                self.a_rules = self.fig_rules.add_subplot(2, 2, i+1)
                self.a_rules.axes.get_xaxis().set_visible(False)
                self.a_rules.axes.get_yaxis().set_visible(False) 
                self.a_rules.margins(y=0.2)
                self.a_rules.margins(x=0.2)
                self.a_rules.axes.set_aspect('equal')
                self.a_rules.set_xlim (-1.2,1.2)
                self.a_rules.set_ylim (-1.2,1.2)
                 
                if i <= 1:
                    ind = 1 
                    rule_point = r1
                else:
                    ind = 2
                    rule_point = r2
                
                if i%2 == 0:
                    extra_label = "LHS\n"
                    rule_side = rule_point.lhs
                else:
                    extra_label = "RHS\n"
                    rule_side =  rule_point.rhs

                self.a_rules.set_title(extra_label + 'Rule {}'.format(ind), fontsize = 11)
                
                # Plots planes in rules
                for k in rule_side.elements[1]:
                    pts = self.data_prep(k)
                    if k.attribute.elements[0].value == 'black':
                        patch = plt.Polygon(pts, facecolor = "black")
                    elif k.attribute.elements[0].value == 'white':
                        patch = plt.Polygon(pts, facecolor = 'white')
                    self.a_rules.add_patch(patch)
                
                # Plots points in rules
                for j in rule_side.elements[0]:
                    if i == 1 or i == 2:
                        self.a_rules.plot(j.value.x, j.value.y, 'o', c = 'black', markersize = 5, markeredgecolor='orange', markeredgewidth=0.5) 
                    else:
                        self.a_rules.plot(j.value.x, j.value.y, 'wo', markersize = 5, markeredgecolor='orange', markeredgewidth=0.5) 
                    # Annotates points with labels
                    self.a_rules.annotate(j.attribute.elements[0].value, xy = (j.value.x, j.value.y), xycoords='data', xytext=(-2.7, 2.7), size = 6.5, textcoords='offset points', ha = 'right', va = 'bottom', bbox = dict(boxstyle = 'round,pad=0.2', fc = 'orange')) 
                
        # Draws plane match when mouse cursor is hovering over axis
        # Called in self.on_motion()
        def viz_match (self, matchF, face_color, edge_color = 'orange'):  
            pts = self.data_prep(matchF)
            patch = plt.Polygon(pts, facecolor = face_color, edgecolor = edge_color, linewidth = 1.5)
            self.a_main.add_patch(patch)
            self.fig_main.canvas.draw_idle()
            
        # Plots result and changes frame label
        def viz_result (self):
            patchList = []
            self.rearrange_planes()
            if self.count%2 == 0: # Determines while rule has been applied
                ind = 1
            else: 
                ind = 2
                
            self.Frame1_label.configure(text = "Current Drawing, Iteration {}".format(self.count)+", applied Rule {}".format(ind))
            
            for c, i in enumerate(self.rearranged_list): # Plots form planes after rule has been applied
                pts = self.data_prep(i)
                patch = plt.Polygon(pts, facecolor = self.color_list[c])#, edgecolor = self.color_list[c])
                self.a_main.add_patch(patch)
                patchList.append(patch)
                
            # Determines which plane is the match in the next iteration
            if self.count%2 == 0: 
                self.over_patch = patchList[-1]
            else: 
                self.over_patch = patchList[len(self.blackPlanes) - 1]
            
            self.draw_point_label() # Plots points with their accompanying labels
            self.fig_main.canvas.draw_idle() # Redraws plot
        
        # Draws point/s and accompanying label/s onto plane/s
        def draw_point_label(self):
            for i in self.formMain.elements[0].elements:
                # Adjusts label size and point size depending on iteration, as squares becomes smaller
                if self.count >= 6:
                    size_label = 4.5
                    marker_size = 3
                else:
                    size_label = 6
                    marker_size = 5.5
                    
                # Changes color of point based on count number (and thus, rule being applied)    
                if self.count%2 == 0:
                    self.a_main.plot(i.value.x, i.value.y, 'o', c = 'black', markersize = marker_size, markeredgecolor='orange', markeredgewidth=0.5) 
                else:
                    self.a_main.plot(i.value.x, i.value.y, 'wo', markersize = marker_size, markeredgecolor='orange', markeredgewidth=0.5)  
                # Annotates points with label
                self.a_main.annotate(i.attribute.elements[0].value, xy = (i.value.x, i.value.y), xycoords='data', xytext=(-2.7, 2.7), size = size_label, textcoords='offset points', ha = 'right', va = 'bottom', bbox = dict(boxstyle = 'round,pad=0.2', fc = 'orange')) 
        
        # Separates black and white planes from each other and recombines them as one list for easier calling
        def rearrange_planes(self):
            self.blackPlanes = []
            self.whitePlanes = []
            color_list1 = []
            color_list2 = []
            
            # Determines which list the planes go to
            for count, plane in enumerate(list(reversed(self.formMain.elements[1].elements))):
                if plane.attribute.elements[0].value == 'black':
                    self.blackPlanes.append(plane)
                    color_list1.append('black')
                if plane.attribute.elements[0].value == 'white':
                    self.whitePlanes.append(plane) 
                    color_list2.append('white')  
                    
            # Assembles rearranged planes for plotting
            self.rearranged_list = self.blackPlanes + self.whitePlanes
            self.color_list = color_list1 + color_list2
         
        # Prepares plane vertices for plotting as patches; appends each vector coordinate pair to vertex_list  
        def data_prep(self, vertex_source):
            vertex_list = []
            for j in vertex_source.boundaries[0].polygonal:
                vertex_list.append([j.x, j.y])
            return vertex_list

        # Connects mouse event functions to the 'Current Drawing' figure (fig_main)
        def connect(self):
            self.cidpress = self.fig_main.canvas.mpl_connect('button_press_event', self.on_press)
            self.cidmotion_on = self.fig_main.canvas.mpl_connect('axes_enter_event', self.on_motion)
            self.cidmotion_off = self.fig_main.canvas.mpl_connect('axes_leave_event', self.off_motion)
         
        # Calls functions to draw result of rule application 
        def on_press(self, event):
            #if self.over_patch.contains(event)[0]:
            self.count += 1
            self.generate(self.count) # Applies rule based on current count
            self.fig_main.clf() # Clears figure
            self.call_axes() # Redraws axis
            self.viz_result() # Draws result 
           # else:
                #self.error_box()

        # Identifies which subplot the mouse is hovering over, changes edgecolor of plane match to orange
        def on_motion(self, event):
            if event.inaxes == self.a_main:
                if self.count%2 == 0:
                    self.viz_match(r2.detection.matches()[0].elements[1].elements[0], face_color = 'white') # Shows rule 2 match
                else: 
                    r1.detect(self.formMain) # Detects rule 1 match in formM1
                    self.viz_match(r1.detection.matches()[0].elements[1].elements[0], face_color = 'black') # Shows rule 1 match
                    
                
        # Removes orange color from edges of plane match when  mouse cursor is not hovering over axis
        def off_motion(self, event):
            if event.inaxes == self.a_main:
                if self.count%2 == 0:
                    self.viz_match(r2.detection.matches()[0].elements[1].elements[0], face_color = 'white', edge_color = 'white')
                else: 
                    self.viz_match(r1.detection.matches()[0].elements[1].elements[0], face_color = 'black', edge_color = 'black')
                
        # Generates error box if mouse click is not within area of match
        def error_box(self):
            error_box = Tk()
            error_box.title("Input Error")
            Label(error_box, text = "Please click directly on the match.", padx = 5, pady = 15).pack()
            Button(error_box, text= "OK", padx = 5, pady = 5, command= error_box.destroy).pack()  
            error_box.update()

        # Menu Commands: Save - saves PDF of 'Current Drawing' figure, Exit - closes window
        def menu(self):
            menubar = Menu (root)  # Saves PDF of 'Current Drawing' only
            menubar.add_command(label="Save Current Drawing", command=self.save) # Closes window
            menubar.add_command(label="Start Over", command=self.start_over) # Closes window
            menubar.add_command(label="Exit", command=root.destroy)
            root.config(menu=menubar)
            
        # When Start Over button is selected on menu
        # GUI starts from original form
        def start_over(self):
            # Resets figure to original form
            self.formMain = formM1.duplicate()
            
            # Rearranges planes in separate black and white lists
            self.rearrange_planes()
            
            # Resets iteration count
            self.count = 1
            
            
            # Drawing figure
            self.fig_main.clf() # Clears figure
            self.call_axes() # Initalizes subplot in 'fig_main' and adjusts plot settings
            self.initialization() # Plots initial form onto 'fig_main', 'a_main'
            self.connect() # Activates mouseover, mouseoff and mouseclick functions
            self.Frame1_label.configure(text =  "Current Drawing, Iteration {}".format(self.count)) # Resets Iteration label
            self.fig_main.canvas.draw_idle() # Redraws plot
            
        # Saves file of 'Current Drawing'
        # File type may be changed by change the .pdf extension to e.g.: .jpg, .png 
        def save(self):
            self.fig_main.savefig('currentdrawing.pdf', bbox_inches='tight')
            
    root = Tk()
    # Adjusts window size; width x height + x + y (x,y coordinates for where window will be placed on screen)
    root.geometry("1300x700+30+20")
    app = Application(master=root)
    # Calls and opens GUI
    app.mainloop()
    
