# -*- coding: utf-8 -*-
"""
Created on Fri Dec  2 14:20:55 2016

@author: akibdpt
"""

from sortal import (transformSet, rule, primitiveSort, weight,
                    disjunctiveSort, affineTransform, vector2D, 
                    planeSegment2D, point2D, attributeSort, coForm, label)
from sympy import pi, sin, cos
from copy import deepcopy as copy


# API Imports
from api.api_create import create # Imports create methods to use create.polygon_form to make LHS and RHS in rules
from api.sort_types import sort_types # Imports sort types
from api.form_algebra import form_algebra # Imports form algebra functions
from api.transform import transform
    
create = create()
algebra = form_algebra()

def rule5a(rotatingAngle):
    
    ###########################################################################
    ############################ LEFT HAND SIDE ###############################
    ###########################################################################
    # In the case of rotating 15 degree:
    # Consisting of: a plane segment (-1,-1)-(-1,1)-(1,1)-(1,-1) with weight 255
    #                2 points: 1 at (1,1) with label white
    #                          1 at (sqrt(3)/3,1) with label white
    
    ##### CREATING THE PLANE SEGMENT
    # Initialize the plane segment
    v1 = vector2D(-1, -1)
    v2 = vector2D(-1, 1)
    v3 = vector2D(1, 1)
    v4 = vector2D(1, -1)
    
    bound = [v1,v2,v3,v4] # Creating a list of vector defining the plane
    # Initialize the label: white
    label1 = label(sort_types['label'], 'white')
    # Initialize the other label: black
    label2 = label(sort_types['label'], 'black')
    
    # Creating the plane with a weight attribute with boundaries v1, v2, v3, v4
    formP1 = create.planeForm(sort_types['planeSort2D'], bound) 
    formP1.elements[0].addAttribute(label2) 
    
    
    
    # Initialize the two points
    # First create two point at (1,1), add one to the form first then transform the form (rotating by a degree),
    # The point inside the form will become (sqrt(3)/3,1), then add the other point at (1,1) to the form, 
    # Thus we have 1 point at (sqrt(3)/3,1) and one point at (1,1)
    point1 = point2D(sort_types['pointSort2D'], 1, 1)    
    
    # Add the points into its form(discrete)
    # Creating the discrete form for containing the points
    # Initially insert one point at (1,1) into the form
    formPt1 = create.pointForm (sort_types['pointSort2D'], 1, 1) 
    formPt1 = transform(formPt1, rotatingAngle) # Transforming the form with the transformSet defined above, the point inside will become (sqrt(3)/3,1)
    formPt1.insert(point1) # Add the other point at (1,1) into the form
    
    formPt1 = create.addAttributeAll(formPt1, label1) # Add white label to all the points inside the form
        
    ###########################################################################
    ############################ RIGHT HAND SIDE ##############################
    ###########################################################################
    # In the case of 15 degree rotation:
    # Consisting of: 5 plane segments:1: (sqrt(3)/3,1)-(1,-sqrt(3)/3)-(1,1)
    #                                 2: (-sqrt(3)/3,-1)-(1,-1)-(1,-sqrt(3)/3)
    #                                 3: (-1,sqrt(3)/3)-(sqrt(3)/3,1)-(-1,1)
    #                                 4: (-1,-1)-(-sqrt(3)/3,-1)-(-1,sqrt(3)/3)
    #                                 5: (-sqrt(3)/3,-1)-(1,-sqrt(3)/3)-(sqrt(3)/3,1)-(-1,sqrt(3)/3)
    #                                  * plane segment 1-4 have weight of 255 and plane segment 5 has weight of 0                        
    #                2 points: (-1/3 + sqrt(3)/3, 1/3 + sqrt(3)/3) with label black
    #                          (sqrt(3)/3, 1) with label black
        
    
    ##### Creating the plane segment
    """  Each of these plane segment below must be contained by a different 
         coform to be able to perform sum/difference operation
         _________________             _________________             _________________
        |(1)              |           |(2)              |           |(3)              |             
        |                 |           |                 |           |                 |         
        |                 |           |     rotated     |           |     rotated     |           
        |      black      |     -     |      black      |     +     |      white      |   
        |      (255)      |           |      (255)      |           |       (0)       |           
        |                 |           |                 |           |                 |           
        |_________________|           |_________________|           |_________________|       
    ==> After performing the operations, the coform containing the plane segment (1)
        will contain the 5 plane segments we need
    """

    # Initialize the plane boundary
    v1 = vector2D(-1, -1)
    v2 = vector2D(-1, 1)
    v3 = vector2D(1, 1)
    v4 = vector2D(1, -1)
    
    bound1 = [v1,v2,v3,v4]
    
    # Create plane segment (2) 
    # Create a coform (2) for containing the plane segment (2)
    formP2 = create.planeForm(sort_types['planeSort2D'], bound1)
    formP2 = transform(formP2, rotatingAngle) # Rotate coform (2) --> rotating plane segment (2)
    formP3 = copy(formP2) # Make a deep copy of coform (2) as coform (3)
    # We now have coform (2) and coform (3) both containing a rotated plane segment
    
    formP2 = create.addAttributeAll(formP2, label2) # Add "black" to plane segment (2)
    formP3 = create.addAttributeAll(formP3, label1) # Add "white" to plane segment (3)

    # Create plane segment (4)
    # Add weight 255 to plane segment (4)
    # Create coform (4)
    # Insert plane segment (4) into coform (4)
    formP4 = create.planeForm(sort_types['planeSort2D'], bound1)
    formP4.elements[0].addAttribute(label2)
    # Perform the operation: plane segment (1) - plane segment (2) + plane segment (3)
    formP4.difference(formP2)
    formP4.sum(formP3)
    # formP4 (coform containing plane segment (1)) now contain the 5 plane segments:
    #                                 1: (sqrt(3)/3,1)-(1,-sqrt(3)/3)-(1,1)
    #                                 2: (-sqrt(3)/3,-1)-(1,-1)-(1,-sqrt(3)/3)
    #                                 3: (-1,sqrt(3)/3)-(sqrt(3)/3,1)-(-1,1)
    #                                 4: (-1,-1)-(-sqrt(3)/3,-1)-(-1,sqrt(3)/3)
    #                                 5: (-sqrt(3)/3,-1)-(1,-sqrt(3)/3)-(sqrt(3)/3,1)-(-1,sqrt(3)/3)
    #####
    
    ##### Creating the two points
    
    # Initialize the two points
    # First create two point at (1,1), add one to the form first then transform the form (rotating by a degree),
    # The point inside the form will become (sqrt(3)/3,1), then add the other point at (1,1) to the form, 
    # Thus we have 1 point at (sqrt(3)/3,1) and one point at (1,1),
    # Rotate the form one more time,
    # We will then have 1 point at (-1/3 + sqrt(3)/3, 1/3 + sqrt(3)/3) and one point at (sqrt(3)/3, 1) 
    point2 = point2D(sort_types['pointSort2D'], 1, 1)   
    
    formPt2 = create.pointForm (sort_types['pointSort2D'], 1, 1) # Initially add one point at (1,1) to the form --> form has 1 point: (1,1)
    formPt2 = transform(formPt2, rotatingAngle) # Transform the form --> form has 1 point: (sqrt(3)/3,1)
    formPt2.insert(point2) # Add the other point at (1,1) to the form --> form has 2 points: (sqrt(3)/3,1) and (1,1)
    formPt2 = transform(formPt2, rotatingAngle) # Transform the form one more time --> form has 2 points: (-1/3 + sqrt(3)/3, 1/3 + sqrt(3)/3) and (sqrt(3)/3, 1)
    formPt2 = create.addAttributeAll(formPt2, label2) # Add black label to both points
    
    ###########################################################################
    ####################### CREATE METAFORMS AND RULE #########################
    ###########################################################################
    
    # LHS 
    forms1 = [formP1, formPt1]
    formM1 = create.metaForm(sort_types['disjunctive2D'], forms1)
    
    # RHS 
    forms2 = [formP4, formPt2]
    formM2 = create.metaForm(sort_types['disjunctive2D'], forms2)
    
    return rule('Rule 5a', 'Transforms black square into white square, rotates by an additional (rotatingAngle) degrees', formM1, formM2)
    
    
###############################################################################
###############################################################################
###############################################################################
# same as the above rule 5a, only the weights and label values are switched
def rule5b(rotatingAngle):
    
    label1 = label(sort_types['label'], 'black')
    label2 = label(sort_types['label'], 'white')
    
    v1 = vector2D(-1, -1)
    v2 = vector2D(-1, 1)
    v3 = vector2D(1, 1)
    v4 = vector2D(1, -1)
    
    bound = [v1,v2,v3,v4]

    formP1 = create.planeForm(sort_types['planeSort2D'], bound)
    formP1.elements[0].addAttribute(label2)
    

    point1 = point2D(sort_types['pointSort2D'], 1, 1) 
    
    formPt1 = create.pointForm(sort_types['pointSort2D'], 1, 1) # Initially inserts one point at (1,1) into the form
    formPt1 = transform(formPt1, rotatingAngle) # Transform the form --> form has 1 point: (sqrt(3)/3,1)
    formPt1.insert(point1)                      # Adds another point at (1,1) into the form
    formPt1 = create.addAttributeAll(formPt1, label1) # Adds 'black' label to all the points inside the form
    
    v1 = vector2D(-1, -1)
    v2 = vector2D(-1, 1)
    v3 = vector2D(1, 1)
    v4 = vector2D(1, -1)
    
    bound1 = [v1,v2,v3,v4]

    # Create plane segment (2) 
    # Create a coform (2) for containing the plane segment (2)
    formP2 = create.planeForm(sort_types['planeSort2D'], bound1)
    formP2 = transform(formP2, rotatingAngle)# Rotate coform (2) --> rotating plane segment (2)
    formP3 = copy(formP2) # Make a deep copy of coform (2) as coform (3)
    # We now have coform (2) and coform (3) both containing a rotated plane segment

    formP2 = create.addAttributeAll(formP2, label2) # Adds 'black' weight attribute to all points in formP2
    formP3 = create.addAttributeAll(formP3, label1) # Adds 'white' weight attribute to all points in formP3
    
    # Create plane segment (4)
    # Add weight 255 to plane segment (4)
    # Create coform (4)
    # Insert plane segment (4) into coform (4)
    formP4 = create.planeForm(sort_types['planeSort2D'], bound1)
    formP4.elements[0].addAttribute(label2)
    # Perform the operation: plane segment (1) - plane segment (2) + plane segment (3)
    formP4.difference(formP2)
    formP4.sum(formP3)
    # formP4 (coform containing plane segment (1)) now contain the 5 plane segments:
    #                                 1: (sqrt(3)/3,1)-(1,-sqrt(3)/3)-(1,1)
    #                                 2: (-sqrt(3)/3,-1)-(1,-1)-(1,-sqrt(3)/3)
    #                                 3: (-1,sqrt(3)/3)-(sqrt(3)/3,1)-(-1,1)
    #                                 4: (-1,-1)-(-sqrt(3)/3,-1)-(-1,sqrt(3)/3)
    #                                 5: (-sqrt(3)/3,-1)-(1,-sqrt(3)/3)-(sqrt(3)/3,1)-(-1,sqrt(3)/3)
    #####

    ##### Creating the two points
    point2 = point2D(sort_types['pointSort2D'], 1, 1)    

    formPt2 = create.pointForm(sort_types['pointSort2D'], 1, 1) # Initially add one point at (1,1) to the form --> form has 1 point: (1,1)
    formPt2 = transform(formPt2, rotatingAngle) # Transform the form --> form has 1 point: (sqrt(3)/3,1)
    formPt2.insert(point2) # Add the other point at (1,1) to the form --> form has 2 points: (sqrt(3)/3,1) and (1,1)
    formPt2 = transform(formPt2, rotatingAngle) # Add the other point at (1,1) to the form --> form has 2 points: (sqrt(3)/3,1) and (1,1)
    formPt2 = create.addAttributeAll(formPt2, label2) # Add black label to both points
    
    ###########################################################################
    ####################### CREATE METAFORMS AND RULE #########################
    ###########################################################################
    
    # LHS 
    forms1 = [formP1, formPt1]
    formM1 = create.metaForm(sort_types['disjunctive2D'], forms1)
    
    # RHS 
    forms2 = [formP4, formPt2]
    formM2 = create.metaForm(sort_types['disjunctive2D'], forms2)
    
    return rule('Rule 5b', 'Transforms white square into black square, rotates by an additional (rotatingAngle) degrees', formM1, formM2)
