# -*- coding: utf-8 -*-
"""
Created on Mon Aug 22 15:24:49 2016

@author: akibdpt
"""
from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division
from __future__ import absolute_import

from builtins import super
"""
from future import standard_library
standard_library.install_aliases()
"""
from .multiplyForm import multiplyForm
from ..ind.plane import plane2D

"""
 * INTERVAL is a behavioral category for Sorts (see 'Sort.java').
 * An interval sort is a multiply-associated sort for which
 * ...
 * Therefore, the IntervalForm class extends on the MultiplyForm
 * class (see 'MultiplyForm.java') and implements the algebra methods
 * from the Form interface (see 'Form.java').
 *
 * ASSUMPTIONS: the following algebraic property is assumed on forms
 *   a + b = b <=> a < b
 * 
"""
class cyclical(multiplyForm):
    type = 'form-cyclical'
    
    def __init__(self, sort, *args, **kwargs):
        super(cyclical, self).__init__(sort)
    
    def printE(self, string = ''):
        print(string + '|' + self.sort.name+':',len(self.elements),'elements')
        for ind in self.elements:
            print(string, end = ' _ ')
            print(ind.printVal())
            if ind.hasAttribute():
                ind.attribute.printE(string+'  ')
        return
        
    
    def insert(self, ind, start = 0):
        if not self.sort == ind.sort:
            raise ValueError('Inserting an object of sort '+ind.sort.name+' into a form of sort '+self.sort.name)
        if ind.type == self.type:
            for i in ind:
                self.insert(i)
            return
        index = start
        thisInd = self.get(index)
        while thisInd is not None and thisInd.compareInterval(ind) < 1:
            if thisInd.compareInterval(ind) == 0:
                if thisInd.hasAttribute() or ind.hasAttribute():
                    thisInd.attribute.sum(ind.attribute)
                return
            else:
                index = index+1
                thisInd = self.get(index)
        self.add(index, ind)
        self.maximal = False
        return
        
        
    def oneStepForward(self, index):
        current = self.get(index)
        nextInd = self.get(index+1)
        if (nextInd is not None and current.touches(nextInd) and 
            nextInd.attribute == current.attribute and nextInd.end != current.start):
                nextInd.combine(current)
                self.remove(current)
        else:
            index += 1
        return index


    def combineIndividuals(self, ind1, ind2):
        if ind1 is ind2:
            return
        cmpI = ind2.compareInterval(ind1)
        if cmpI == 0:
            if ind2.hasAttribute() or ind1.hasAttribute():
                ind2.attribute.sum(ind1.attribute)                
            self.remove(ind1)
        elif ind2.touches(ind1):
            if (not ind2.hasAttribute() and not ind1.hasAttribute()) or ind2.attribute == ind1.attribute:
                self.remove(ind1)
                ind2.combine(ind1)
        elif ind2.disjoint(ind1):
            return
        elif ind2.contains(ind1):
            if (not ind2.hasAttribute() and not ind1.hasAttribute()) or ind2.attribute == ind1.attribute:
                self.remove(ind1)
            else:
                temp = ind2.complement(ind1)
                if ind2.hasAttribute() or ind1.hasAttribute():
                    ind1.attribute.sum(ind2.attribute)
                    
                self.remove(ind2)
                
                for i in temp:
                    self.insert(i)
                
        else:
            if (not ind2.hasAttribute() and not ind1.hasAttribute()) or ind2.attribute == ind1.attribute:
                self.remove(ind1)
                ind2.combine(ind1)
            else:
                temp1 = ind2.complement(ind1)[0]
                temp2 = ind1.complement(ind2)[0]
                
                tempCommon = ind2.common(ind1)
                        
                self.remove(ind1)
                self.remove(ind2)
                
                self.insert(temp1)
                self.insert(temp2)
                for com in tempCommon:
                    self.insert(com)
                    
            
    def splitArcs(self):
        if len(self.elements) == 0:
            return self
        if 'Arc' not in self.elements[0].type:
            return self
        i = 0
        while i < len(self):
            rev = self.get(i)
            l = rev.splitArc()
            if l is not None:
                self.remove(rev)
                self.insert(l[0])
                self.insert(l[1])
                i += 1
            i+= 1
                        
        return self
                    
        
    """
    maximalize the form:
    _remove Nil elements
    _reorder the elements in the form in ascending order
    """
    def maximalize(self, ignoreMaximal = False):
        if self.maximal and ignoreMaximal == False:
            return

        self.splitArcs()
        
        #self.printE('ooo')

        if len(self.elements) == 0:
            return
        ind = self.first()
        
        index = 1
        nextInd = self.get(index)
        
        disorganized = []
        while nextInd is not None:
            #print('\n"""',ind.printVal(),'\n|||', nextInd.printVal(),'\n')
            if nextInd.isNil():
                self.remove(nextInd)
                nextInd = self.get(index)
                continue
            elif ind.isNil():
                self.remove(ind)
                ind = nextInd
                nextInd = self.get(index)
                continue
            cmp = ind.compareValue(nextInd)
            if cmp == 0:
                cmpI = ind.compareInterval(nextInd)
                if cmpI == 0:
                    if ind.hasAttribute() or nextInd.hasAttribute():
                        ind.attribute.sum(nextInd.attribute)                
                    self.remove(nextInd)
                    nextInd = self.get(index)
                elif ind.touches(nextInd):
                    if nextInd.end == ind.start:
                        ind = nextInd
                        index += 1
                        nextInd = self.get(index)
                        continue
                    if (not ind.hasAttribute() and not nextInd.hasAttribute()) or ind.attribute == nextInd.attribute:
                        self.remove(nextInd)
                        ind.combine(nextInd)
                        nextInd = self.get(index)
                    else:
                        if cmpI == 1:
                            self.remove(nextInd)
                            disorganized.append(nextInd)
                            nextInd = self.get(index)
                        else:
                            ind = nextInd
                            index += 1
                            nextInd = self.get(index)
                elif ind.disjoint(nextInd):
                    if cmpI == 1:
                        self.remove(nextInd)
                        disorganized.append(nextInd)
                        nextInd = self.get(index)
                    else:
                        ind = nextInd
                        index += 1
                        nextInd = self.get(index)
                    
                elif ind.contains(nextInd):
                    if (not ind.hasAttribute() and not nextInd.hasAttribute()) or ind.attribute == nextInd.attribute:
                        self.remove(nextInd)
                        nextInd = self.get(index)
                    else:
                        temp = ind.complement(nextInd)
                        if ind.hasAttribute() or nextInd.hasAttribute():
                            nextInd.attribute.sum(ind.attribute)
                            
                        self.remove(ind)
                        
                        ind = nextInd
                        
                        for i in temp:
                            self.insert(i)
                            
                        index += 1
                        nextInd = self.get(index)
                        
                elif nextInd.contains(ind):
                    if (not ind.hasAttribute() and not nextInd.hasAttribute()) or ind.attribute == nextInd.attribute:
                        self.remove(ind)
                        ind = nextInd
                        nextInd = self.get(index)
                    else:
                        temp = nextInd.complement(ind)
                        if ind.hasAttribute() or nextInd.hasAttribute():
                            ind.attribute.sum(nextInd.attribute)
                        self.remove(nextInd)
                        if temp[0].compareInterval(ind) == -1:
                            ind = temp[0]
                        for i in temp:
                            self.insert(i)
                        nextInd = self.get(index)
                        
                else:
                    if (not ind.hasAttribute() and not nextInd.hasAttribute()) or ind.attribute == nextInd.attribute:
                        self.remove(nextInd)
                        ind.combine(nextInd)
                        nextInd = self.get(index)
                    else:
                        temp1 = ind.complement(nextInd)[0]
                        temp2 = nextInd.complement(ind)[0]
                        
                        tempCommon = ind.common(nextInd)
                        self.remove(nextInd)
                        self.remove(ind)
                        
                        ind = temp1
                        self.insert(temp1)
                        self.insert(temp2)
                        for com in tempCommon:
                            com.attribute.sum(nextInd.attribute)
                            self.insert(com)
                        ind = self.get(index-1)
                        nextInd = self.get(index)

            elif cmp == 1:                
                self.remove(nextInd)
                disorganized.append(nextInd)
                nextInd = self.get(index)
            else:
                
                ind = nextInd
                index += 1
                nextInd = self.get(index)
                
        for ins in disorganized:
            self.insert(ins)
            
        for ind in self.elements:
            if ind.hasAttribute():
                ind.attribute.maximalize()   
        
        index = 0
        while index < len(self.elements):
            index = self.oneStepForward(index)
            
        self.maximal = True
        return
    
        
    
    """
    sum two forms (OR)
    e.g. formA: a, b, c
         formB:    b, c, d
        _formA.sum(formB)
      -->formA: a, b, c, d
    """
    def sum(self, other):
        if self is other:
            raise ValueError('Operation on two instances of the same form')
        if not self.sort.__eq__(other.sort):
            raise ValueError('Operation on two forms of different sorts '+str(self.sort)+'  and '+str(other.sort))
            
        other.maximalize()
        
        for i in other.elements:
            self.insert(i)
                
        self.maximalize()
        return

    #####
    #    going through the two form and split any arc segment that goes over  
    #    the anchor point (last segment of a codescriptor) and do normal 
    #    difference then go through the list once again to join the segments together
    #####
    def difference(self, other):
        if self is other:
            raise ValueError('Operation on two instances of the same form')
        if not self.sort.__eq__(other.sort):
            raise ValueError('Operation on two forms of different sorts '+str(self.sort)+'  and '+str(other.sort))
        
        self.maximalize()
        other.maximalize()
            
        thisInd = self.first()       
        otherInd = other.first()
        
        thisIndex = 0
        otherIndex = 0
        
        while thisInd is not None and otherInd is not None:
            #print('\n"""',thisInd.printVal(),'\n|||', otherInd.printVal(),'\n')
            cmp = thisInd.compareValue(otherInd) 
            if cmp == 0:
                cmpI = thisInd.compareInterval(otherInd)
                if thisInd.disjoint(otherInd) or thisInd.touches(otherInd):
                    if  cmpI == 1:                
                        otherIndex += 1
                        otherInd = other.get(otherIndex)
                    else:     
                        thisIndex += 1
                        thisInd = self.get(thisIndex)
                        
                elif cmpI == 0:
                    if thisInd.hasAttribute() or otherInd.hasAttribute():
                        thisInd.attribute.difference(otherInd.attribute)
                    if not thisInd.hasAttribute():
                        self.remove(thisInd)
                        thisInd = self.get(thisIndex)
                    otherIndex += 1
                    otherInd = other.get(otherIndex)
                    
                elif thisInd.contains(otherInd):
                    temp = thisInd.complement(otherInd)
                    if thisInd.hasAttribute() or otherInd.hasAttribute():
                        thisInd.attribute.difference(otherInd.attribute)
                        
                        if thisInd.hasAttribute():
                            self.remove(thisInd)
                            thisInd = thisInd.common(otherInd)[0]
                            self.insert(thisInd)
                            for i in temp:
                                self.insert(i)
                            thisIndex += 1
                            if thisInd == self.get(thisIndex):
                                thisIndex += 1
                                thisInd = self.get(thisIndex)
                        else:
                            self.remove(thisInd)
                            for i in temp:
                                self.insert(i)
                            thisInd = self.get(thisIndex)
                    elif not thisInd.hasAttribute():
                        self.remove(thisInd)
                        for i in temp:
                            self.insert(i)
                        thisInd = self.get(thisIndex)
   
                elif otherInd.contains(thisInd):
                    if thisInd.hasAttribute() or otherInd.hasAttribute():
                        thisInd.attribute.difference(otherInd.attribute)
                    if not thisInd.hasAttribute():
                        self.remove(thisInd)
                        thisInd = self.get(thisIndex)
                    else:
                        thisIndex += 1
                        thisInd = self.get(thisIndex)
                elif cmpI == -1:   
                    temp = thisInd.complement(otherInd)
                    if thisInd.hasAttribute() or otherInd.hasAttribute():
                        thisInd.attribute.difference(otherInd.attribute)
                    if thisInd.hasAttribute():
                        tempCommon = thisInd.common(otherInd)
                        self.remove(thisInd)
                        for com in tempCommon:
                            self.insert(com)
                        self.insert(temp[0])
                        thisIndex += 1 + len(tempCommon)
                        thisInd = self.get(thisIndex)
                    else:
                        self.remove(thisInd)
                        self.insert(temp[0])
                        thisIndex += 1
                        thisInd = self.get(thisIndex)
                else:
                    temp = thisInd.complement(otherInd)
                    if thisInd.hasAttribute() or otherInd.hasAttribute():
                        thisInd.attribute.difference(otherInd.attribute)
                    if thisInd.hasAttribute():
                        tempCommon = thisInd.common(otherInd)
                        self.remove(thisInd)
                        for com in tempCommon:
                            self.insert(com)
                        thisIndex += len(tempCommon)
                    else:
                        self.remove(thisInd)
                    self.insert(temp[0])
                    thisInd = self.get(thisIndex)
            elif cmp == 1: 
                otherIndex += 1
                otherInd = other.get(otherIndex)
            else:            
                thisIndex += 1
                thisInd = self.get(thisIndex)
        
        index = 0
        while index < len(self.elements):
            index = self.oneStepForward(index)
        return

    def product(self, other):
        if self is other:
            raise ValueError('Operation on two instances of the same form')
        if not self.sort.__eq__(other.sort):
            raise ValueError('Operation on two forms of different sorts '+str(self.sort)+'  and '+str(other.sort))
            
        self.maximalize()
        other.maximalize()
        
        thisIndex = 0
        otherIndex = 0
        thisInd = self.first()
        otherInd = other.first()
        while thisInd is not None and otherInd is not None:
            if thisInd.compareValue(otherInd) == 0:
                if thisInd.disjoint(otherInd) or thisInd.touches(otherInd):
                    if thisInd.compareInterval(otherInd) == 1:
                        otherIndex += 1
                        otherInd = other.get(otherIndex)
                    else:
                        self.pop(thisIndex)
                        thisInd = self.get(thisIndex)
                else:
                    checkSame = False
                    if thisInd.hasAttribute() and otherInd.hasAttribute():
                        temp = thisInd.duplicateInd()
                        temp.attribute.product(otherInd.attribute)
                        if temp.hasAttribute():
                            tempCommon = temp.common(otherInd)
                            for com in tempCommon:
                                self.insert(com)
                    else:
                        tempCommon = thisInd.common(otherInd)
                        for com in tempCommon:
                            if com == thisInd:
                                checkSame = True
                            else:
                                self.insert(com)
                            
                    #######
                    if thisInd.compareEnd(otherInd) < 1:
                        if not checkSame:
                            self.pop(thisIndex)
                            thisIndex -= 1
                        thisIndex += 1
                        thisInd = self.get(thisIndex)
                    else:
                        otherIndex += 1
                        otherInd = other.get(otherIndex)
  
            elif thisInd.compareValue(otherInd) == 1: 
                otherIndex += 1
                otherInd = other.get(otherIndex)
            else:         
                self.pop(thisIndex)
                thisInd = self.get(thisIndex)
                
        while thisInd is not None:   
            self.pop(thisIndex)
            thisInd = self.get(thisIndex)
            
        index = 0
        while index < len(self.elements):
            index = self.oneStepForward(index)
            
        return


        
    """
    check if this form contains a certain individual
    """
    def containsInd(self, ind):
        com = None
        if self.sort == ind.sort:
            for i in self.elements:
                if not i.compareValue(ind) == 0:
                    continue
                if i.containsInd(ind):
                    return True
                elif i.containsPoint(ind.start) and self.sort.type == 'sort-attribute':
                    if ind.hasAttribute():
                        if not ind.attribute.partOf(i.attribute):
                            continue
                    com = i.duplicateInd()
                elif com is not None:
                    if i.start == com.end and i.containsPoint(ind.end) and ind.attribute.partOf(i.attribute):
                        com.combine(i)
                        if com.contains(ind):
                            return True
                    elif i.start == com.end and ind.contains(i) and ind.attribute.partOf(i.attribute):
                        com.combine(i)
                    
        return False
        
        
    """
    check if this form is a part of other form
    """
    def partOf(self, other):
        if self is other:
            raise ValueError('Operation on two instances of the same form')
        if not self.sort == other.sort:
            if other.sort.contains(self.sort):
                for i in other:
                    if i.sort == self.sort or i.sort.contains(self.sort) and self.partOf(i):
                        return True
                return False
            raise ValueError('Operation on two forms of different sorts '+str(self.sort)+'  and '+str(other.sort))
        self.maximalize()
        other.maximalize()
        thisInd = self.first()
        otherInd = other.first()
        for thisInd in self.elements:
            check = False
            com = None
            for otherInd in other.elements:
                if thisInd == otherInd or otherInd.contains(thisInd):
                    if thisInd.hasAttribute() and not thisInd.attribute.partOf(otherInd.attribute):
                        return False
                    check = True
                    break

                elif otherInd.containsPoint(thisInd.start):
                    if thisInd.hasAttribute():
                        if not thisInd.attribute.partOf(otherInd.attribute):
                            continue
                    com = otherInd.duplicateInd()
                elif com is not None:
                    if (otherInd.start == com.end and otherInd.containsPoint(thisInd.end) 
                    and thisInd.attribute.partOf(otherInd.attribute)):
                        com.combine(otherInd)
                        if com.contains(thisInd):
                            check = True
                            break
                    elif (otherInd.start == com.end and thisInd.contains(otherInd) 
                    and thisInd.attribute.partOf(otherInd.attribute)):
                        com.combine(otherInd)

            if check is False:
                return False
        return True