# -*- coding: utf-8 -*-
"""
Created on Tue Apr 11 14:16:14 2017

@author: akibdpt
"""
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from builtins import range
from builtins import super

from colorsys import rgb_to_hsv, hsv_to_rgb
from .special import special
from ..sort.parameter.colorMode import colorMode
from ..struct.coordinate import coordinate as co, c0
from ..io.parseReader import ParsingError

class color(special):
    
    type = 'ind-color'
    DEFAULT_MODE = colorMode('RGB,')
    RGB_NOTATION = "RGB"
    HSV_NOTATION = "HSV"
    
    def __init__(self, sort, c1, c2, c3, alpha = -1, RGB = True, *args, **kwargs):
        super(color, self).__init__(sort)
        mode = sort.baseSort().arguments
        if mode is None:
            mode = self.DEFAULT_MODE
        if mode.behavior == colorMode.Behavior.RGB_ALPHA:
            if alpha == -1:
                self.alpha = mode.max_alpha
            else:
                self.alpha = co(alpha)
        self.components = [0,0,0]
        
        c1 = co(c1)
        c2 = co(c2)
        c3 = co(c3)
        
        if RGB and mode.model == colorMode.Model.HSV:
            self.components = self.RGBtoHSV(c1, c2, c3)
            print(self.components)
            self.reduceValues(mode)
        elif (not RGB) and mode.model == colorMode.Model.RGB:
            self.components = self.HSVtoRGB(c1, c2, c3)
            self.reduceValues(mode)
        else:
            self.components[0] = c1
            self.components[1] = c2
            self.components[2] = c3
            self.checkValues(mode)
            
        
            
    def __str__(self):
        if self.isNil():
            return '[NIL]'
        mode = self.sort.baseSort().arguments
        if mode is None:
            mode = self.DEFAULT_MODE
        s = '('+str(self.components[0])+', '+str(self.components[1])+', '+str(self.components[2])
        if mode.behavior == colorMode.Behavior.RGB_ALPHA:
            s += ', '+str(self.alpha)
        return s + ')'
        
    def getValue(self):
        return self.components
        
    def printVal(self):
        return str(self)
            
    def isNil(self):
        return self.components == [-1,-1,-1]
    
    def colorMode(self):
        return self.sort.baseSort().arguments
        
    def model(self):
        if self.colorMode() is not None:
            return self.colorMode().model
        return self.DEFAULT_MODE.model
            
    def behavior(self):
        if self.colorMode() is not None:
            return self.colorMode().behavior
        return self.DEFAULT_MODE.behavior
            
    def max_alpha(self):
        if self.colorMode() is not None:
            return self.colorMode().max_alpha
        return self.DEFAULT_MODE.max_alpha
            
    def max_cvalues(self):
        if self.colorMode() is not None:
            return self.colorMode().max_cvalues
        return self.DEFAULT_MODE.max_cvalues
            
    @staticmethod
    def RGBtoHSV(c1, c2, c3):
        result = rgb_to_hsv(c1.value,c2.value,c3.value)
        comp = []
        for i in result:
            comp.append(co(i))
        return comp
        
    @staticmethod
    def HSVtoRGB(c1, c2, c3):
        result = hsv_to_rgb(c1.value,c2.value,c3.value)
        comp = []
        for i in result:
            comp.append(co(i))
        return comp
        
    def reduceValues(self, mode):
        if mode.behavior == colorMode.Behavior.RGB_ALPHA:
            if self.alpha <= 0 or self.alpha > colorMode.MAX_ALPHA:
                raise ValueError("Alpha value must lie between 0 and " + str(colorMode.MAX_ALPHA))
            if self.alpha == colorMode.MAX_ALPHA:
                self.alpha = self.alpha * colorMode.max_alpha / colorMode.MAX_ALPHA
        notation = self.RGB_NOTATION
        max_cvalues = colorMode.MAX_RGB_VALUES
        if mode.model == colorMode.Model.HSV:
            notation = self.HSV_NOTATION
            max_cvalues = colorMode.MAX_HSV_VALUES
        if len(max_cvalues) > 1:
            for i in range(3):
                if self.components[i] < 0 or self.components[i] > max_cvalues[i]:
                    raise ValueError(str(notation[i])+' value must lie between 0 and ' + str(max_cvalues[i]))
        else:
            for i in range(3):
                if self.components[i] < 0 or self.components[i] > max_cvalues[0]:
                    raise ValueError(str(notation[i])+' value must lie between 0 and ' + str(max_cvalues[0]))

        modeMaxCvalues = mode.max_cvalues
        if modeMaxCvalues != max_cvalues:
            if len(max_cvalues) > 1 and len(modeMaxCvalues)> 1:
                for i in range(3):
                    self.components[i] = round(self.components[i] * modeMaxCvalues[i] / max_cvalues[i])
            elif len(max_cvalues) > 1 and len(modeMaxCvalues) == 1:
                for i in range(3):
                    self.components[i] = round(self.components[i] * modeMaxCvalues[i] / max_cvalues[i])
            elif len(max_cvalues) == 1 and len(modeMaxCvalues) > 1:
                for i in range(3):
                    self.components[i] = round(self.components[i] * max_cvalues[0] / modeMaxCvalues[i])
            else:
                for i in range(3):
                    self.components[i] = round(self.components[i] * max_cvalues[0] / modeMaxCvalues[0])
                

    def checkValues(self, mode):
        if mode.behavior == colorMode.Behavior.RGB_ALPHA and (self.alpha <= 0 or self.alpha > mode.max_alpha):
            raise ValueError("Alpha value must lie between 0 and " + str(colorMode.MAX_ALPHA))
        notation = self.RGB_NOTATION
        if mode.model == colorMode.Model.HSV:
            notation = self.HSV_NOTATION
        max_cvalues = mode.max_cvalues
        if len(max_cvalues) > 1:
            for i in range(3):
                if self.components[i] < 0 or self.components[i] > max_cvalues[i]:
                    raise ValueError(str(notation[i])+' value must lie between 0 and ' + str(max_cvalues[i]))
        else:
            for i in range(3):
                if self.components[i] < 0 or self.components[i] > max_cvalues[0]:
                    raise ValueError(str(notation[i])+' value must lie between 0 and ' + str(max_cvalues[i]))
                
    def RGB(self):
        mode = self.sort.baseSort().arguments
        if mode is None:
            mode = self.DEFAULT_MODE
        if mode.model == colorMode.Model.HSV:
            comps = self.components
            comps = self.HSVtoRGB(comps[0],comps[1],comps[2])
        else:
            comps = self.components
        return '0x'+format(int((comps[0] * 256 + comps[1]) * 256 + comps[2]), '06x')
        
    def equals(self, other):
        return self.components == other.components
        
    def compareValue(self, other):
        for i in range(3):
            c = self.components[i] - other.components[i]
            if c < 0:
                return -1
            if c > 0:
                return 1
        mode = self.sort.baseSort().arguments
        if mode is None:
            mode = self.DEFAULT_MODE
        if mode.behavior == colorMode.Behavior.RGB_ALPHA:
            c = self.alpha - other.alpha
            if c < 0:
                return -1
            if c > 0:
                return 1
        return 0
    
    def contains(self, other):
        if self.sort != other.sort:
            raise ValueError('Colors must be of the same sort')
        if self.isNil():
            return False
        if other.isNil():
            return True
        return (self.components[0] >= other.components[0] and 
                self.components[1] >= other.components[1] and
                self.components[2] >= other.components[2])
        
    def combine(self, other):
        if self.sort != other.sort:
            raise ValueError('Colors must be of the same sort')
        if self.isNil() and other.isNil():
            return False
        if other.isNil():
            return True
            
        mode = self.sort.baseSort().arguments
        if mode is None:
            mode = self.DEFAULT_MODE
            
        thisComps = self.components
        otherComps = other.components
        if mode.model == colorMode.Model.HSV:
            thisComps = self.HSVtoRGB(thisComps[0],thisComps[1],thisComps[2])
            otherComps = self.HSVtoRGB(otherComps[0],otherComps[1],otherComps[2])

        if mode.behavior == colorMode.Behavior.RGB_ALPHA:
            factor = 1 - other.alpha
            if self.alpha == 0 or other.alpha == 1:
                for i in range(3):
                    self.components[i] = otherComps[i]
                self.alpha = other.alpha
            elif self.alpha == 1:
                for i in range(3):
                    self.components[i] = round(otherComps[i] * other.alpha + thisComps[i] * factor)
            elif other.alpha != 0:
                _alpha = other.alpha + self.alpha * factor
                for i in range(3):
                    self.components[i] = round(otherComps[i] * other.alpha + thisComps[i] * factor * self.alpha) / _alpha
                self.alpha = _alpha
                
        elif mode.behavior == colorMode.Behavior.RGB_AVG:
            for i in range(3):
                self.components[i] = round((otherComps[i] + thisComps[i]) / 2)
                
        elif mode.behavior == colorMode.Behavior.RGB_SUM:
            for i in range(3):
                self.components[i] = otherComps[i] + thisComps[i]
            maxCVals = mode.max_cvalues
            if mode.model == colorMode.Model.HSV:
                maxCVals = colorMode.MAX_RGB_VALUES
            if len(maxCVals) > 1:
                for i in range(3):
                    if self.components[i] > maxCVals[i]:
                        self.components[i] = maxCVals[i]
            else:
                for i in range(3):
                    if self.components[i] > maxCVals[0]:
                        self.components[i] = maxCVals[0]
                        
        elif mode.behavior == colorMode.Behavior.RGB_MAX:
            for i in range(3):
                self.components[i] = max(otherComps[i], thisComps[i])
                
                
        if mode.model == colorMode.Model.HSV:
            self.components = self.RGBtoHSV(self.components[0], self.components[1], self.components[2])
            self.reduceValues(mode)
        
        return True


    def common(self, other):
        if self.sort != other.sort:
            raise ValueError('Colors must be of the same sort')
        if self.isNil() or other.isNil():
            self.components = [-1,-1,-1]
            return False

        mode = self.sort.baseSort().arguments
        if mode is None:
            mode = self.DEFAULT_MODE
            
        thisComps = self.components
        otherComps = other.components
        if mode.model == colorMode.Model.HSV:
            thisComps = self.HSVtoRGB(thisComps[0],thisComps[1],thisComps[2])
            otherComps = self.HSVtoRGB(otherComps[0],otherComps[1],otherComps[2])
            
        if mode.behavior == colorMode.Behavior.RGB_ALPHA:
            self.alpha = min(self.alpha, other.alpha)
            """
        elif mode.behavior == colorMode.Behavior.RGB_AVG:
            pass
        elif mode.behavior == colorMode.Behavior.RGB_SUM:
            pass
        elif mode.behavior == colorMode.Behavior.RGB_MAX:
            """
        else:
            for i in range(3):
                self.components[i] = min(otherComps[i], thisComps[i])
                
        if mode.model == colorMode.Model.HSV:
            self.components = self.RGBtoHSV(self.components[0], self.components[1], self.components[2])
            self.reduceValues(mode)
        return True
        
    def complement(self, other):
        if self.sort != other.sort:
            raise ValueError('Colors must be of the same sort')

        mode = self.sort.baseSort().arguments
        if mode is None:
            mode = self.DEFAULT_MODE
            
        thisComps = self.components
        otherComps = other.components
        if mode.model == colorMode.Model.HSV:
            thisComps = self.HSVtoRGB(thisComps[0],thisComps[1],thisComps[2])
            otherComps = self.HSVtoRGB(otherComps[0],otherComps[1],otherComps[2])
            
            
        if mode.behavior == colorMode.Behavior.RGB_ALPHA:
            factor = 1 - other.alpha
            if self.alpha == 0 or other.alpha == 1:
                for i in range(3):
                    self.components[i] = otherComps[i]
                self.alpha = other.alpha
            elif self.alpha == 1:
                for i in range(3):
                    self.components[i] = round(otherComps[i] * other.alpha + thisComps[i] * factor)
            elif other.alpha != 0:
                _alpha = other.alpha + self.alpha * factor
                for i in range(3):
                    self.components[i] = round(otherComps[i] * other.alpha + thisComps[i] * factor * self.alpha) / _alpha
                self.alpha = _alpha
                
        elif mode.behavior == colorMode.Behavior.RGB_AVG:
            for i in range(3):
                self.components[i] = round((otherComps[i] + thisComps[i]) / 2)
                
        elif mode.behavior == colorMode.Behavior.RGB_SUM:
            for i in range(3):
                if thisComps[i] < otherComps[i]:
                    self.components[i] = c0
                else:
                    self.components[i] = thisComps[i] - otherComps[i]
                
        elif mode.behavior == colorMode.Behavior.RGB_MAX:
            for i in range(3):
                if thisComps[i] <= otherComps[i]:
                     self.components[i] = c0
                
        if mode.model == colorMode.Model.HSV:
            self.components = self.RGBtoHSV(self.components[0], self.components[1], self.components[2])
            self.reduceValues(mode)
            
        if self.components == [0,0,0]:
            return False
            
        return True
        
    def replicateInd(self):
        mode = self.sort.baseSort().arguments
        if mode is None:
            mode = self.DEFAULT_MODE
        if mode.behavior == colorMode.Behavior.RGB_ALPHA:
            duplicate = color(self.sort, self.components[0],self.components[1],self.components[2], self.alpha, mode.model == colorMode.Model.RGB)
        else:
            duplicate = color(self.sort, self.components[0],self.components[1],self.components[2], RGB = mode.model == colorMode.Model.RGB)
        return duplicate
        
    @staticmethod
    def parse(reader, sort):
        mode = sort.baseSort().arguments
        if reader.token() != '(':
            color.parseError(reader,mode)
        r = reader.newValue()
        if reader.newToken() != ',':
            color.parseError(reader,mode)
        g = reader.newValue()
        if reader.newToken() != ',':
            color.parseError(reader,mode)
        b = reader.newValue()
        if mode.behavior == colorMode.Behavior.RGB_ALPHA:
            if reader.newToken() != ',' :
                color.parseError(reader,mode)
            alpha = reader.newValue()
        if reader.newToken() != ')':
            color.parseError(reader,mode)
        if mode.behavior == colorMode.Behavior.RGB_ALPHA:
            ind = color(sort, r, g, b, alpha)
        elif mode.model == colorMode.Model.HSV:
            ind = color(sort, r, g, b, False)
        else:
            ind = color(sort, r, g, b)
        reader.newToken()
        return ind
        
    @staticmethod
    def parseError(reader, mode):
        if mode.behavior == colorMode.Behavior.RGB_ALPHA:
            raise ParsingError(reader, 'color has to be in format: "(Red, Green, Blue, Alpha)"')
        elif mode.model == colorMode.Model.HSV:
            raise ParsingError(reader, 'color has to be in format: "(Hue, Saturation, Value)"')
        else:
            raise ParsingError(reader, 'color has to be in format: "(Red, Green, Blue)"')
