# -*- coding: utf-8 -*-
"""
Created on Thu Oct 12 09:56:47 2017

@author: akibdpt
"""
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from builtins import range
from builtins import super
from builtins import str


import itertools

from copy import copy
from random import randint

from .paramSet import paramSet

from .mapping import mapping
from .graphNode import graphNode3D

from ..transform.graphTransform import graphTransform3D
from ..transform.affineTransform import affineTransform3D

from ..ind.line import line3D
from ..ind.description import description
from ..ind.plane import plane3D
from ..ind.planeSegmentP import planeSegmentP3D

from ..sort.sort import sort as sortClass

from ..struct.planeStruct import bSegment, boundary
from ..struct.vector import vector3D
from ..struct.coordinate import coordinate, c0, c1, pi, cos, sin, sqrt, acos
from ..struct.key import key



ANY = 0
PERPENDICULAR = 1
PARALLEL = 2
SKEW = 3
EMBEDDED = 4

OUTSIDE = 0
BOUND = 1
INSIDE = 2
STRICT_BOUND = 3

NONE = '[None]'

vz = vector3D(0,0,0)
"""
ANY = '~'
PARALLEL = '='
PERPENDICULAR = '+'

OUTSIDE = '0'
BOUND = 'b'
INSIDE = 'i'
STRICT_BOUND = 's'

NONE = '[--------]'
"""



class graphMap3D(mapping):
    type = 'map-graphMap'    
    
    def __init__(self, *args, **kwargs):
        super(graphMap3D, self).__init__()     
        self.nodeList = graphNode3D()
        self.matrices = [[],[],[],[]]
        self.matchingNodeList = graphNode3D()
        self.matchingMatrices = [[],[],[],[]]
        self.postAnalysisProcess = True
        
        self.rhsLines = []
        
        self.predicates = []
        self.directives = []
        self.postCheck = False

    def clear(self):
        self.matchingNodeList.clear()
        self.matchingMatrices = [[],[],[],[]]
        
    @staticmethod
    def getInstance():
        return graphMap3D()
        
    def addPredicate(self, cond, keyList):
        cond = cond.lower()
        self.predicates.append([cond,keyList])
        
    def addDirective(self, cond, keyList):
        cond = cond.lower()
        self.directives.append([cond,keyList])
    
    def generateNodes(self, ind, matching = False):
        if matching:
            nodeList = self.matchingNodeList
        else:
            nodeList = self.nodeList
        if ind.type == 'ind-pointP3D':
            nodeList.addPoint(ind)
        elif ind.type == 'ind-planeSegmentP3D':
            element = ind.parametricNode()
            nodeList.addPlane(element, ind)
        elif ind.type == 'ind-lineSegmentP3D':
            element = ind.parametricNode()
            nodeList.addLine(element, ind)
        elif ind.type == 'ind-circleP3D' or ind.type == 'ind-ellipseP3D' or ind.type == 'ind-bezierCurveP3D':
            nodeList.addCurve(ind)
        return nodeList
        
    def updateMatrices(self, matching = True):
        self.updatePlaneMatrix(matching)
        self.updateLineMatrix(matching)
        self.updatePlaneLineMatrix(matching)
        
    def updatePlaneMatrix(self, matching):
        if matching is False:
            nodeList = self.nodeList
            matrices = self.matrices
        else:
            nodeList = self.matchingNodeList
            matrices = self.matchingMatrices
        matrices[0] = [[None for i in range(len(nodeList.planes))] for j in range(len(nodeList.planes))]
        for i in range(len(nodeList.planes)):
            pli = nodeList.planes[i].nodeElement
            for j in range(i+1, len(nodeList.planes)):
                plj = nodeList.planes[j].nodeElement
                edge = ANY
                if plj.parallel(pli):
                    edge = PARALLEL
                elif plj.perpendicular(pli):
                    edge = PERPENDICULAR
                    
                    
                positionOne = OUTSIDE
                positionTwo = OUTSIDE

                if edge < PARALLEL:
                    iDir, iRoot = plj.intersection(pli)
                    for k in nodeList.planes[i].forms:
                        check = False
                        for line in k.boundaries[0].canonical:
                            if iDir.parallel(line.direction) and iRoot == line.root():
                                if positionOne == BOUND:
                                    positionOne = INSIDE
                                    check = True
                                    break
                                else:
                                    positionOne = BOUND
                                    break
                            elif iDir.parallel(line.direction):
                                continue
                            else:
                                intsn = line.intersectionLine(iRoot, iDir)
                                if line.containsPoint(intsn):
                                    positionOne = INSIDE
                                    check = True
                                    break
                        if check:
                            break
                    for k in nodeList.planes[j].forms:
                        check = False
                        for line in k.boundaries[0].canonical:
                            if iDir.parallel(line.direction) and iRoot == line.root():
                                if positionTwo == BOUND:
                                    positionTwo = INSIDE
                                    check = True
                                    break
                                else:
                                    positionTwo = BOUND
                                    break
                            elif iDir.parallel(line.direction):
                                continue
                            else:
                                intsn = line.intersectionLine(iRoot, iDir)
                                if line.containsPoint(intsn):
                                    positionTwo = INSIDE
                                    check = True
                                    break
                        if check:
                            break

                matrices[0][i][j] = [edge, positionOne]
                matrices[0][j][i] = [edge, positionTwo]
                
    def updateLineMatrix(self, matching):
        if matching is False:
            nodeList = self.nodeList
            matrices = self.matrices
        else:
            nodeList = self.matchingNodeList
            matrices = self.matchingMatrices
        matrices[3] = [[None for i in range(len(nodeList.lines))] for j in range(len(nodeList.lines))]
        for i in range(len(nodeList.lines)):
            li = nodeList.lines[i].nodeElement
            for j in range(i+1, len(nodeList.lines)):
                lj = nodeList.lines[j].nodeElement
                edge = ANY
                if lj.skew(li):
                    edge = SKEW
                elif lj.parallel(li):
                    edge = PARALLEL
                elif lj.perpendicular(li):
                    edge = PERPENDICULAR
                    
                positionOne = OUTSIDE
                positionTwo = OUTSIDE

                if edge < PARALLEL:
                    intersection = lj.intersection(li)
                    newIntersection = graphNode3D.lineIntersection(intersection)
                    nodeList.lines[i].addIntersection(newIntersection)
                    nodeList.lines[j].addIntersection(newIntersection)
                    for k in nodeList.lines[i].forms:
                        if k.start == intersection or k.end == intersection:
                            if positionOne == BOUND:
                                positionOne = INSIDE
                                newIntersection.iType = 'i'
                                break
                            else:
                                positionOne = BOUND
                                newIntersection.iType = 'i'
                        elif k.containsPoint(intersection):
                            positionOne = INSIDE
                            newIntersection.iType = 'i'
                            break
                    for k in nodeList.lines[j].forms:
                        if k.start == intersection or k.end == intersection:
                            if positionTwo == BOUND:
                                positionTwo = INSIDE
                                newIntersection.iType = 'i'
                                break
                            else:
                                positionTwo = BOUND
                                newIntersection.iType = 'i'
                        elif k.containsPoint(intersection):
                            positionTwo = INSIDE
                            newIntersection.iType = 'i'
                            break
                matrices[3][i][j] = [edge, positionOne]
                matrices[3][j][i] = [edge, positionTwo]
                
            
        
    def updatePlaneLineMatrix(self, matching):
        if matching is False:
            nodeList = self.nodeList
            matrices = self.matrices
        else:
            nodeList = self.matchingNodeList
            matrices = self.matchingMatrices
        matrices[1] = [[None for i in range(len(nodeList.lines))] for j in range(len(nodeList.planes))]
        matrices[2] = [[None for i in range(len(nodeList.planes))] for j in range(len(nodeList.lines))]

        for i in range(len(nodeList.planes)):
            pli = nodeList.planes[i].nodeElement
            for j in range(len(nodeList.lines)):
                lj = nodeList.lines[j].nodeElement
                
                edge = ANY
                if pli.parallelLine(lj):
                    edge = PARALLEL
                    if lj.root == pli.pRoot or lj.root.subtract(pli.pRoot).dotproduct(pli.normal) == 0:
                        edge = EMBEDDED
                        nodeList.planes[i].addLine(nodeList.lines[j])
                elif pli.perpendicularLine(lj):
                    edge = PERPENDICULAR
                
                positionOne = OUTSIDE
                positionTwo = OUTSIDE
                
                if edge < PARALLEL:
                    intersection = pli.intersectionLine(lj)
                    newIntersection = graphNode3D.lineIntersection(intersection)
                    nodeList.planes[i].addIntersection(newIntersection)
                    for k in nodeList.planes[i].forms:
                        if k.containsPoint(intersection):
                            positionOne = INSIDE
                            break
                    for k in nodeList.lines[j].forms:
                        if k.start == intersection or k.end == intersection:
                            if positionTwo == BOUND:
                                positionTwo = INSIDE
                                newIntersection.iType = 'i'
                                break
                            else:
                                positionTwo = BOUND
                                newIntersection.iType = 'i'
                        elif k.containsPoint(intersection):
                            positionTwo = INSIDE
                            newIntersection.iType = 'i'
                            break
                    
                matrices[1][i][j] = [edge, positionOne]
                matrices[2][j][i] = [edge, positionTwo]
            
        
    def map(self, fromInd, toInd):
        self.generateNodes(toInd, matching = True)
        return True
        
    def distinguish(self, ind):
        node = self.generateNodes(ind)
        return node
        
    def potentialTransforms(self):
        if self.postAnalysisProcess:
            self.postAnalysisProcess = False
            self.nodeList.generateLinesPlanes(True)
            self.updateMatrices(matching=False)
            self.nodeList.classifyObjs(True)
            for node in self.nodeList.lines:
                node.addEndPoints()
                #node.classifyIntersections()
                node.organizeForm()
        self.matchingNodeList.generateLinesPlanes()
        self.updateMatrices()
        self.matchingNodeList.classifyObjs()
        for matchingNode in self.matchingNodeList.lines:
            matchingNode.addEndPoints()
            #matchingNode.classifyIntersections()
            matchingNode.organizeForm()

        """
        print('\n0. Plane lines')
        for j in self.nodeList.planes:
            print(len(j.lines))
            
        print('\n1. Plane matrix')
        for j in self.matrices[0]:
            for k in j:
                if k is None:
                    print(NONE, end=' ')
                else:
                    print (k,end=' ')
            print()
        print('\n2. Plane Line matrix')
        for j in self.matrices[1]:
            for k in j:
                if k is None:
                    print(NONE, end=' ')
                else:
                    print (k,end=' ')
            print()
        print('\n3. Line Plane matrix')
        for j in self.matrices[2]:
            for k in j:
                if k is None:
                    print(NONE, end=' ')
                else:
                    print (k,end=' ')
            print()
        print('\n4. Line matrix')
        for j in self.matrices[3]:
            for k in j:
                if k is None:
                    print(NONE, end=' ')
                else:
                    print (k,end=' ')
            print()
        print()
        print('\n\n-----------------------------------------------------------------\n\n')
        print('\n0. Plane lines')
        for j in self.matchingNodeList.planes:
            print(len(j.lines))
        print('\n1. Plane matrix')
        for j in self.matchingMatrices[0]:
            for k in j:
                if k is None:
                    print(NONE, end=' ')
                else:
                    print (k,end=' ')
            print()
        print('\n2. Plane Line matrix')
        for j in self.matchingMatrices[1]:
            for k in j:
                if k is None:
                    print(NONE, end=' ')
                else:
                    print (k,end=' ')
            print()
        print('\n3. Line Plane matrix')
        for j in self.matchingMatrices[2]:
            for k in j:
                if k is None:
                    print(NONE, end=' ')
                else:
                    print (k,end=' ')
            print()
        print('\n4. Line matrix')
        for j in self.matchingMatrices[3]:
            for k in j:
                if k is None:
                    print(NONE, end=' ')
                else:
                    print (k,end=' ')
            print()
        print()
        """
        
        transforms = []
        matchLists = []
        self.detectTransform(matchLists)
        
        noInt = True

        ptransf = [graphTransform3D()]
        
        if not self.resolveFreePoints(ptransf):
            return []
        for matches in matchLists:
            #print('\n------------------------------------------------------------')
            #print(matches)
            planeMatch = matches[0]
            lineMatch = matches[1]
            transf = [] # initialize a graph transform object for this particular match
            for tr in ptransf:
                transf.append(tr.duplicate())

            check = True             
            # for each node in lhs:
            singleIntNodes = []
            vectChecks = []
                
            for nodeIndex in range(len(self.nodeList.planes)):
                nodeI = self.nodeList.planes[nodeIndex]
                matchNodeI = self.matchingNodeList.planes[planeMatch[nodeIndex]]
                self.resolvePointsOnPlane(nodeI, matchNodeI, transf)
            
            for nodeIndex in range(len(self.nodeList.lines)):
                nodeI = self.nodeList.lines[nodeIndex]
                matchNodeI = self.matchingNodeList.lines[lineMatch[nodeIndex]]
                
                for j in range(nodeIndex+1, len(self.nodeList.lines)):
                    if self.matchingMatrices[3][lineMatch[nodeIndex]][lineMatch[j]][0] < PARALLEL:
                        noInt = False
                        vectFrom = self.getIntersection(nodeI, self.nodeList.lines[j]).value
                        vectTo = self.getIntersection(matchNodeI,self.matchingNodeList.lines[lineMatch[j]]).value
                        ind = 0
                        while ind < len(transf):
                            tr = transf[ind]
                            if not tr.add(vectFrom, vectTo):
                                transf.pop(ind)
                                ind -=1
                            ind += 1
                if noInt:
                    break
                if len(transf) == 0 or not self.orderingCheck(transf, nodeI):
                    check = False
                    break
                
                # if the node doesnt have any end points --> all points on line have a match --> continue onto other ndoe
                if (not nodeI.endP) and len(nodeI.linePoints) == 0:
                    continue
                
                # check if the node and matchingNode are matched in reverse
                # i.e.  node:           --A---B------C----
                #       matchingNode:   ----X--Y--Z----W------
                #       A matches with W, B matches with Z, C matches with X
                #       --> the two nodes match in reverse order
                startI = 0
                while (not transf[0].contains(nodeI.intersections[startI].value)):
                #while (not transf[0].contains(nodeI.intersections[startI].value)) or nodeI.intersections[startI].value < nodeI.forms[0].start:
                    startI += 1
                start = transf[0].apply(nodeI.intersections[startI].value)
                
                endI = -1
                while not transf[0].contains(nodeI.intersections[endI].value):
                #while not transf[0].contains(nodeI.intersections[endI].value) or nodeI.intersections[endI].value > nodeI.forms[-1].end:
                    endI -= 1
                end = transf[0].apply(nodeI.intersections[endI].value)
                
                if start == end:
                    if nodeI.intersections[startI] != nodeI.intersections[endI]:
                        check = False
                        break
                    singleIntNodes.append([nodeI, matchNodeI, startI])
                    self.addVectChecks(vectChecks, [vectFrom, vectTo])
                    if not self.vectChecking(vectChecks):
                        check = False
                        singleIntNodes = []
                        break
                    continue
                elif start > end:
                    reverse = True
                else:
                    reverse = False

                # resolve points that are on the node but not an intersection
                if len(nodeI.linePoints) > 0:
                    if len(matchNodeI.linePoints) < len(nodeI.linePoints):
                        check = False
                        break
                    self.resolvePointsOnLine(nodeI, matchNodeI, reverse, transf)
                    
                if not nodeI.endP:
                    continue
                
                nodeInts = transf[0].getCheckPoints(nodeI.nodeElement)
                lineForm = []
                specialLineForm = []
                lineIndex = 0
                while lineIndex < len(nodeI.forms):
                    if transf[0].hasCheckPointOf(nodeI.forms[lineIndex]):
                        lineForm.append(nodeI.forms[lineIndex])
                        prevCheck = lineIndex
                        while prevCheck > 0 and nodeI.forms[prevCheck-1].end == nodeI.forms[prevCheck].start:
                            lineForm.append(nodeI.forms[prevCheck-1])
                            prevCheck -= 1
                        while lineIndex < len(nodeI.forms)-1 and nodeI.forms[lineIndex].end == nodeI.forms[lineIndex+1].start:
                            lineForm.append(nodeI.forms[lineIndex+1])
                            lineIndex += 1
                    lineIndex += 1
                for line in nodeI.forms:
                    if line not in lineForm:
                        specialLineForm.append(line)
                for line in specialLineForm:
                    posMatches = self.resolveSpecialLine(line, nodeInts, reverse, nodeI, matchNodeI)
                    if len(posMatches) == 0:
                        check = False
                    else:
                        transfIndex = 0
                        while transfIndex < len(transf):
                            tr = transf.pop(transfIndex)
                            for posM in posMatches:
                                additionalTransf = tr.duplicate()
                                if not reverse:
                                    additionalTransf.add(line.start, posM.start)
                                    additionalTransf.add(line.end, posM.end)                            
                                else:
                                    additionalTransf.add(line.start, posM.end)
                                    additionalTransf.add(line.end, posM.start)
                                transf.insert(transfIndex, additionalTransf)
                                transfIndex += 1
                if not check:
                    break
                
                ind = 0
                while ind < len(transf):
                    tr = transf[ind]
                    trCheck = True
                    for line in lineForm:
                        hasMatch = self.resolveLine(line, reverse, nodeI, matchNodeI, tr)
                        if not hasMatch:
                            transf.pop(ind)
                            trCheck = False
                            break
                    if not trCheck:
                        break
                    for intersection in nodeI.intersections:
                        if intersection.iType != 'o' and not tr.contains(intersection.value):
                            transf.pop(ind)
                            ind -=1
                            break
                    ind += 1  
            if not self.checkIntersectionPoints(transf):
                continue
            if noInt:
                transf = self.resolveNoIntersectionNodes(lineMatch, transf)
                if transf is False:
                    continue
            if len(singleIntNodes) == 2 and singleIntNodes[0][0].nodeElement.intersect(singleIntNodes[1][0].nodeElement):
                self.resolveSingleIntersectionNodes([singleIntNodes[0]], transf)
                self.resolveSingleIntersectionNodes([singleIntNodes[1]], transf)
            else:
                self.resolveSingleIntersectionNodes(singleIntNodes, transf)
            for i in range(len(self.nodeList.planes)):
                planeNode = self.nodeList.planes[i]
                matchingPlaneNode = self.matchingNodeList.planes[planeMatch[i]]
                self.resolvePlanes(transf, planeNode, matchingPlaneNode, lineMatch)
                self.resolveCurves(transf, planeNode, matchingPlaneNode)
            self.resolveCurves(transf, self.nodeList, self.matchingNodeList)
            for tr in transf:
                transforms.append(tr)
        return transforms
        
    def detectTransform(self, matchLists):
        if len(self.matrices[0]) == 0:
            for subset in itertools.permutations([k for k in range(len(self.matchingNodeList.lines))], len(self.nodeList.lines)):
                matchLists.append([(),subset])
            return matchLists
            
        planeMatches = []
        for subset in itertools.permutations([k for k in range(len(self.matchingNodeList.planes))], len(self.nodeList.planes)):
            check = True
            ###########################################################
            for indI, i in enumerate(subset):
                numLines = len(self.nodeList.planes[indI].lines)
                matchingNumLines = len(self.matchingNodeList.planes[i].lines)
                if numLines > matchingNumLines:
                    check = False
                    break
            if not check:
                continue
            for indI, i in enumerate(subset):
                for indJ, j in enumerate(subset):
                    if i == j:
                        continue
                    lhsEdge = self.matrices[0][indI][indJ]
                    matchEdge = self.matchingMatrices[0][i][j]
                    
                    if ((lhsEdge[0] == PARALLEL and lhsEdge[0] != matchEdge[0])
                    or (matchEdge[0] == PARALLEL and lhsEdge[1] != OUTSIDE)
                    or (lhsEdge[0] > matchEdge[0]) 
                    or lhsEdge[1] > matchEdge[1]):
                        check = False
                        break
                if not check:
                    break
            ###########################################################
            if check:
                planeMatches.append(subset)
                
        lineMatches = []
        for subset in itertools.permutations([k for k in range(len(self.matchingNodeList.lines))], len(self.nodeList.lines)):
            check = True
            for indI, i in enumerate(subset):
                for indJ, j in enumerate(subset):
                    if i == j:
                        continue
                    lhsEdge = self.matrices[3][indI][indJ]
                    matchEdge = self.matchingMatrices[3][i][j]
                    if (lhsEdge[0] == SKEW and matchEdge[0] == SKEW):
                        continue
                    
                    if ((lhsEdge[0] == PARALLEL and lhsEdge[0] != matchEdge[0])
                    or (matchEdge[0] == PARALLEL and lhsEdge[1] != OUTSIDE)
                    or (lhsEdge[0] > matchEdge[0]) 
                    or lhsEdge[1] > matchEdge[1]):
                        check = False
                        break
                if not check:
                    break
            if check:
                lineMatches.append(subset)

        for plMatch in planeMatches:
            for lnMatch in lineMatches:
                check = True

                for indI, i in enumerate(plMatch):
                    for indJ, j in enumerate(lnMatch):
                        lhsPL = self.matrices[1][indI][indJ]
                        matchPL = self.matchingMatrices[1][i][j]
                        
                        if ((lhsPL[0] == EMBEDDED and matchPL[0] != EMBEDDED)
                        or (lhsPL[0] == PARALLEL and lhsPL[0] != matchPL[0])
                        or (matchPL[0] == PARALLEL and lhsPL[1] != OUTSIDE)
                        or (lhsPL[0] > matchPL[0]) 
                        or lhsPL[1] > matchPL[1]):
                            check = False
                            break

                        lhsLP = self.matrices[2][indJ][indI]
                        matchLP = self.matchingMatrices[2][j][i]
                        
                        if ((lhsLP[0] == EMBEDDED and matchLP[0] != EMBEDDED)
                        or (lhsLP[0] == PARALLEL and lhsLP[0] != matchLP[0])
                        or (matchLP[0] == PARALLEL and lhsLP[1] != OUTSIDE)
                        or (lhsLP[0] > matchLP[0]) 
                        or lhsLP[1] > matchLP[1]):
                            check = False
                            break
                        
                    if not check:
                        break
                if check:
                    #matchLists.append(lnMatch)
                    matchLists.append([plMatch, lnMatch])

        return matchLists
        
        
    def resolveSpecialLine(self, line, nodeInts, reverse, nodeI, matchNodeI):
        posMatch = []
        if line.start < nodeInts[0][0]:
            checkEnd = False
            for lineCheck in nodeI.forms:
                if lineCheck.containsPoint(nodeInts[0][0]) and nodeInts[0][0] != lineCheck.start:
                    checkEnd = True
            for mLine in matchNodeI.forms:
                if line.hasAttribute() and not line.attribute.checkPartOf(mLine.attribute):
                    continue
                if (not reverse) and mLine.start >= nodeInts[0][1]:
                    continue
                elif reverse and mLine.end <= nodeInts[0][1]:
                    continue
                if not checkEnd:
                    if not reverse:
                        matchStart = mLine.start
                        matchEnd = nodeInts[0][1] if mLine.containsPoint(nodeInts[0][1]) else mLine.end
                    else:
                        matchStart = nodeInts[0][1] if mLine.containsPoint(nodeInts[0][1]) else mLine.start
                        matchEnd = mLine.end
                    newLine = mLine.__class__(mLine.sort, matchStart, matchEnd)
                else:
                    if checkEnd and mLine.containsPoint(nodeInts[0][1]):
                        continue
                    newLine = mLine
                mergeable = False
                for matchIndex in range(len(posMatch)):
                    if newLine.start > posMatch[matchIndex].end or newLine.end < posMatch[matchIndex].start:
                        continue
                    if posMatch[matchIndex].squareLength() < newLine.squareLength():
                        posMatch[matchIndex] = newLine
                        mergeable = True
                if not mergeable:
                    posMatch.append(newLine)
        elif line.start > nodeInts[-1][0]:
            checkStart = False
            for lineCheck in nodeI.forms:
                if lineCheck.containsPoint(nodeInts[-1][0]) and nodeInts[-1][0] != lineCheck.end:
                    checkStart = True
            for mLine in matchNodeI.forms:
                #print(mLine)
                if line.hasAttribute() and not line.attribute.checkPartOf(mLine.attribute):
                    continue
                if (not reverse) and mLine.end <= nodeInts[-1][1]:
                    continue
                elif reverse and mLine.start >= nodeInts[-1][1]:
                    continue
                if not checkStart:
                    if not reverse:
                        matchStart = nodeInts[-1][1] if mLine.containsPoint(nodeInts[-1][1]) else mLine.start
                        matchEnd = mLine.end
                    else:
                        matchStart = mLine.start
                        matchEnd = nodeInts[-1][1] if mLine.containsPoint(nodeInts[-1][1]) else mLine.end
                    newLine = mLine.__class__(mLine.sort, matchStart, matchEnd)
                    
                else:
                    if checkStart and mLine.containsPoint(nodeInts[-1][1]):
                        continue
                    newLine = mLine
                mergeable = False
                for matchIndex in range(len(posMatch)):
                    if newLine.start > posMatch[matchIndex].end or newLine.end < posMatch[matchIndex].start:
                        continue
                    if posMatch[matchIndex].squareLength() < newLine.squareLength():
                        posMatch[matchIndex] = newLine
                        mergeable = True
                if not mergeable:
                    posMatch.append(newLine)
            #raise ValueError('.... not implemented yet ....')
        else:
            index = 1
            while line.start > nodeInts[index][0]:
                index += 1
            if reverse:
                upperBound = nodeInts[0][1]
                for indexC in range(1,index):
                    if upperBound > nodeInts[indexC][1]:
                        upperBound = nodeInts[indexC][1]
                lowerBound = nodeInts[-1][1]
                for indexC in range(index,len(nodeInts)-1):
                    if lowerBound < nodeInts[indexC][1]:
                        lowerBound = nodeInts[indexC][1]
                """
                upperBound =  nodeInts[index-1][1]
                lowerBound =  nodeInts[index][1]
                """
            else:
                lowerBound = nodeInts[0][1]
                for indexC in range(1,index):
                    if lowerBound < nodeInts[indexC][1]:
                        lowerBound = nodeInts[indexC][1]
                upperBound = nodeInts[-1][1]
                for indexC in range(index,len(nodeInts)-1):
                    if upperBound > nodeInts[indexC][1]:
                        upperBound = nodeInts[indexC][1]
                """
                upperBound =  nodeInts[index][1]
                lowerBound =  nodeInts[index-1][1]
                """
            #print(' nodeI:', line)
            #print(' lowerBound-upperBound:', lowerBound, upperBound)
            checkStart = False
            checkEnd = False
            for lineCheck in nodeI.forms:
                if lineCheck.containsPoint(nodeInts[index-1][0]) and nodeInts[index-1][0] != lineCheck.end:
                    checkStart = True
                if lineCheck.containsPoint(nodeInts[index][0]) and nodeInts[index][0] != lineCheck.start:
                    checkEnd = True
            #print('   .....___',checkStart, checkEnd)
            for mLine in matchNodeI.forms:
                if line.hasAttribute() and not line.attribute.checkPartOf(mLine.attribute):
                    continue
                if mLine.end <= lowerBound or mLine.start >= upperBound:
                    continue
                #print('   .....', mLine)
                if not (checkStart or checkEnd):
                    if not reverse:
                        matchStart = nodeInts[index-1][1] if mLine.containsPoint(nodeInts[index-1][1]) else mLine.start
                        matchEnd = nodeInts[index][1] if mLine.containsPoint(nodeInts[index][1]) else mLine.end
                    else:
                        matchStart = nodeInts[index-1][1] if mLine.containsPoint(nodeInts[index-1][1]) else mLine.end
                        matchEnd = nodeInts[index][1] if mLine.containsPoint(nodeInts[index][1]) else mLine.start
                    newLine = mLine.__class__(mLine.sort, matchStart, matchEnd)
                else:
                    if checkStart and mLine.containsPoint(nodeInts[index-1][1]):
                        continue
                    if checkEnd and mLine.containsPoint(nodeInts[index][1]):
                        continue
                    newLine = mLine
                mergeable = False
                for matchIndex in range(len(posMatch)):
                    if newLine.start > posMatch[matchIndex].end or newLine.end < posMatch[matchIndex].start:
                        continue
                    if posMatch[matchIndex].squareLength() < newLine.squareLength():
                        posMatch[matchIndex] = newLine
                        mergeable = True
                if not mergeable:
                    posMatch.append(newLine)
                    #print(posMatch)
        return posMatch
        
    def resolveLine(self, line, reverse, nodeI, matchNodeI, transf):
        checkStart = transf.contains(line.start)
        checkEnd = transf.contains(line.end)
        if checkStart and checkEnd:
            return True
        if checkStart:
            unmatchedInts = [line.end]
        elif checkEnd:
            unmatchedInts = [line.start]
        else:
            unmatchedInts = [line.start, line.end]
        
        
        hasMatch = False
        segmentInts = []


        nodeInts = []
        for nInt in nodeI.intersections:
            if nInt.endP:
                if line.containsPoint(nInt.value) and transf.contains(nInt.value):
                    pass
                else:
                    continue
            matchingInt = transf.apply(nInt.value)
            if line.start == nInt.value or line.end == nInt.value:
                alignment = BOUND
                segmentInts.append([nInt.value, matchingInt, alignment])
            elif line.containsPoint(nInt.value):
                alignment = INSIDE
                segmentInts.append([nInt.value, matchingInt, alignment])
            nodeInts.append([nInt.value, matchingInt])
            
        """
        nodeInts = transf.getCheckPoints(nodeI.nodeElement)
        for nInt in nodeI.intersections:
            if line.containsPoint(nInt[0]):
                if line.start == nInt[0] or line.end == nInt[0]:
                    alignment = BOUND
                else:
                    alignment = INSIDE
                segmentInts.append([nInt[0], nInt[1], alignment])
                
        if len(segmentInts) == 0:
            return False
        """
        
        #print('\n____ line:',line, line.attribute)
        # checking against all line segments in the matching node
        for mLine in matchNodeI.forms:
            # if line's attribute is not a part of matching line's attribute --> continue
            if line.hasAttribute() and not line.attribute.checkPartOf(mLine.attribute):
                continue
            
            #print('   ____ mLine:', mLine, mLine.attribute)
            
            # check if the matching line has the contains the matching intersections of intersections on line
            lineCheck = True
            for segInt in segmentInts:
                if (mLine.containsPoint(segInt[1]) 
                or (segInt[2] == BOUND and (mLine.start == segInt[1] or mLine.end == segInt[1]))):
                    continue
                lineCheck = False
                break
            if not lineCheck:
                continue
            hasMatch = True
            #print('     ~~~~ match ~~~~ ')
            # for each unmatched points
            for pointIndex, point in enumerate(unmatchedInts):
                posMatch = None # possible match
                orderingPriority = False
                if point < segmentInts[0][0]:
                    if point > nodeInts[0][0]:
                        if reverse:
                            # point is smaller than first matched point on segment
                            # but bigger than first matched point on node
                            # and the node is matched in reverse
                            # --> matching point must be:
                            #       bigger than first matched point on matching segment
                            #       smaller or equals to first matched point on matching node
                            #       be on the matching segment
                            orderingPriority = '>'
                            for mInt in matchNodeI.intersections:
                                if (mInt.value > segmentInts[0][1]  
                                and mInt.value <= nodeInts[0][1]
                                and mLine.containsPoint(mInt.value)):
                                    posMatch = mInt.value
                        else:
                            # point is smaller than first matched point on segment
                            # but bigger than first matched point on node
                            # and the node is NOT matched in reverse
                            # --> matching point must be:
                            #       smaller than first matched point on matching segment
                            #       bigger or equals to first matched point on matching node
                            #       be on the matching segment
                            orderingPriority = '<'
                            for mInt in reversed(matchNodeI.intersections):
                                if (mInt.value < segmentInts[0][1]  
                                and mInt.value >= nodeInts[0][1]
                                and mLine.containsPoint(mInt.value)):
                                    posMatch = mInt.value
                    else:
                        if reverse:
                            # point is smaller than first matched point on segment,
                            # also smaller than first matched point on node 
                            # and the node is matched in reverse
                            # --> matching point must be:
                            #       bigger than first matched point on matching node
                            #       be on the matching segment
                            orderingPriority = '>'
                            for mInt in matchNodeI.intersections:
                                if (mInt.value > nodeInts[0][1]  
                                and mLine.containsPoint(mInt.value)):
                                    posMatch = mInt.value
                        else:
                            # point is smaller than first matched point on segment,
                            # also smaller than first matched point on node 
                            # and the node is NOT matched in reverse
                            # --> matching point must be:
                            #       smaller than first matched point on matching node
                            #       be on the matching segment
                            orderingPriority = '<'
                            for mInt in reversed(matchNodeI.intersections):
                                if (mInt.value < nodeInts[0][1]  
                                and mLine.containsPoint(mInt.value)):
                                    posMatch = mInt.value
                elif point > segmentInts[-1][0]:
                    if point < nodeInts[-1][0]:
                        if reverse:
                            # point is bigger than last matched point on segment,
                            # but smaller than last matched point on node 
                            # and the node is matched in reverse
                            # --> matching point must be:
                            #       smaller than last matched point on matching segment
                            #       bigger or equals to last matched point on matching node
                            #       be on the matching segment
                            orderingPriority = '<'
                            for mInt in reversed(matchNodeI.intersections):
                                if (mInt.value < segmentInts[-1][1]  
                                and mInt.value >= nodeInts[-1][1] ###
                                and mLine.containsPoint(mInt.value)):
                                    posMatch = mInt.value
                        else:
                            # point is bigger than last matched point on segment,
                            # but smaller than last matched point on node 
                            # and the node is NOT matched in reverse
                            # --> matching point must be:
                            #       bigger than last matched point on matching segment
                            #       smaller or equals to last matched point on matching node
                            #       be on the matching segment
                            orderingPriority = '>'
                            for mInt in matchNodeI.intersections:
                                if (mInt.value > segmentInts[-1][1]  
                                and mInt.value <= nodeInts[-1][1] ###
                                and mLine.containsPoint(mInt.value)):
                                    posMatch = mInt.value
                    else:
                        if reverse:
                            # point is bigger than last matched point on segment,
                            # also bigger than last matched point on node 
                            # and the node is matched in reverse
                            # --> matching point must be:
                            #       smaller than last matched point on matching node
                            #       be on the matching segment
                            orderingPriority = '<'
                            for mInt in reversed(matchNodeI.intersections):
                                if (mInt.value < nodeInts[-1][1]  
                                and mLine.containsPoint(mInt.value)):
                                    posMatch = mInt.value
                        else:
                            # point is bigger than last matched point on segment,
                            # also bigger than last matched point on node 
                            # and the node is NOT matched in reverse
                            # --> matching point must be:
                            #       bigger than last matched point on matching node
                            #       be on the matching segment
                            orderingPriority = '>'
                            for mInt in matchNodeI.intersections:
                                if (mInt.value > nodeInts[-1][1]  
                                and mLine.containsPoint(mInt.value)):
                                    posMatch = mInt.value
                else:
                    # point is in between the first and last matched points on segment
                    # narrow down the range in which the point can be 
                    index = 1
                    while point > segmentInts[index][0] and index < len(segmentInts):
                        index += 1
                    # the point must be now between segmentInts[index-1] and segmentInts[index]
                    # i.e. point is bigger than point on node at (index-1) 
                    #          and smaller than point on node at (index) 
                    if reverse:
                        # the node is matched in reverse
                        # --> matching point must be:
                        #       smaller than matched point on matching node at (index-1)
                        #       bigger than matched point on matching node at (index)
                        #       be on the matching segment
                        for mInt in reversed(matchNodeI.intersections):
                            if (mInt.value < segmentInts[index-1][1]  
                            and mInt.value > segmentInts[index][1]
                            and mLine.containsPoint(mInt.value)):
                                posMatch = mInt.value
                    else:
                        # the node is NOT matched in reverse
                        # --> matching point must be:
                        #       bigger than matched point on matching node at (index-1)
                        #       smaller than matched point on matching node at (index)
                        #       be on the matching segment
                        for mInt in matchNodeI.intersections:
                            if (mInt.value > segmentInts[index-1][1]  
                            and mInt.value < segmentInts[index][1]
                            and mLine.containsPoint(mInt.value)):
                                posMatch = mInt.value
                if posMatch is None:
                    continue
                transf.add(point, posMatch, orderingPriority)
                """
                addCheck = False
                for j in range(len(nodeInts)):
                    if point < nodeInts[j][0]:
                        nodeInts.insert(j, [point, posMatch])
                        addCheck = True
                        break
                if addCheck is False:
                    nodeInts.append([point, posMatch])
                    """
        for i in unmatchedInts:
            if not transf.contains(i):
                return False
        return hasMatch
        
    def resolveSingleIntersectionNodes(self, singleIntNodes, transf):
        if len(singleIntNodes) == 0:
            return transf
            
        transf1 = copy(transf)
        for sn in singleIntNodes:
            transfIndex = 0
            while transfIndex < len(transf):
                tr = transf.pop(transfIndex)
                newTransf = self.resolveSingleIntersectionLine(sn[0],sn[1],tr,sn[2], True)
                for i in newTransf:
                    transf.insert(transfIndex, i)
                    transfIndex += 1
            transfIndex1 = 0
            while transfIndex1 < len(transf1):
                tr = transf1.pop(transfIndex1)
                newTransf = self.resolveSingleIntersectionLine(sn[0],sn[1],tr,sn[2], False)
                for i in newTransf:
                    transf1.insert(transfIndex1, i)
                    transfIndex1 += 1
        transf += transf1
        return transf
    
    def resolveSingleIntersectionLine(self, nodeI, matchNodeI, tr, anchor, reverse):
        if len(nodeI.forms) > len(matchNodeI.forms) or len(nodeI.linePoints) > len(matchNodeI.linePoints):
            return []
                        
        anchorFrom = nodeI.intersections[anchor].value
        anchorTo = tr.apply(nodeI.intersections[anchor].value)
        anchorCheck = False
        for line in nodeI.forms:
            if line.containsPoint(anchorFrom):
                anchorCheck = True
                break

        lineMatches = []
        for line in nodeI.forms:
            if line.containsPoint(anchorFrom):
                lineCheck = 0
            elif line.start > anchorFrom:
                lineCheck = 1
            else:
                lineCheck = -1
            posMatch = []
            for mLine in matchNodeI.forms:
                if line.hasAttribute() and not line.attribute.checkPartOf(mLine.attribute):
                    continue
                if not reverse:
                    if lineCheck == 0 and mLine.containsPoint(anchorTo):
                        if line.start == anchorFrom:
                            if anchorTo == mLine.end: continue
                            newLine = mLine.__class__(mLine.sort, anchorTo, mLine.end)
                        elif line.end == anchorFrom:
                            if anchorTo == mLine.start: continue
                            newLine = mLine.__class__(mLine.sort, mLine.start, anchorTo)
                        else:
                            newLine = mLine
                    elif (not anchorCheck) and lineCheck == 1 and mLine.containsPoint(anchorTo):
                        if mLine.end == anchorTo: continue
                        newLine = mLine.__class__(mLine.sort, anchorTo, mLine.end)
                    elif (not anchorCheck) and lineCheck == -1 and mLine.containsPoint(anchorTo):
                        if mLine.start == anchorTo: continue
                        newLine = mLine.__class__(mLine.sort, mLine.start, anchorTo)
                    elif lineCheck == 1 and mLine.start > anchorTo:
                        newLine = mLine
                    elif lineCheck == -1 and mLine.end < anchorTo:
                        newLine = mLine
                    else:
                        continue
                else:
                    if lineCheck == 0 and mLine.containsPoint(anchorTo):
                        if line.start == anchorFrom:
                            if anchorTo == mLine.start: continue
                            newLine = mLine.__class__(mLine.sort, mLine.start, anchorTo)
                        elif line.end == anchorFrom:
                            if anchorTo == mLine.end: continue
                            newLine = mLine.__class__(mLine.sort, anchorTo, mLine.end)
                        else:
                            newLine = mLine
                    elif (not anchorCheck) and lineCheck == 1 and mLine.containsPoint(anchorTo):
                        if mLine.start == anchorTo: continue
                        newLine = mLine.__class__(mLine.sort, mLine.start, anchorTo)
                    elif (not anchorCheck) and lineCheck == -1 and mLine.containsPoint(anchorTo):
                        if mLine.end == anchorTo: continue
                        newLine = mLine.__class__(mLine.sort, anchorTo, mLine.end)
                    elif lineCheck == 1 and mLine.end < anchorTo:
                        newLine = mLine
                    elif lineCheck == -1 and mLine.start > anchorTo:
                        newLine = mLine
                    else:
                        continue
                mergeable = False
                for matchIndex in range(len(posMatch)):
                    if newLine.start > posMatch[matchIndex].end or newLine.end < posMatch[matchIndex].start:
                        continue
                    if posMatch[matchIndex].squareLength() < newLine.squareLength():
                        posMatch[matchIndex] = newLine
                        mergeable = True
                    
                if not mergeable:
                    posMatch.append(newLine)
            
            if len(posMatch) == 0:
                return []
            lineMatches.append((line, posMatch))
            
        indexing = [0] * len(lineMatches)
        transf = []
        transfIndex = 0
        while True:
            additionalTransf = tr.duplicate()
            addCheck = True
            for l in range(len(indexing)):
                line = lineMatches[l][0]
                posM = lineMatches[l][1][indexing[l]]
                if not reverse:
                    if not (additionalTransf.add(line.start, posM.start)
                    and additionalTransf.add(line.end, posM.end)):
                        addCheck = False
                        break
                else:
                    if not (additionalTransf.add(line.start, posM.end)
                    and additionalTransf.add(line.end, posM.start)):
                        addCheck = False
                        break
                    
            if addCheck:
                transf.insert(transfIndex, additionalTransf)
            transfIndex += 1
            indexing[0] += 1
            for i in range(len(indexing)-1):
                if indexing[i] == len(lineMatches[i][1]):
                    indexing[i] = 0
                    indexing[i+1] += 1
                else:
                    break
            if indexing[-1] == len(lineMatches[-1][1]):
                break
        self.resolvePointsOnLine(nodeI, matchNodeI, reverse, transf)
        return transf
        
        
        
    def resolveNoIntersectionNodes(self, match, transf):
        if len(self.nodeList.lines) == 0:
            return transf

        transf1 = copy(transf)
            
        """
        line0 = self.nodeList.lines[0].nodeElement
        mLine0 = self.matchingNodeList.lines[match[0]].nodeElement
        intersector = matchIntersector = line0.__class__(line0.sort, vector3D(0,0), vector3D(0,1))
        if not intersector.intersect(line0):
            intersector = line0.__class__(line0.sort, vector3D(0,0), vector3D(1,0))
        if not matchIntersector.intersect(mLine0):
            matchIntersector = line0.__class__(line0.sort, vector3D(0,0), vector3D(1,0))
        """
        
        vectChecks = []
        for nodeIndex in range(len(self.nodeList.lines)):
            if (len(self.nodeList.lines[nodeIndex].forms) > len(self.matchingNodeList.lines[match[nodeIndex]].forms)
            or len(self.nodeList.lines[nodeIndex].linePoints) > len(self.matchingNodeList.lines[match[nodeIndex]].linePoints)):
                return False
                
            lineI = self.nodeList.lines[nodeIndex].nodeElement
            matchLineI = self.matchingNodeList.lines[match[nodeIndex]].nodeElement
            
            v0 = lineI.root
            v1 = matchLineI.root
            self.addVectChecks(vectChecks, [v0,v1])
        if not self.vectChecking(vectChecks):
            return False

        for nodeIndex in range(len(self.nodeList.lines)):
            
            nodeI = self.nodeList.lines[nodeIndex]
            matchNodeI = self.matchingNodeList.lines[match[nodeIndex]]
            
            transfIndex = 0
            while transfIndex < len(transf):
                tr = transf.pop(transfIndex)
                newTransf = self.resolveNoIntersectionLine(nodeI,matchNodeI,tr,True)
                for i in newTransf:
                    transf.insert(transfIndex, i)
                    transfIndex += 1
            self.resolvePointsOnLine(nodeI, matchNodeI, True, transf)
                    
            transfIndex1 = 0
            while transfIndex1 < len(transf1):
                tr = transf1.pop(transfIndex1)
                newTransf = self.resolveNoIntersectionLine(nodeI,matchNodeI,tr,False)
                for i in newTransf:
                    transf1.insert(transfIndex1, i)
                    transfIndex1 += 1
            self.resolvePointsOnLine(nodeI, matchNodeI, False, transf1)
        transf += transf1
        
        
        return transf
        
        
    def resolveNoIntersectionLine(self, nodeI, matchNodeI, tr, reverse):
        transf = []
        for i in itertools.combinations(list(range(len(matchNodeI.forms))), len(nodeI.forms)):
            check = True
            for j in range(len(i)):
                if nodeI.forms[j].sort.type == 'sort-attribute' :
                    if (not reverse) and not nodeI.forms[j].attribute.checkPartOf(matchNodeI.forms[i[j]].attribute):
                        check = False
                        break
                    elif reverse and not nodeI.forms[len(i) - j - 1].attribute.checkPartOf(matchNodeI.forms[i[j]].attribute):
                        check = False
                        break
            if not check:
                continue
            newtr = tr.duplicate()
            if not reverse:
                for j in range(len(i)):
                    newtr.add(nodeI.forms[j].start, matchNodeI.forms[i[j]].start)
                    newtr.add(nodeI.forms[j].end, matchNodeI.forms[i[j]].end)
            else:
                for j in range(len(i)):
                    k = len(i) - j - 1
                    newtr.add(nodeI.forms[k].start, matchNodeI.forms[i[j]].end)
                    newtr.add(nodeI.forms[k].end, matchNodeI.forms[i[j]].start)
            transf.append(newtr)
        return transf
        
        
    def resolveFreePoints(self, transf):
        fPoints = []
        for p in self.nodeList.points:
            if p.oType == 'f':
                fPoints.append(p.value)
        if len(fPoints) == 0:
            return True
        mFPoints = []
        for p in self.matchingNodeList.points:
            if p.oType == 'f':
                mFPoints.append(p.value)
        if len(fPoints)>len(mFPoints):
            return False
        transfIndex = 0
        while transfIndex < len(transf):
            tr = transf.pop(transfIndex)
            
            for i in itertools.permutations(mFPoints, len(fPoints)):
                newTr = tr.duplicate()
                check = True
                for j in range(len(fPoints)):
                    if not newTr.add(fPoints[j], i[j]):
                        check = False
                if not check:
                    continue
                transf.insert(transfIndex, newTr)
                transfIndex += 1
        return True

    def resolvePointsOnPlane(self, nodeI, matchNodeI, transf):
        transfIndex = 0
        while transfIndex < len(transf):
            tr = transf.pop(transfIndex)
            for subset in itertools.permutations([k for k in range(len(matchNodeI.points))], len(nodeI.points)):
                newTransf = tr.duplicate()
                check = True
                for i,j in enumerate(subset):
                    if not newTransf.add(nodeI.points[i].value, matchNodeI.points[j].value):
                        check = False
                        break
                if check:
                    transf.insert(transfIndex, newTransf)
                    transfIndex+=1        
        
    def resolvePointsOnLine(self, nodeI, matchNodeI, reverse, transf):
        for p in nodeI.linePoints:
            transfIndex = 0
            while transfIndex < len(transf):
                tr = transf.pop(transfIndex)
                nodeInts = tr.getCheckPoints(nodeI.nodeElement)
                
                if not reverse:
                    if p.value < nodeInts[0][0]:
                        posRange = (None, nodeInts[0][1])
                    elif p.value > nodeInts[-1][0]:
                        posRange = (None, nodeInts[-1][1])
                    else:
                        rangeIndex = 1
                        while p.value > nodeInts[rangeIndex][0]:
                            rangeIndex += 1
                        posRange = (nodeInts[rangeIndex-1][1], nodeInts[rangeIndex][1])
                else:
                    if p.value < nodeInts[0][0]:
                        posRange = (nodeInts[0][1], None)
                    elif p.value > nodeInts[-1][0]:
                        posRange = (nodeInts[-1][1], None)
                    else:
                        rangeIndex = 1
                        while p.value > nodeInts[rangeIndex][0]:
                            rangeIndex += 1
                        posRange = (nodeInts[rangeIndex][1], nodeInts[rangeIndex-1][1])
                posMatch = []
                for mP in matchNodeI.linePoints:
                    if ((posRange[0] is not None and mP.value<posRange[0])
                    or (posRange[1] is not None and mP.value>posRange[1])):
                        continue
                    posMatch.append(mP.value)
                    
                for i in posMatch:
                    newTransf = tr.duplicate()
                    if (newTransf.containsMatch(i) and not newTransf.contains(p.value)) or not newTransf.add(p.value, i):
                        continue
                    transf.insert(transfIndex, newTransf)
                    transfIndex+=1
                    
    def resolvePlanes(self, transf, planeNode, matchingPlaneNode, match):
        for plane in planeNode.forms:
            transfIndex = 0
            while transfIndex < len(transf):
                tr = transf.pop(transfIndex)
                for otherPlane in matchingPlaneNode.forms:
                    
                    if (otherPlane.sort != plane.sort or (plane.hasAttribute() 
                    and (not plane.attribute == otherPlane.attribute))
                    or len(plane.boundaries[0].polygonal) != len(otherPlane.boundaries[0].polygonal)):
                        continue

                    """
                    print('.............')
                    print(plane.printRhino())
                    print(otherPlane.printRhino())
                    """
                    
                    posMatch = []
                    check = True
                    for i in range(len(self.nodeList.lines)):
                        nodeI = self.nodeList.lines[i]
                        if not plane.containsLine(nodeI.nodeElement):
                            continue
    
                        polygons = []
                        for p in plane.boundaries[0].polygonal:
                            if nodeI.nodeElement.containsPoint(p):
                                polygons.append(p)
                                
                        if len(polygons) == 0:
                            continue
                        polygons = sorted(polygons)
                        checkProceed = False
                        for p in polygons:
                            if not tr.contains(p):
                                checkProceed = True
                        if not checkProceed:
                            continue
                        matchNodeI = self.matchingNodeList.lines[match[i]]
                        
                        nodeInts = tr.getCheckPoints(nodeI.nodeElement)
                        if nodeInts[0][1] > nodeInts[-1][1]:
                            reverse = True
                        else:
                            reverse = False
                            
                        mPolygons = []
                        for p in otherPlane.boundaries[0].polygonal:
                            if matchNodeI.nodeElement.containsPoint(p):
                                mPolygons.append(p)
                        if len(mPolygons) != len(polygons):
                            check = False
                            break
                        mPolygons = sorted(mPolygons)
                        if reverse:
                            mPolygons = list(reversed(mPolygons))
                        for pI in range(len(polygons)):
                            mP = tr.transform(polygons[pI])
                            if mP is not False:
                                if mPolygons[pI] == mP:
                                    continue

                                check = False
                                break
                            if polygons[pI] < nodeInts[0][0]:
                                if ((not reverse and mPolygons[pI] > nodeInts[0][1])
                                or (reverse and mPolygons[pI] < nodeInts[0][1])):
                                    check = False
                                    break
                            elif polygons[pI] > nodeInts[-1][0]:
                                if ((not reverse and mPolygons[pI] < nodeInts[-1][1])
                                or (reverse and mPolygons[pI] > nodeInts[-1][1])):
                                    check = False
                                    break
                            else:
                                intI = 0
                                while polygons[pI] > nodeInts[intI][0]:
                                    intI += 1
                                if not ((not reverse and nodeInts[intI-1][1] <= mPolygons[pI] <= nodeInts[intI][1])
                                or (reverse and nodeInts[intI][1] <= mPolygons[pI] <= nodeInts[intI-1][1])):
                                    check = False
                                    break
                            posMatch.append([polygons[pI], mPolygons[pI]])
                    if not check:
                        continue
                    newTr = tr.duplicate()
                    for pM in posMatch:
                        newTr.add(pM[0], pM[1])
                    transf.insert(transfIndex, newTr)
                    transfIndex += 1      
                    
    def resolveCurves(self, transf, planeNode, matchingPlaneNode):
        if len(planeNode.curves) == 0:
            return True
        if len(planeNode.curves) > len(matchingPlaneNode.curves):
            for i in range(len(transf)):
                transf.pop(0)
        if len(transf) == 0:
            return False
        tr = transf[0]
        circleList = []
        ellipseList = []
        bezierList = []
        insertSpotCircle = 0
        insertSpotEllipse = 0
        for i in planeNode.curves:
            containedVects = []
            if i.value.type == 'ind-bezierCurveP3D':
                bezierList.append(i)
                continue
            for j in tr.vects:
                if i.value.containsPoint(j[0]):
                    containedVects.append(j[0])
            if len(containedVects) >= 2:
                i.oType = 't'
                if i.value.type == 'ind-ellipseP3D':
                    ellipseList.insert(0,i)
                if i.value.type == 'ind-circleP3D' and not tr.contains(i.value.position(i.value.center)):
                    circleList.insert(0,i)
            elif len(containedVects) == 1:
                i.oType = 'o'
                if i.value.type == 'ind-ellipseP3D':
                    ####
                    ellipseList.append(len(ellipseList)-insertSpotEllipse,i)
                if i.value.type == 'ind-circleP3D':
                    circleList.insert(len(circleList)-insertSpotCircle,i)
            else:
                if i.value.type == 'ind-ellipseP3D':
                    ellipseList.append(i)
                    insertSpotEllipse += 1
                if i.value.type == 'ind-circleP3D':
                    circleList.append(i)
                    insertSpotCircle += 1
            i.assocObj = containedVects
        otherCircleList = []
        otherEllipseList = []
        otherBezierList = []
        for i in matchingPlaneNode.curves:
            if i.value.type == 'ind-bezierCurveP3D':
                otherBezierList.append(i)
            elif i.value.type == 'ind-circleP3D':
                otherCircleList.append(i)
            elif i.value.type == 'ind-ellipseP3D':
                otherEllipseList.append(i)
                
                
        transfIndex = 0
        while transfIndex < len(transf):
            tr = transf.pop(transfIndex)
            for i in itertools.permutations(otherBezierList, len(bezierList)):
                newTr = tr.duplicate()
                check = True
                for j in range(len(bezierList)):
                    op1 = bezierList[j].misc[0]
                    op2 = bezierList[j].misc[1]
                    op3 = bezierList[j].misc[2]
                    mp1 = i[j].misc[0]
                    mp2 = i[j].misc[1]
                    mp3 = i[j].misc[2]
                    
                    if newTr.contains(op2):
                        if newTr.apply(op2) != mp2:
                            check = False
                            break
                        mp2 = None
                    if newTr.contains(op1):
                        p1 = newTr.apply(op1)
                        if p1 == mp1:
                            mp1 = None
                        elif p1 == mp3:
                            mp3 = mp1
                            mp1 = None
                        else:
                            check = False
                            break
                    if newTr.contains(op3):
                        p3 = newTr.apply(op3)
                        if p3 == mp3:
                            mp3 = None
                        elif mp1 is not None and p3 == mp1:
                            mp1 = mp3
                            mp3 = None
                        else:
                            check = False
                            break
                    if mp1 is not None:
                        newTr.add(op1, mp1)
                    if mp2 is not None:
                        newTr.add(op2, mp2)
                    if mp3 is not None:
                        newTr.add(op3, mp3)
                if not check:
                    continue
                transf.insert(transfIndex, newTr)
                transfIndex += 1
            
            
        transfIndex = 0
        while transfIndex < len(transf):
            tr = transf.pop(transfIndex)
            for i in itertools.permutations(otherCircleList, len(circleList)):
                newTr = tr.duplicate()
                check = True
                for j in range(len(circleList)):
                    nodeCent = circleList[j].misc[0]
                    nodeStart = circleList[j].misc[1]
                    nodeEnd = circleList[j].misc[2]
                    matchCent = i[j].misc[0]
                    matchStart = i[j].misc[1]
                    matchEnd = i[j].misc[2]
                    if circleList[j].oType == 't':
                        for v0 in circleList[j].assocObj:
                            v1 = tr.apply(v0)
                            if not i[j].value.containsPoint(v1):
                                check = False
                                break
                        if not check:
                            break
                        newTr.add(nodeCent, matchCent, override = True)
                    elif circleList[j].oType == 'o':
                        v1 = tr.apply(circleList[j].assocObj[0])
                        if (not i[j].value.containsPoint(v1) 
                        or not newTr.add(nodeCent, matchCent)):
                            check = False
                            break
                        if nodeStart != circleList[j].assocObj[0] and nodeEnd != circleList[j].assocObj[0]:
                            newTr.add(nodeStart, matchStart)
                            newTr.add(nodeEnd, matchEnd)
                        else:
                            nodeStart = nodeEnd if nodeStart == circleList[j].assocObj[0] else nodeStart
                            matchStart = matchEnd if matchStart == v1 else matchStart
                            newTr.add(nodeStart, matchStart)
                    else:
                        if (not newTr.add(nodeCent, matchCent)
                        or not newTr.add(nodeStart, matchStart) 
                        or not newTr.add(nodeEnd, matchEnd)):
                            check = False
                            break
                if not check:
                    continue
                transf.insert(transfIndex, newTr)
                transfIndex += 1

        transfIndex = 0
        while transfIndex < len(transf):
            tr = transf.pop(transfIndex)
            for i in itertools.permutations(otherEllipseList, len(ellipseList)):
                newTr = tr.duplicate()
                check = True
                for j in range(len(ellipseList)):
                    nodeF1 = ellipseList[j].misc[0]
                    nodeF2 = ellipseList[j].misc[1]
                    nodeStart = ellipseList[j].misc[2]
                    nodeEnd = ellipseList[j].misc[3]
                    matchF1 = i[j].misc[0]
                    matchF2 = i[j].misc[1]
                    matchStart = i[j].misc[2]
                    matchEnd = i[j].misc[3]
                    if ellipseList[j].oType == 't':
                        for v0 in ellipseList[j].assocObj:
                            v1 = tr.apply(v0)
                            if not i[j].value.containsPoint(v1):

                                check = False
                                break
                        if not check:
                            break
                        newTr.add(nodeF1, matchF1, override = True)
                        newTr.add(nodeF2, matchF2, override = True)
                    elif ellipseList[j].oType == 'o':
                        v1 = tr.apply(ellipseList[j].assocObj[0])
                        if (not i[j].value.containsPoint(v1) 
                        or not newTr.add(nodeF1, matchF1)
                        or not newTr.add(nodeF2, matchF2)):
                            check = False
                            break
                        if nodeStart != ellipseList[j].assocObj[0] and nodeEnd != ellipseList[j].assocObj[0]:
                            newTr.add(nodeStart, matchStart)
                            newTr.add(nodeEnd, matchEnd)
                        else:
                            nodeStart = nodeEnd if nodeStart == ellipseList[j].assocObj[0] else nodeStart
                            matchStart = matchEnd if matchStart == v1 else matchStart
                            newTr.add(nodeStart, matchStart)
                    else:
                        if (not newTr.add(nodeF1, matchF1)
                        or not newTr.add(nodeF2, matchF2) 
                        or not newTr.add(nodeStart, matchStart) 
                        or not newTr.add(nodeEnd, matchEnd)):
                            check = False
                            break
                if not check:
                    continue
                transf.insert(transfIndex, newTr)
                transfIndex += 1
        return True    
        
        
    """

    """
        
        
    @staticmethod
    def getIntersection(line1, line2):
        for i in line1.intersections:
            for j in line2.intersections:
                if i == j:
                    return i
        return None
        
    def checkIntersectionPoints(self, transf):
        ind = 0
        while ind < len(transf):
            tr = transf[ind]
            for p in self.nodeList.points:
                if p.oType != 'i':
                    continue
                transformedPoint = tr.apply(p.value)
                check = False
                for mP in self.matchingNodeList.points:
                    if mP.oType != 'i':
                        continue
                    if mP.value == transformedPoint:
                        check = True
                        break
                if not check:
                    transf.pop(ind)
                    ind -= 1
                    break
            ind += 1
        if len(transf) == 0:
            return False
        return True
        
    def orderingCheck(self, transf, nodeI):
        ind = 0
        originalInts = []
        for ints in nodeI.intersections:
            if ints.iType == 'o':
                continue
            originalInts.append(ints.value)
        originalInts = sorted(originalInts)
        while ind < len(transf):
            tr = transf[ind]
            nodeInts = []
            for i in originalInts:
                nodeInts.append([i,tr.apply(i)])
            #nodeInts = tr.getCheckPoints(nodeI.nodeElement)
            if len(nodeInts) < 2:
                ind += 1
                continue
            reverse = False
            index = 1
            while index < len(nodeInts) and nodeInts[index][1] == nodeInts[index-1][1]:
                index += 1
            if index == len(nodeInts):
                ind += 1
                continue
            if nodeInts[index-1][1] < nodeInts[index][1]:
                reverse = False
            else:
                reverse = True
            for index in range(1,len(nodeInts)):
                if ((not reverse and nodeInts[index-1][1] > nodeInts[index][1])
                or (reverse and nodeInts[index-1][1] < nodeInts[index][1])):
                    transf.pop(ind)
                    ind -= 1
                    break
            ind += 1
        if len(transf) == 0 :
            return False
        return True
        
    
    def addVectChecks(self, vectChecks, item):
        for i in range(len(vectChecks)):
            if vectChecks[i][0] > item[0]:
                vectChecks.insert(i, item)
                return
        vectChecks.append(item)
                
    def vectChecking(self, vectChecks):
        if len(vectChecks) > 2:
            a = vectChecks[0][0]<vectChecks[1][0]
            b = vectChecks[0][1]<vectChecks[1][1]
            if (a and b) or not (a or b):
                reverse = False
            else:
                reverse = True
            for index in range(2, len(vectChecks)):
                a = vectChecks[index-1][0]<vectChecks[index][0]
                b = vectChecks[index-1][1]<vectChecks[index][1]
                if (a and b) or not (a or b):
                    if reverse:
                        return False
                        break
                elif not reverse:
                    return False
                    break
        return True

        
    def analyzeRHS(self, rhs, shape, transf):
        transfIndex = 0
        while transfIndex < len(transf):
            residuals = []
            if transf[transfIndex].transformable(rhs, residuals):
                transfIndex += 1
                continue
            tr = transf.pop(transfIndex)
            newTransf = [tr]
            for r in residuals:
                if r.type == 'ind-lineSegmentP3D':
                    containedPoints = []
                    
                    ######
                    check = True
                    for i in self.nodeList.lines:
                        if i.nodeElement.compareValue(r) == 0:
                            check = False
                    if check and r not in self.rhsLines:
                        self.rhsLines.append(r)
                    ######
                    
                    for i in tr.vects:
                        if r.containsPoint(i[0]):
                            containedPoints.append(i)
                    if len(containedPoints) == 1:
                        parallelLines = self.getLineProperties('p', rhs, r)
                        i = 0
                        while i < len(parallelLines):
                            if (not tr.contains(parallelLines[i].start)) or (not tr.contains(parallelLines[i].end)):
                                parallelLines.pop(i)
                            else:
                                i+=1
                        if len(parallelLines) == 0:
                            for t in newTransf:
                                if not t.contains(r.start):
                                    t.addUnresolved(r.start)
                                if not t.contains(r.end):
                                    t.addUnresolved(r.end)
                            continue
                        intersectionLines = self.getLineProperties('iE', rhs, r)
                        posMatch = []
                        if len(intersectionLines) >= 2:
                            lr = line3D(r.sort, containedPoints[0][1], containedPoints[0][1].add((tr.apply(parallelLines[0].start).subtract(tr.apply(parallelLines[0].end)))))
                            l0 = line3D(r.sort, tr.apply(intersectionLines[0].start), tr.apply(intersectionLines[0].end))
                            l1 = line3D(r.sort, tr.apply(intersectionLines[-1].start), tr.apply(intersectionLines[-1].end))
                            if not (lr.intersect(l0) and lr.intersect(l1)):
                                newTransf = []
                                continue
                            start = lr.intersection(l0)
                            end = lr.intersection(l1)
                            if start == end:
                                intersectionLines = intersectionLines[0:1]
                            else:
                                posMatch.append([[r.start, start],[r.end, end]])
                        elif len(intersectionLines) == 1:
                            lr = line3D(r.sort, containedPoints[0][1], containedPoints[0][1].add((tr.apply(parallelLines[0].start).subtract(tr.apply(parallelLines[0].end)))))
                            l0 = line3D(r.sort, tr.apply(intersectionLines[0].start), tr.apply(intersectionLines[0].end))
                            if not lr.intersect(l0):
                                newTransf = []
                                continue
                            intersectionP = lr.intersection(l0)
                            
                            for pLine in parallelLines:
                                if pLine in residuals:
                                    continue
                                
                                intPoint = pLine.intersection(intersectionLines[0])
                                intProj = r.root.add(r.direction.scale(intPoint.subtract(r.root)))
                                if r.containsPoint(intProj):
                                    containsProj = True
                                else:
                                    containsProj = False
                                mPLine = pLine.__class__(pLine.sort, tr.apply(pLine.start), tr.apply(pLine.end))
                                mIntPoint = mPLine.intersection(l0)
                                mIntProj = lr.root.add(lr.direction.scale(mIntPoint.subtract(lr.root)))
                                
                                
                                start = tr.apply(pLine.start)
                                end = tr.apply(pLine.end)

                                nStart = containedPoints[0][1].add(start.subtract(end).scale(start.subtract(containedPoints[0][1])))
                                nEnd = containedPoints[0][1].add(start.subtract(end).scale(end.subtract(containedPoints[0][1])))
                                if nStart > nEnd:
                                    nStart, nEnd = nEnd, nStart
                                if (nStart < intersectionP) != (nEnd < intersectionP):
                                    if containsProj == (mIntProj == nStart):
                                        if intersectionP != nStart:
                                            posMatch.append([[r.start, nStart],[r.end, intersectionP]])
                                    else:
                                        if intersectionP != nEnd:
                                            posMatch.append([[r.start, intersectionP],[r.end, nEnd]])
                                elif nStart > intersectionP and containsProj == (mIntProj == nEnd):
                                    if intersectionP != nEnd:
                                        posMatch.append([[r.start, intersectionP],[r.end, nEnd]])
                                elif containsProj == (mIntProj == nStart):
                                    if intersectionP != nStart:
                                        posMatch.append([[r.start, nStart],[r.end, intersectionP]])
                                elif intersectionP == mIntProj:
                                    if intersectionP != nStart:
                                        posMatch.append([[r.start, nStart],[r.end, intersectionP]])
                                    elif intersectionP != nEnd:
                                        posMatch.append([[r.start, intersectionP],[r.end, nEnd]])
                        else:
                            for pLine in parallelLines:
                                start = tr.apply(pLine.start)
                                end = tr.apply(pLine.end)
                                nStart = containedPoints[0][1].add(start.subtract(end).scale(start.subtract(containedPoints[0][1])))
                                nEnd = containedPoints[0][1].add(start.subtract(end).scale(end.subtract(containedPoints[0][1])))
                                posMatch.append([[r.start, nStart],[r.end, nEnd]])
                        lenNewTransf = len(newTransf)
                        for count in range(lenNewTransf):
                            for m in posMatch:
                                newtr = newTransf[0].duplicate()
                                check = True
                                for item in m:
                                    if not newtr.add(item[0],item[1]):
                                        check = False
                                        break
                                if check:
                                    newTransf.append(newtr)
                                newtr = newTransf[0].duplicate()
                                check = True
                                for itemI in range(len(m)):
                                    if not newtr.add(m[itemI][0],m[itemI-1][1]):
                                        check = False
                                        break
                                if check:
                                    newTransf.append(newtr)
                            newTransf.pop(0)
                    elif len(containedPoints) > 1:
                        raise NotImplementedError
                    else:
                        for t in newTransf:
                            t.addUnresolved(r.start)
                            t.addUnresolved(r.end)
                if r.type == 'ind-pointP3D':
                    for t in newTransf:
                        t.addUnresolved(r.value)
            for newTr in newTransf:
                transf.insert(transfIndex, newTr)
                transfIndex += 1
        return transf
        
    
            
    @staticmethod
    def getLineProperties(prop, element, lineSeg):
        if 'ind-lineSegment3D' == element.type or 'ind-lineSegmentP3D' == element.type:
            if (prop == 'p' and type(element) == type(lineSeg) and element != lineSeg 
            and element.parallel(lineSeg) and not element.start.colinear(lineSeg.start, lineSeg.end)):
                return element
            if prop == 'i' and type(element) == type(lineSeg) and element != lineSeg and element.intersectSegment(lineSeg):
                return element
            if prop == 'iE' and type(element) == type(lineSeg) and element != lineSeg:
                if (lineSeg.start.colinear(element.start, element.end)
                or lineSeg.end.colinear(element.start, element.end)):
                    return element
            return None
        elif 'ind-point3D' == element.type:
            if prop == 'cP' and lineSeg.containsPoint(element.value):
                return element
            return None
        elif 'form' in element.type:
            if element.type == 'form-metaForm':
                p = []
                for f in element.elements:
                    p += graphMap3D.getLineProperties(prop, f, lineSeg)
                return p
            else:
                p = []
                for l in element.elements:
                    check = graphMap3D.getLineProperties(prop, l, lineSeg)
                    if check is not None:
                        p.append(check)
                return p
            return []
            
    def getTransform(self, tfs):
        string = self.id.split("=|=")
        if len(string) > 1:
            tfs.current =  string[0]
        return tfs.retrieveTransform(graphTransform3D)
    
        
    def evaluatePredicate(self, transfSet, lhs, tlhs, form):   
        graphTransf = self.getTransform(transfSet)
        self.postCheck = True
        for cond in self.predicates:
            if cond[0] == 'maxline':
                for keyI in cond[0]:
                    ind = keyI.ind
                    transfInd = ind.transform(transfSet)
                    form.accept(self, direction = [cond[0], transfInd])
            elif cond[0] == 'bound':
                for keyI in cond[1]:
                    """
                    # bound for lhs after transformation
                    ind = keyI.ind.transform(transfSet)
                    form.accept(self, direction = [cond[0], ind, keyI.properties])
                    """
                    ind = keyI.ind.transform(transfSet)
                    transfStart = graphTransf.apply(keyI.ind.start)
                    transfEnd = graphTransf.apply(keyI.ind.end)
                    form.accept(self, direction = [cond[0], ind, keyI.properties, transfStart, transfEnd])
            elif cond[0] == 'void':
                bound = boundary()
                if cond[1][-1] == cond[1][0]:
                    cond[1].pop(-1)
                for keyI in cond[1]:
                    newInd = graphTransf.transform(keyI)
                    if newInd is False:
                        graphTransf.printTransf()
                        raise ValueError('matching for vector '+str(keyI)+' is not found')
                    bound.add(newInd)
                for i in range(len(bound.polygonal)):
                    bound.add(bSegment(bound.polygonal[i],bound.polygonal[i-1]))
                planeSeg = planeSegmentP3D(lhs.sort, [bound], True)
                lines = []
                form.accept(self, direction = [cond[0], planeSeg, lines])
                if not self.postCheck:
                    return False
                check = planeSeg.classify_line(lines)
                if not check.notInside():
                    self.postCheck = False
            elif cond[0] == 'voidof':
                sorts = cond[1][0]
                bdr = cond[1][1:]
                bound = boundary()
                if bdr[-1] == bdr[0]:
                    bdr.pop(-1)
                for keyI in bdr:
                    newInd = graphTransf.transform(keyI)
                    if newInd is False:
                        graphTransf.printTransf()
                        raise ValueError('matching for vector '+str(keyI)+' is not found')
                    bound.add(newInd)

                i = 0
                while i < len(bound.polygonal):
                    if bound.polygonal[i] == bound.polygonal[i-1]:
                        bound.polygonal.pop(i)
                    else:
                        i += 1
                    
                for i in range(len(bound.polygonal)):
                    bound.add(bSegment(bound.polygonal[i],bound.polygonal[i-1]))
                planeSeg = planeSegmentP3D(lhs.sort, [bound], True)
                lines = []
                form.accept(self, direction = [cond[0], planeSeg, lines, sorts])
                if not self.postCheck:
                    return False
                check = planeSeg.classify_line(lines)
                if not check.notInside():
                    self.postCheck = False
            elif cond[0] == 'shortest_line':
                for keyI in cond[1]:
                    ind = keyI.ind.transform(transfSet)
                    tlhs.accept(self, direction = [cond[0], ind.squareLength()])
            elif cond[0] == 'nolabel':
                for keyI in cond[1]:
                    ind = keyI.ind.transform(transfSet)
                    form.accept(self, direction = [cond[0], ind])
            if self.postCheck is False:
                return False
        return True
            
        
    def solveDirectives(self, transfSet, lhs, tlhs, rhs, form, param = None):
        transf = [transfSet]
        if param is None:
            param = paramSet()
        param.lhs = tlhs
        param.rhs = rhs
        
        for cond in self.directives:
            if cond[0] != 'distance':
                continue
            i = 0
            while i < len(cond[1]):
                if len(cond[1][i]) > 4:
                    i+=1
                    continue
                for j in range(i+1, len(cond[1])):
                    if cond[1][i][1].ind == cond[1][j][1].ind:
                        dir2 = cond[1].pop(j)
                        cond[1][i].append(None)
                        for k in dir2:
                            cond[1][i].append(k)
                        break
                i+=1
                
        for cond in self.directives:
            if cond[0] == 'point_on_line':
                for i0 in range(len(transf)):
                    tfSet = transf.pop(0)
                    graphTransf = self.getTransform(tfSet)
                    for p in range(len(cond[1])):
                        bounds = self.evaluateDesc(cond[1][p], param)
                        if bounds[-2].__class__ != coordinate:
                            bounds.append(bounds[-1])
                        if len(bounds) == 2:
                            continue
                        for i in range(0,len(bounds)-2):
                            obj = cond[1][p][i].ind
                            self.solvePointOnLine(graphTransf, obj, bounds)
                    transf.append(tfSet)
            elif cond[0] == 'normal':
                for p in range(len(cond[1])):
                    for i0 in range(len(transf)):
                        tfSet = transf.pop(0)
                        newSet = tfSet.duplicate()
                        tr = self.getTransform(newSet)
                        bounds = self.evaluateDesc(cond[1][p], param)
                        obj = bounds[0].ind
                        obj = obj.transform(newSet)
                        
                        
                        if len(bounds) == 5:
                            length = obj.length()
                        elif len(bounds) == 6:
                            length = bounds[-1]
                        elif len(bounds) == 7:
                            rand = randint(0, 100000)
                            length = (bounds[-1]-bounds[-2])*rand/100000 + bounds[-2]
                            
                        objSort = sortClass.getSort(bounds[1])
                        if objSort.type in (None, 'sort-disjunctiveSort','sort-compoundSort'):
                            return []
                        start = tr.apply(vector3D(bounds[2], bounds[3], bounds[4]))
                        normal = obj.normal
                        end1 = start.add(normal.scale(length))
                        end2 = start.subtract(normal.scale(length))
                        newLine1 = objSort.baseSort().characteristic(objSort, start, end1)
                        newLine2 = objSort.baseSort().characteristic(objSort, start, end2)
                        tr.addExtra(newLine1)
                        tr2 = self.getTransform(tfSet)
                        tr2.addExtra(newLine2)
                        transf.append(tfSet)
                        transf.append(newSet)
                        
                """
                for p in range(len(cond[1])):
                    for i0 in range(len(transf)):
                        tfSet = transf.pop(0)
                        ntfSet = tfSet.duplicate()
                        tr = self.getTransform(ntfSet)
                        print('???',tr.unresolved)
                        bounds = self.evaluateDesc(cond[1][p], param)
                        print('!!!',bounds)
                        if len(bounds) == 4:
                            length = 1
                        elif len(bounds) == 5:
                            length = bounds[-2]
                        elif len(bounds) == 6:
                            rand = randint(0, 100000)
                            length = (bounds[-2]-cond[-3])*rand/100000 + bounds[-3]
                        obj = bounds[1].ind
                        obj = obj.transform(tfSet)
                        normal = obj.normal.unit()
                        start = tr.apply(vector3D(bounds[0], bounds[1], bounds[2]))
                        end = start.add(normal.scale(length))
                """
            elif cond[0] == 'angle':
                for p in range(len(cond[1])):
                    for i0 in range(len(transf)):
                        tfSet = transf.pop(0)
                        bounds = self.evaluateDesc(cond[1][p], param)
                        pl = None
                        if len(self.nodeList.planes) > 0:
                            for i in self.nodeList.planes:
                                if not (i.nodeElement.containsLine(bounds[1].ind) and i.nodeElement.containsLine(bounds[0].ind)):
                                    continue
                                pl = i.nodeElement.transform(tfSet)
                        if pl is None:
                            b1 = bounds[0].ind.transform(tfSet)
                            vects = [b1.start, b1.end, b1.end.add(vector3D(1,0,0))]
                            if vects[-1].colinear(vects[0],vects[1]):
                                vects[-1] = b1.end.add(vector3D(0,1,0))
                            pl = plane3D(bounds[1].ind.sort.getSort('__dummyPlane3D__'), vects)
                            
                        posMatch = self.solveAngle(tfSet, bounds, lhs, pl)
                        for j in posMatch:
                            ntfSet = tfSet.duplicate()
                            tr = self.getTransform(ntfSet)
                            adding = True
                            count = 0
                            for i in j:
                                if count >=2:
                                    tr.add(i[0],i[1])
                                    continue
                                present = False
                                for uI in range(len(tr.unresolved)):
                                    u = tr.unresolved[uI]
                                    if u == i[0]:
                                        tr.add(i[0],i[1])
                                        tr.unresolved.pop(uI)
                                        present = True
                                        break
                                count += 1
                                if not present:
                                    vCheck = False
                                    for vec in tr.vects:
                                        if vec[0] == i[0] and vec[1] == i[1]:
                                            vCheck = True
                                            break
                                    if not vCheck:
                                        adding = self.unifyDirectives(i, j, tr)
                            if adding:
                                tr.addDirectiveResult(j)
                                transf.append(ntfSet)
                            """
                            ntfSet = tfSet.duplicate()
                            tr = ntfSet.retrieveTransform(graphTransform3D)
                            for i in j:
                                for uI in range(len(tr.unresolved)):
                                    u = tr.unresolved[uI]
                                    if u == i[0]:
                                        tr.add(i[0],i[1])
                                        tr.unresolved.pop(uI)
                                        break
                            transf.append(ntfSet)
                            """                            

            elif cond[0] == 'distance':
                for p in range(len(cond[1])):
                    for i0 in range(len(transf)):
                        tfSet = transf.pop(0)
                        bounds = self.evaluateDesc(cond[1][p], param)
                        if len(bounds) > 4:
                            cond1 = []
                            cond2 = []
                            index = 0
                            while bounds[index] is not None:
                                cond1.append(bounds[index])
                                index += 1
                            index += 1
                            while index < len(bounds):
                                cond2.append(bounds[index])
                                index += 1
                            d1 = self.evaluateBounds(cond1, tfSet)
                            d2 = self.evaluateBounds(cond2, tfSet)

                            if 'lineSegment' in cond1[1].ind.type:
                                
                                pl = None
                                if len(self.nodeList.planes) > 0:
                                    for i in self.nodeList.planes:
                                        if i.nodeElement.containsLine(cond1[1].ind) and i.nodeElement.containsPoint(cond1[0].ind.value) and i.nodeElement.containsPoint(cond2[0].ind.value):
                                            pl = i.nodeElement.transform(tfSet)
                                            break
                                if pl is not None:
                                    posMatch = self.solveDoublePointLineDistance(tfSet, d1, d2, cond1[0].ind, cond2[0].ind, cond1[1].ind, pl)
                                else:
                                    posMatch = self.solveDoublePointLineDistance(tfSet, d1, d2, cond1[0].ind, cond2[0].ind, cond1[1].ind, None)
                                
                            elif cond1[0].ind.type != cond2[0].ind.type:
                                posMatch = self.solvePointPointLinePointDistance(tfSet, d1, d2, cond1[0].ind, cond2[0].ind, cond1[1].ind)
                            elif 'point' in cond1[0].ind.type:
                                pl = None
                                if len(self.nodeList.planes) > 0:
                                    for i in self.nodeList.planes:
                                        if i.nodeElement.containsPoint(cond1[1].ind.value) and i.nodeElement.containsPoint(cond1[0].ind.value) and i.nodeElement.containsPoint(cond2[0].ind.value):
                                            pl = i.nodeElement.transform(tfSet)
                                            break
                                if pl is not None:
                                    posMatch = self.solveDoublePointPointDistance(tfSet, d1, d2, cond1[0].ind, cond2[0].ind, cond1[1].ind, pl)
                                else:
                                    posMatch = self.solveDoublePointPointDistance(tfSet, d1, d2, cond1[0].ind, cond2[0].ind, cond1[1].ind, None)
                            elif 'lineSegment' in cond1[0].ind.type:
                                pl = None
                                if len(self.nodeList.planes) > 0:
                                    for i in self.nodeList.planes:
                                        if i.nodeElement.containsPoint(cond1[1].ind.value) and i.nodeElement.containsLine(cond1[0].ind) and i.nodeElement.containsLine(cond2[0].ind):
                                            pl = i.nodeElement.transform(tfSet)
                                            break
                                if pl is not None:
                                    posMatch = self.solveDoubleLinePointDistance(tfSet, d1, d2, cond1[0].ind, cond2[0].ind, cond1[1].ind, pl)
                                else:
                                    posMatch = []
                        
                        elif 'lineSegment' in cond[1][p][0].ind.type and 'lineSegment' in cond[1][p][1].ind.type:
                            pl = None
                            if len(self.nodeList.planes) > 0:
                                for i in self.nodeList.planes:
                                    if i.nodeElement.containsLine(bounds[1].ind):
                                        pl = i.nodeElement.transform(tfSet)
                                        if i.nodeElement.containsLine(bounds[0].ind):
                                            check = True
                                        else:
                                            check = False
                                        break
                            if pl is not None:
                                posMatch = self.solveLineLineDistance(tfSet, bounds, pl, check)
                            else:
                                posMatch = []
                        elif 'lineSegment' in cond[1][p][0].ind.type and 'point' in cond[1][p][1].ind.type:
                            pl = None
                            if len(self.nodeList.planes) > 0:
                                for i in self.nodeList.planes:
                                    if not (i.nodeElement.containsPoint(bounds[1].ind.value) and i.nodeElement.containsLine(bounds[0].ind)):
                                        continue
                                    pl = i.nodeElement.transform(tfSet)
                            posMatch = self.solveLinePointDistance(tfSet, bounds, pl)
                        elif 'point' in cond[1][p][0].ind.type and 'lineSegment' in cond[1][p][1].ind.type:
                            pl = None
                            if len(self.nodeList.planes) > 0:
                                for i in self.nodeList.planes:
                                    if i.nodeElement.containsLine(bounds[1].ind):
                                        pl = i.nodeElement.transform(tfSet)
                                        if i.nodeElement.containsPoint(bounds[0].ind.value):
                                            check = True
                                        else:
                                            check = False
                                        break
                            if pl is not None:
                                posMatch = self.solvePointLineDistance(tfSet, bounds, pl, check)
                            else:
                                posMatch = []
                            
                        elif 'point' in cond[1][p][0].ind.type and 'point' in cond[1][p][1].ind.type:
                            posMatch = self.solvePointPointDistance(tfSet, bounds)
                        for j in posMatch:
                            ntfSet = tfSet.duplicate()
                            tr = self.getTransform(ntfSet)
                            for i in j:
                                for uI in range(len(tr.unresolved)):
                                    u = tr.unresolved[uI]
                                    if u == i[0]:
                                        tr.add(i[0],i[1], override = True)
                                        tr.unresolved.pop(uI)
                                        break
                            transf.append(ntfSet)
                
        for directive in self.directives:
            if 'length' == directive[0]:
                for p in range(len(directive[1])):
                    for i0 in range(len(transf)):
                        tfSet = transf.pop(0)
                        bounds = self.evaluateDesc(directive[1][p], param)
                        posMatch = self.readjustLength(tfSet, bounds, lhs, rhs)
                        for j in posMatch:
                            ntfSet = tfSet.duplicate()
                            tr = self.getTransform(ntfSet)
                            for i in j:
                                tr.add(i[0],i[1], override = True)
                            transf.append(ntfSet)
        i0 = 0
        while i0 < len(transf):
            if len(self.getTransform(transf[i0]).unresolved) > 0:
                tfSet = transf.pop(i0)
                graphTransf = self.getTransform(tfSet)
                lhs.accept(self, direction = ['point_on_line', graphTransf, True])
                if len(graphTransf.unresolved) > 0:
                    continue
                transf.append(tfSet)
            i0 += 1
        return transf
        
    def unifyDirectives(self, coordMatch, posMatch, transf):
        if len(coordMatch) < 2:
            return False
        otherResults = transf.getDirectiveResults(coordMatch)
        intersection = None
        for i in otherResults:
            l1 = line3D(sortClass.getSort('__dummyLine3D__'), i[0][1], i[1][1])
            l2 = line3D(sortClass.getSort('__dummyLine3D__'), posMatch[0][1], posMatch[1][1])
            if not l1.intersect(l2):
                continue
            ints = l1.intersection(l2)
            if intersection is None:
                intersection = ints
            elif intersection != ints:
                return False
        if intersection is None:
            return False
        transf.add(coordMatch[0], intersection, override = True)
        return True

    def visitEnter(self, obj, direction = 0):
        if direction[0] == 'maxline':
            if obj.__class__ == direction[1].__class__ and obj.contains(direction[1]) and obj != direction[1] :
                self.postCheck = False
                return False
        elif direction[0] == 'bound':
            if obj.__class__ == direction[1].__class__ and obj.contains(direction[1]):
                if 0 in direction[2] and obj.start != direction[3] and obj.end != direction[3]:
                    self.postCheck = False
                    return False
                if 1 in direction[2] and obj.start != direction[4] and obj.end != direction[4]:
                    self.postCheck = False
                    return False
        elif direction[0] == 'void':
            if 'ind-planeSegment' in obj.type :
                if not obj.compareValue(direction[1]) == 0:
                    return True
                check = obj.classify(direction[1])
                if check.disjoint() or check.touches():
                    return True
                self.postCheck = False
                return False
            if 'ind-lineSegment' in obj.type:
                if direction[1].containsLine(obj):
                    direction[2].append(bSegment(obj.start, obj.end))
            if 'ind-point' in obj.type:
                check = False
                for line in direction[1].boundaries[0].canonical:
                    if line.containsPoint(obj.value):
                        check = True
                if check:
                    return True
                if direction[1].boundaries[0].containsPointInside(obj.value):
                    self.postCheck = False
                    return False
        elif direction[0] == 'voidof':
            if 'ind-planeSegment' in obj.type :
                if not (obj.sort.name in direction[-1] or obj.sort.baseSort().name in direction[-1]):
                    return True
                if ('2D' in obj.type) == ('3D' in direction[1].type):
                    raise ValueError('There exists a mix of 2D and 3D objects in the shape')
                check = obj.classify(direction[1])
                if check.disjoint() or check.touches():
                    return True
                self.postCheck = False
                return False
            if 'ind-lineSegment' in obj.type:
                if not (obj.sort.name in direction[-1] or obj.sort.baseSort().name in direction[-1]):
                    return True
                if direction[1].containsLine(obj):
                    direction[2].append(bSegment(obj.start, obj.end))
            if 'ind-point' in obj.type:
                if not (obj.sort.name in direction[-1] or obj.sort.baseSort().name in direction[-1]):
                    return True
                check = False
                for line in direction[1].boundaries[0].canonical:
                    if line.containsPoint(obj.value):
                        check = True
                if check:
                    return True
                if direction[1].boundaries[0].containsPointInside(obj.value):
                    self.postCheck = False
                    return False
        elif direction[0] == 'shortest_line':
            if 'ind-lineSegment' in obj.type and obj.squareLength() < direction[1]:
                self.postCheck = False
                return False
        elif direction[0] == 'nolabel':
            if 'ind-lineSegment' in obj.type and obj.type == direction[1].type and obj.contains(direction[1]):
                if obj.hasAttribute():
                    obj.attribute.accept(self, direction = ['nolabel*'])
            if 'ind-planeSegment' in obj.type and obj.type == direction[1].type and obj.contains(direction[1]):
                if obj.hasAttribute():
                    obj.attribute.accept(self, direction = ['nolabel*'])
            if 'ind-point' in obj.type and obj.type == direction[1].type and obj == direction[1]:
                if obj.hasAttribute():
                    obj.attribute.accept(self, direction = ['nolabel*'])
        elif direction[0] == 'nolabel*':
            if obj.type == 'ind-label':
                self.postCheck = False
                return False
        elif direction[0] == 'point_on_line':
            if 'ind-lineSegment' in obj.type:
                bounds = [c0, c1]
                self.solvePointOnLine(direction[1], obj, bounds, direction[2])
        elif direction[0] == 'containsCheck':
            if 'ind-lineSegment' in obj.type:
                if len(direction) > 3 and (direction[3] is not None) and (obj == direction[3]):
                    return False
                elif obj.start == direction[1] or obj.end == direction[1]:
                    direction[2][0] = True
            if 'ind-point' in obj.type:
                if obj.value == direction[1]:
                    direction[2][0] = True
        return True
        
    def visitLeave(self, obj, direction = 0):
        return
        
        
    @staticmethod
    def ccw(pl,A0,B0,C0):
        """
        if pl is None:
            pl = plane3D(sort.getSort('__dummyPlane3D__'), [A0,B0,C0])
        """
        A = pl.localize(A0)
        B = pl.localize(B0)
        C = pl.localize(C0)
        return (C.y-A.y) * (B.x-A.x) > (B.y-A.y) * (C.x-A.x)
        
    @staticmethod
    def evaluateDesc(boundList, param):
        newList = []
        for i in boundList:
            if isinstance(i, description):
                val = i.evaluateInd(param).getValue()[1:-1]
                newList.append(coordinate(val))
            else:
                newList.append(i)
        return newList
        
    @staticmethod
    def evaluateVariable(varList, tfs):
        if varList[0] == 'distance':
            key1 = key.retrieveKey(varList[1]).ind.edit(tfs)
            key2 = key.retrieveKey(varList[2]).ind.edit(tfs)
            if 'point' in key2.type and 'lineSegment' in key1.type:
                key1, key2 = key2, key1
            if key1.type == key2.type:
                if 'point' in key1.type:
                    return key1.value.subtract(key2.value).length()
                elif 'lineSegment' in key1.type:
                    if key1.parallel(key2):
                        return key1.root.subtract(key2.projection(key1.root)).length()
            else:
                return key1.value.subtract(key2.projection(key1.value)).length()
        elif varList[0] == 'angle':
            key1 = key.retrieveKey(varList[1]).ind.edit(tfs)
            key2 = key.retrieveKey(varList[2]).ind.edit(tfs)
            if 'lineSegment' not in key1.type or 'lineSegment' not in key2.type:
                return 0.01
            v0 = key1.direction
            v1 = key2.direction
            nom = v0.x*v1.x + v0.y*v1.y + v0.z*v1.z
            dem = ((v0.x*v0.x + v0.y*v0.y + v0.z*v0.z)*(v1.x*v1.x + v1.y*v1.y + v1.z*v1.z)).sqrt()
            if dem == c0:
                raise ZeroDivisionError('One of the vectors are (0,0) --> division by zero')
            cosTheta = nom / dem
            return coordinate(acos(cosTheta.value))
        return 0
    
    @staticmethod
    def evaluateBounds(bounds, tfSet):
        if len(bounds) == 4:
            if isinstance(bounds[-1], list):
                b1 = graphMap3D.evaluateVariable(bounds[-1], tfSet)
            else:
                b1 = bounds[-1]
            if isinstance(bounds[-2], list):
                b0 = graphMap3D.evaluateVariable(bounds[-2], tfSet)
            else:
                b0 = bounds[-2]
            rand = randint(0, 100000)
            distance = (b1-b0)*rand/100000 + b0
        elif len(bounds) == 3:
            if isinstance(bounds[-1], list):
                distance = graphMap3D.evaluateVariable(bounds[-1], tfSet)
            else:
                distance = bounds[-1]
        return distance
        
    @staticmethod
    def solvePointOnLine(graphTransf, obj, bounds, adaptive = False):
        i = 0
        while i < len(graphTransf.unresolved):
            u = graphTransf.unresolved[i]
            startP = obj.start
            endP = obj.end
            if obj.containsPoint(u) and (not startP == u) and (not endP == u):
                if adaptive:
                    for j in graphTransf.vects:
                        if obj.containsPoint(j[0]):
                            if j[0] < u:
                                if j[0] > startP:
                                    startP = j[0]
                            else:
                                if j[0] < endP:
                                    endP = j[0]
                
                graphTransf.unresolved.pop(i)
                nStart = graphTransf.apply(startP)
                nEnd = graphTransf.apply(endP)
                if obj.reverse:
                    nStart, nEnd = nEnd, nStart
                rand = randint(0, 100000)
                ratio = (bounds[-1]-bounds[-2])*rand/100000 + bounds[-2]
                p = nStart.add(nEnd.subtract(nStart).scale(ratio))
                graphTransf.add(u, p)
            else:
                i += 1
                
    def solveAngleEnd(self, comparand, intList, pos, reverse):
        if reverse:
            pos = -1 if pos == 0 else 0
        intersect = intList.pop(pos)
        while comparand == intersect:
            intersect = intList.pop(pos)
        return intersect

                
    def solveAngle(self, tfSet, bounds, lhs, cPlane):
        # getting angle from the directive range
        angle = self.evaluateBounds(bounds, tfSet)
        while angle > 180:
            angle -= 360
        angle = angle * pi / 180 
        
        base = bounds[0].ind
        new = bounds[1].ind
        
        graphTransf = self.getTransform(tfSet)
        
        intsn = base.intersection(new)
        
        
        
        
        if intsn == base.end:
            angle = angle - pi
        
        anchorVect = None
        containedPoints = []
        for i in graphTransf.vects:
            if i[0].subtract(new.root).parallel(new.direction):
                containedPoints.append(i[0])
            if anchorVect is None and not i[0].colinear(base.start, base.end):
                anchorVect = i
            if anchorVect is not None and len(containedPoints) > 0:
                break
        if len(containedPoints) == 0:
            raise ValueError('line with tag '+bounds[1].name+' does not pass through any point/line intersection point/line end point on the lhs')
        
        intStart = None
        intEnd = None
        intersections = {}
        for i in self.nodeList.lines:
            if i.nodeElement.intersect(new):
                ints = i.nodeElement.intersection(new)
                if ints == new.start:
                    intStart = (ints, i.forms[0])
                if ints == new.end:
                    intEnd = (ints, i.forms[0])
                if new.containsPoint(ints):
                    intersections[ints] = i.forms[0]
                                    
        nBaseS = graphTransf.apply(base.start)
        nBaseE = graphTransf.apply(base.end)
                
        if nBaseS == nBaseE:
            return []
            
        reverse = (nBaseS > nBaseE) 
        
        p0 =  cPlane.position(cPlane.localize(vz))
        d1 = cPlane.localize(nBaseE.subtract(nBaseS))
            
        cp = containedPoints[0]
        nCP = graphTransf.apply(cp)
                
        
        if anchorVect is not None and self.ccw(cPlane, base.end, base.start, anchorVect[0]) != self.ccw(cPlane, nBaseE, nBaseS, anchorVect[1]):
            angle = -angle
            
        ca = cos(angle.value)
        sa = sin(angle.value)
        
        d2 = cPlane.position(d1.__class__(ca*d1.x-sa*d1.y, sa*d1.x+ca*d1.y)).subtract(p0)
        
        
        pM = []
        
        
        if graphTransf.contains(new.start) and graphTransf.contains(new.end):
            nLen = (nBaseE.subtract(nBaseS).length() * new.length()) / (base.length() * d2.length())
            cStart = False
            cEnd = False
            for i in graphTransf.directiveResults:
                if i[0][0] == new.start or i[1][0] == new.start:
                    cStart = True
                elif i[0][0] == new.end or i[1][0] == new.end:
                    cEnd = True
                if cStart and cEnd:
                    if new.start == intsn:
                        pLine = new.__class__(new.sort, nCP, nCP.add(d2))
                        nBase = base.edit(tfSet)
                        nStart = nBase.intersection(pLine)
                        nEnd = nStart.add(d2.scale(nLen))
                    elif new.end == intsn:
                        pLine = new.__class__(new.sort, nCP, nCP.add(d2))
                        nBase = base.edit(tfSet)
                        nEnd = nBase.intersection(pLine)
                        nStart = nEnd.add(d2.scale(nLen))
                    else:
                        nLen = nLen / 2
                        pLine = new.__class__(new.sort, nCP, nCP.add(d2))
                        nBase = base.edit(tfSet)
                        intersection = pLine.intersection(nBase)
                        nStart = intersection.subtract(d2.scale(nLen))
                        nEnd = intersection.add(d2.scale(nLen))
                elif cStart:
                    nEnd = graphTransf.apply(new.end)
                    nStart = nEnd.add(d2.scale(nLen))
                elif cEnd:
                    nStart = graphTransf.apply(new.start)
                    nEnd = nStart.add(d2.scale(nLen))
                break
            pM.insert(0,(new.end, nEnd))
            pM.insert(0,(new.start, nStart))
            posMatch = []
            posMatch.append(pM)
            return posMatch
            
        newLine = line3D(new.sort, nCP, nCP.add(d2))
        nBase = base.edit(tfSet)
        nIntsn = newLine.intersection(nBase)
        
        if intStart is not None and intEnd is not None:
            nSLine = intStart[1].edit(tfSet)
            if nSLine.intersect(newLine):
                nStart = nSLine.intersection(newLine)
            nELine = intEnd[1].edit(tfSet)
            if nELine.intersect(newLine):
                nEnd = nELine.intersection(newLine)
        else:
            newints = []
            for intersection0 in intersections:
                if graphTransf.contains(intersection0):
                    intersection = graphTransf.apply(intersection0)
                    newints.append(intersection)
                else:
                    nIntLine = intersections[intersection0].edit(tfSet) 
                    if not nIntLine.intersect(newLine):
                        continue
                    intersection = nIntLine.intersection(newLine)
                    newints.append(intersection)
                    if intersection0 == new.start or intersection0 == new.end:
                        continue
                    pM.append((intersection0,intersection))
            newints = sorted(newints)
            dirCheck = d2 > vz
            if intsn <= new.start or intsn >= new.end:
                index = 0
                while index < len(newints):
                    if newints[index] == nIntsn or (newints[index] > nIntsn) == dirCheck:
                        index += 1
                    else:
                        newints.pop(index)
            if len(newints) >= 2:
                if intStart is not None:
                    nSLine = intStart[1].edit(tfSet)
                    if nSLine.intersect(newLine):
                        nStart = nSLine.intersection(newLine)
                    if not graphTransf.contains(new.end):
                        nEnd = self.solveAngleEnd(nStart, newints, -1, reverse)
                elif intEnd is not None:
                    nELine = intEnd[1].edit(tfSet)
                    if nELine.intersect(newLine):
                        nEnd = nELine.intersection(newLine)
                    if not graphTransf.contains(new.start):
                        nStart = self.solveAngleEnd(nEnd, newints, 0, reverse)
                elif graphTransf.contains(new.start):
                    nStart = graphTransf.apply(new.start)
                    nEnd = self.solveAngleEnd(nStart, newints, -1, reverse)
                elif graphTransf.contains(new.end):
                    nEnd = graphTransf.apply(new.end)
                    nStart = self.solveAngleEnd(nEnd, newints, 0, reverse)
                else:
                    nStart, nEnd = newints[0], newints.pop(-1)
                    while nEnd == nStart:
                        nEnd = newints.pop(-1)
                    if reverse:
                        nStart, nEnd = nEnd , nStart
            else:
                nLen = (nBaseE.subtract(nBaseS).length() * new.length()) / (base.length() * d2.length())
                if graphTransf.contains(new.start):
                    nStart = graphTransf.apply(new.start)
                    nEnd = nStart.add(d2.scale(nLen))
                elif graphTransf.contains(new.end):
                    nEnd = graphTransf.apply(new.end)
                    nStart = nEnd.add(d2.scale(nLen))
                elif new.start == intsn:
                    pLine = new.__class__(new.sort, nCP, nCP.add(d2))
                    nBase = base.edit(tfSet)
                    nStart = nBase.intersection(pLine)
                    nEnd = nStart.add(d2.scale(nLen))
                elif new.end == intsn:
                    pLine = new.__class__(new.sort, nCP, nCP.add(d2))
                    nBase = base.edit(tfSet)
                    nEnd = nBase.intersection(pLine)
                    nStart = nEnd.add(d2.scale(nLen))
                else:
                    nLen = nLen / 2
                    pLine = new.__class__(new.sort, nCP, nCP.add(d2))
                    nBase = base.edit(tfSet)
                    intersection = pLine.intersection(nBase)
                    nStart = intersection.subtract(d2.scale(nLen))
                    nEnd = intersection.add(d2.scale(nLen))
        
            
        pM.insert(0,(new.end, nEnd))
        pM.insert(0,(new.start, nStart))
        posMatch = []
        posMatch.append(pM)
        
        return posMatch
                
    """
    def solveAngle(self, tfSet, bounds, lhs, cPlane):
        angle = self.evaluateBounds(bounds, tfSet)
        while angle > 180:
            angle -= 360
        angle = angle * pi / 180 
        
        base = bounds[0].ind
        new = bounds[1].ind
        graphTransf = self.getTransform(tfSet)
        
        anchorVect = None
        containedPoints = []
        for i in graphTransf.vects:
            if i[0].subtract(new.root).parallel(new.direction):
                containedPoints.append(i[0])
            if anchorVect is None and not i[0].colinear(base.start, base.end):
                anchorVect = i
        if len(containedPoints) == 0:
            raise ValueError('')
                
        intersections = []
        for i in self.nodeList.lines:
            if i.nodeElement.intersect(new):
                ints = i.nodeElement.intersection(new)
                if new.containsPoint(ints):
                    intersections.append([ints, i.forms[0]])
                    
        
        intsn = base.intersection(new)
        
        delayBase = False
        if base.start >= intsn:
            baseDir = True
            p0 = base.end
        elif base.end <= intsn:
            baseDir = False
            p0 = base.start
        else:
            delayBase = True
        
        if new.start >= intsn:
            p1 = new.end
        elif new.end <= intsn:
            p1 = new.start
        else:
            if delayBase:
                p0 = base.start
                delayBase = False
            endProj = base.projection(new.end)
            if (abs(angle)*2 < pi) == ((endProj > intsn) == (p0 > intsn)):
                p1 = new.end
            else:
                p1 = new.start
            
        if delayBase:
            p1Proj = base.projection(p1)
            if (abs(angle)*2 < pi) == ((p1Proj > intsn) == (base.start > intsn)):
                p0 = base.start
                baseDir = False
            else:
                p0 = base.end
                baseDir = True
        
        if angle > 0 and self.ccw(cPlane,p0, intsn, p1):
            angle = -angle
        elif angle < 0 and not self.ccw(cPlane,p0, intsn, p1) :
            angle = -angle

        nBaseS = graphTransf.apply(base.start)
        nBaseE = graphTransf.apply(base.end)
        
        if anchorVect is not None and self.ccw(cPlane,base.end, base.start, anchorVect[0]) != self.ccw(cPlane,nBaseE, nBaseS, anchorVect[1]):
            angle = -angle
            
        if nBaseS > nBaseE:
            baseDir = not baseDir
            angle = pi-angle
            angle = -angle

        
        p0 =  cPlane.position(cPlane.localize(vz))
        if baseDir:
            d1 = cPlane.localize(nBaseE.subtract(nBaseS))
        else:
            d1 = cPlane.localize(nBaseS.subtract(nBaseE))
            
        cp = containedPoints[0]
        nCP = graphTransf.apply(cp)
        
        ca = cos(angle)
        sa = sin(angle)
        
        d2 = cPlane.position(d1.__class__(ca*d1.x-sa*d1.y, sa*d1.x+ca*d1.y)).subtract(p0)
        
        nLen = (nBaseE.subtract(nBaseS).length() * new.length()) / (base.length() * d2.length())
        if graphTransf.contains(new.start):
            nStart = graphTransf.apply(new.start)
            nEnd = nStart.add(d2.scale(nLen))
        elif graphTransf.contains(new.end):
            nEnd = graphTransf.apply(new.end)
            nStart = nEnd.add(d2.scale(nLen))
        elif new.start == intsn:
            pLine = new.__class__(new.sort, nCP, nCP.add(d2))
            nBase = base.edit(tfSet)
            nStart = nBase.intersection(pLine)
            nEnd = nStart.add(d2.scale(nLen))
        elif new.end == intsn:
            pLine = new.__class__(new.sort, nCP, nCP.add(d2))
            nBase = base.edit(tfSet)
            nEnd = nBase.intersection(pLine)
            nStart = nEnd.add(d2.scale(nLen))
        else:
            nLen = nLen / 2
            pLine = new.__class__(new.sort, nCP, nCP.add(d2))
            nBase = base.edit(tfSet)
            intersection = pLine.intersection(nBase)
            nStart = intersection.subtract(d2.scale(nLen))
            nEnd = intersection.add(d2.scale(nLen))            
        
        nLine = new.__class__(new.sort, nStart, nEnd)
        check = True
        for intersectionSet in intersections:
            intersection = intersectionSet[0]
            if graphTransf.contains(intersection):
                intersection = graphTransf.apply(intersection)
            else:
                intersection = intersectionSet[1].edit(tfSet).intersection(nLine)
            if not intersection.subtract(nLine.root).parallel(nLine.direction):
                check = False
                break
            if intersection < nLine.start:
                nLine.start = intersection
            elif intersection > nLine.end:
                nLine.end = intersection
                
        if not check:
            return []
        
        if not nLine.reverse:
            nStart = nLine.start
            nEnd = nLine.end
        else:
            nStart = nLine.end
            nEnd = nLine.start
            
        posMatch = []
        posMatch.append(((new.start, nStart),(new.end, nEnd)))
        
        if anchorVect is None:
            sa = -sa
            d2 = cPlane.position(d1.__class__(ca*d1.x-sa*d1.y, sa*d1.x+ca*d1.y))
            
            nLen = (nBaseE.subtract(nBaseS).length() * new.length()) / (base.length() * d2.length())
            if graphTransf.contains(new.start):
                nStart = graphTransf.apply(new.start)
                nEnd = nStart.add(d2.scale(nLen))
            elif graphTransf.contains(new.end):
                nEnd = graphTransf.apply(new.end)
                nStart = nEnd.subtract(d2.scale(nLen))
            elif new.start == intsn:
                pLine = new.__class__(new.sort, nCP, nCP.add(d2))
                nBase = base.edit(tfSet)
                nStart = nBase.intersection(pLine)
                nEnd = nStart.add(d2.scale(nLen))
            elif new.end == intsn:
                pLine = new.__class__(new.sort, nCP, nCP.add(d2))
                nBase = base.edit(tfSet)
                nEnd = nBase.intersection(pLine)
                nStart = nEnd.subtract(d2.scale(nLen))
            else:
                nLen = nLen / 2
                pLine = new.__class__(new.sort, nCP, nCP.add(d2))
                nBase = base.edit(tfSet)
                intersection = pLine.intersection(nBase)
                nStart = intersection.subtract(d2.scale(nLen))
                nEnd = intersection.add(d2.scale(nLen))
            
            nLine = new.__class__(new.sort, nStart, nEnd)

            
            if not nLine.reverse:
                nStart = nLine.start
                nEnd = nLine.end
            else:
                nStart = nLine.end
                nEnd = nLine.start
                
            posMatch.append(((new.start, nStart),(new.end, nEnd)))
        return posMatch
    """
        
    
    def solveLineLineDistance(self, tfSet, bounds, cPlane, fullPlane):
        distance = self.evaluateBounds(bounds, tfSet)
        
        base = bounds[0].ind
        new = bounds[1].ind
        if (new.start.y > base.projection(new.start).y) != (distance > 0):
            distance = -distance
            
        graphTransf = self.getTransform(tfSet)
        
        anchorVect = None
        for i in graphTransf.vects:
            if anchorVect is None and not i[0].colinear(base.start, base.end):
                anchorVect = i
                break
        
        intersections = []
        for i in self.nodeList.lines:
            if i.nodeElement.intersect(new):
                ints = i.nodeElement.intersection(new)
                if new.containsPoint(ints):
                    intersections.append([ints, i.forms[0]])
                    
        for i in self.rhsLines:
            if graphTransf.contains(i.start) and graphTransf.contains(i.end) and i.intersect(new):
                ints = i.intersection(new)
                if new.containsPoint(ints):
                    intersections.append([ints, i])
                    
        
        nBaseS = graphTransf.apply(base.start)
        nBaseE = graphTransf.apply(base.end)
        
        if anchorVect is not None and self.ccw(cPlane, base.end, base.start, anchorVect[0]) != self.ccw(cPlane,nBaseE, nBaseS, anchorVect[1]):
            distance = -distance
            
        posMatch = []
        
        if fullPlane:
            if nBaseS > nBaseE:
                d1 = cPlane.localize(nBaseS.subtract(nBaseE))
                distance = -distance
            else:
                d1 = cPlane.localize(nBaseE.subtract(nBaseS))
            p0 =  cPlane.position(cPlane.localize(vz))
            pDir = cPlane.position(d1.__class__(-d1.y, d1.x)).subtract(p0)
            
            sc = ((distance*distance)/(pDir.dotproduct(pDir))).sqrt()
               
            if cPlane.normal.z < 0:
                distance = -distance
               
            if distance < 0:
                sc = -sc
            
            nStart = nBaseS.add(pDir.scale(sc))
            nEnd = nBaseE.add(pDir.scale(sc))
    
            nLine = new.__class__(new.sort, nStart, nEnd)
            check = True
            for intersectionSet in intersections:
                intersection = intersectionSet[0]
                if graphTransf.contains(intersection):
                    intersection = graphTransf.apply(intersection)
                else:
                    intersection = intersectionSet[1].edit(tfSet).intersection(nLine)
                if not intersection.subtract(nLine.root).parallel(nLine.direction):
                    check = False
                    break
                if intersection < nLine.start:
                    nLine.start = intersection
                elif intersection > nLine.end:
                    nLine.end = intersection
            if not check:
                return []
                
            if not nLine.reverse:
                nStart, nEnd = nLine.start, nLine.end
            else:
                nStart, nEnd = nLine.end, nLine.start
                
            posMatch.append([[new.start, nStart],[new.end, nEnd]])
            
            if anchorVect is None:
                sc = -sc
                
                nStart = nBaseS.add(pDir.scale(sc))
                nEnd = nBaseE.add(pDir.scale(sc))
                
                nLine = new.__class__(new.sort, nStart, nEnd)
                
                if not nLine.reverse:
                    nStart, nEnd = nLine.start, nLine.end
                else:
                    nStart, nEnd = nLine.end, nLine.start
                posMatch.append([[new.start, nStart],[new.end, nEnd]])
            return posMatch
        else:
            projStart = cPlane.projection(nBaseS)
            pdir = cPlane.normal.product(nBaseE.subtract(nBaseS))
            c = projStart.subtract(nBaseS).length()
            if c > abs(distance):
                return []
            dis = (distance*distance - c ** 2).sqrt()
            if distance < 0:
                dis = -dis

            pD = pdir.unit().scale(dis)
            nStart = projStart.add(pD)
            nEnd = nStart.add(nBaseE.subtract(nBaseS))
            posMatch.append([[new.start, nStart],[new.end, nEnd]])
            nStart = projStart.subtract(pD)
            nEnd = nStart.add(nBaseE.subtract(nBaseS))
            posMatch.append([[new.start, nStart],[new.end, nEnd]])
            return posMatch
    
    

    def solveLinePointDistance(self, tfSet, bounds, cPlane):
        distance = self.evaluateBounds(bounds, tfSet)
        
        base = bounds[0].ind
        new = bounds[1].ind
        graphTransf = self.getTransform(tfSet)
        
        if (new.value.y > base.projection(new.value).y) != (distance > 0):
            distance = -distance
            
        if cPlane is not None and cPlane.normal.z < 0:
            distance = -distance
            
        nCline = None
        for i in self.nodeList.lines:
            if i.nodeElement.containsPoint(new.value):
                nCline = i.forms[0].edit(tfSet)
                break
        if nCline is None:
            for i in self.rhsLines:
                if graphTransf.contains(i.start) and graphTransf.contains(i.end) and i.containsPoint(new.value):
                    cline = i
                    nCline = cline.edit(tfSet)
                    break            
            if nCline is None:
                return []
            
        
        anchorVect = None
        for i in graphTransf.vects:
            if anchorVect is None and not i[0].colinear(base.start, base.end):
                anchorVect = i
                break
            
            
        if cPlane is not None:
            nBaseS = graphTransf.apply(base.start)
            nBaseE = graphTransf.apply(base.end)
            
            if self.ccw(cPlane,base.end, base.start, anchorVect[0]) != self.ccw(cPlane,nBaseE, nBaseS, anchorVect[1]):
                distance = -distance
                
            if nBaseS > nBaseE:
                d1o = nBaseS.subtract(nBaseE)
                d1 = cPlane.localize(d1o)
                distance = -distance
            else:
                d1o = nBaseE.subtract(nBaseS)
                d1 = cPlane.localize(d1o)
                
            p0 =  cPlane.position(cPlane.localize(vz))
            pDir = cPlane.position(d1.__class__(-d1.y, d1.x)).subtract(p0)
            
            sc = ((distance*distance)/(pDir.dotproduct(pDir))).sqrt()
                
            if distance < 0:
                sc = -sc
            
            nStart = nBaseS.add(pDir.scale(sc))
            nLine = base.__class__(base.sort, nStart, nStart.add(d1o))
            try:       
                nPoint = nLine.intersection(nCline)
            except ValueError:
                return []
                
            return [[[new.value, nPoint]]]
        else:
            nBase = base.transform(tfSet)
            p1 = nBase.root
            p2 = nCline.root
            d1 = nBase.direction
            d2 = nCline.unitDir
            n = d1.product(d2)
            n1 = d1.product(n)
            n2 = d2.product(n)
            f1 = p1.add(d1.scale((p2.subtract(p1).dotproduct(n2))/(d1.dotproduct(n2))))
            f2 = p2.add(d2.scale((p1.subtract(p2).dotproduct(n1))/(d2.dotproduct(n1))))
            c = f1.subtract(f2).length()
            if c > abs(distance):
                return []
            dis = (distance*distance - c ** 2).sqrt()
            
            nPoint = f2.add(d2.scale(dis))
            nPoint1 = f2.add(d2.scale(-dis))
            return [[[new.value, nPoint]], [[new.value, nPoint1]]]
        
    def solvePointLineDistance(self, tfSet, bounds, cPlane, fullPlane):
        distance = self.evaluateBounds(bounds, tfSet)
        
        base = bounds[0].ind
        new = bounds[1].ind

        graphTransf = self.getTransform(tfSet)
        
        cPoint = []
        lineN = new.parametricNode()
        for i in graphTransf.vects:
            if lineN.containsPoint(i[0]):
                cPoint.append(i[0])
                cPoint.append(i[1])
                if new.containsPoint(i[0]): cPoint.append(True)
                else: cPoint.append(False)
                break
        if len(cPoint) == 0:
            raise ValueError
            
        anchorVect = None
        for i in graphTransf.vects:
            if anchorVect is None and not i[0].colinear(base.value, cPoint[0]):
                anchorVect = i
                break
        
        intersections = []
        for i in self.nodeList.lines:
            if i.nodeElement.intersect(new):
                ints = i.nodeElement.intersection(new)
                if new.containsPoint(ints):
                    intersections.append([ints, i.forms[0]])
                    
        for i in self.rhsLines:
            if graphTransf.contains(i.start) and graphTransf.contains(i.end) and i.intersect(new):
                ints = i.intersection(new)
                if new.containsPoint(ints):
                    intersections.append([ints, i])
                    
        
        nBase = graphTransf.apply(base.value)
        
        if fullPlane:
            p0 =  cPlane.position(cPlane.localize(vz))
            d1 = cPlane.localize(nBase.subtract(cPoint[1]))
            d1len = d1.length()
            sa = distance / d1len
            
            try:
                ca = (d1len*d1len - distance*distance).sqrt() / d1len
            except ValueError:
                return []
            
            if anchorVect is not None:
                ccwC = self.ccw(cPlane, base.value, cPoint[0], anchorVect[0])
                if ccwC or ccwC != self.ccw(cPlane, nBase, cPoint[1], anchorVect[1]):
                    sa = -sa
                
            d2 = cPlane.position(d1.__class__(ca*d1.x-sa*d1.y, sa*d1.x+ca*d1.y)).subtract(p0)
                
            nLine = line3D(new.sort, cPoint[1], cPoint[1].add(d2))
            
            newInts = []
            for i in intersections:
                intstn = i[1].edit(tfSet).intersection(nLine)
                if intstn not in newInts:
                    newInts.append(intstn)
                
            posMatch = []
            if cPoint[-1] is True:
                if cPoint[1] not in newInts:
                    newInts.append(cPoint[1])
                if len(newInts) < 2:
                    newInts.append(nLine.projection(nBase))
                newInts = sorted(newInts)
                if new.start == cPoint[0]:
                    if cPoint[1] == newInts[-1]:
                        posMatch.append([[new.end, newInts[0]]])
                    else:
                        posMatch.append([[new.end, newInts[-1]]])
                elif new.end == cPoint[0]:
                    if cPoint[1] == newInts[0]:
                        posMatch.append([[new.start, newInts[-1]]])
                    else:
                        posMatch.append([[new.start, newInts[0]]])
                else:
                    posMatch.append([[new.start, newInts[0]],[new.end, newInts[-1]]])
            else:
                newInts = sorted(newInts)
                if len(newInts) < 2:
                    newInts.append(nLine.projection(nBase))
                    if len(newInts) < 2:
                        newInts.append(cPoint[1])
                posMatch.append([[new.start, newInts[0]],[new.end, newInts[-1]]])
                
            if anchorVect is None:
                sa = -sa
                d2 = cPlane.position(d1.__class__(ca*d1.x-sa*d1.y, sa*d1.x+ca*d1.y))
                nLine = line3D(new.sort, cPoint[1], cPoint[1].add(d2))
                if new.start == cPoint[0]:
                    posMatch.append([[new.end, nLine.projection(nBase)]])
                elif new.end == cPoint[0]:
                    posMatch.append([[new.start, nLine.projection(nBase)]])
                else:
                    posMatch.append([[new.start, cPoint[1]],[new.end, nLine.projection(nBase)]])
            return posMatch
        else:
            p0 = cPlane.projection(vz)
            proj = cPlane.projection(nBase)
            pD = nBase.subtract(cPoint[1])
            pB = nBase.subtract(proj)
            d2 = distance ** 2
            dpD = pD.dotproduct(pD)
            dpB = pB.dotproduct(pB)
            if dpD < d2 or d2 < dpB:
                return []
            disAdj = (dpD - d2).sqrt()
            disHyp = (dpD - dpB).sqrt()
            disOpp = (d2 - dpB).sqrt()
            sinA = disOpp / disHyp
            cosA = disAdj / disHyp
            projDir = cPlane.localize(proj.subtract(cPoint[1]))
            
            nDir1 = projDir.__class__(projDir.x*cosA - projDir.y*sinA, projDir.x*sinA + projDir.y*cosA)
            nDir1 = cPlane.position(nDir1).subtract(p0)
            nDir2 = projDir.__class__(projDir.x*cosA + projDir.y*sinA, projDir.y*cosA - projDir.x*sinA )
            nDir2 = cPlane.position(nDir2).subtract(p0)
            
            d = [nDir1, nDir2]
            posMatch = []
            for nDir in d:
                nLine = line3D(new.sort, cPoint[1], cPoint[1].add(nDir))
                
                newInts = []
                for i in intersections:
                    intstn = i[1].edit(tfSet).intersection(nLine)
                    if intstn not in newInts:
                        newInts.append(intstn)
                if cPoint[-1] is True:
                    if cPoint[1] not in newInts:
                        newInts.append(cPoint[1])
                    if len(newInts) < 2:
                        newInts.append(nLine.projection(nBase))
                    newInts = sorted(newInts)
                    if new.start == cPoint[0]:
                        if cPoint[1] == newInts[-1]:
                            posMatch.append([[new.end, newInts[0]]])
                        else:
                            posMatch.append([[new.end, newInts[-1]]])
                    elif new.end == cPoint[0]:
                        if cPoint[1] == newInts[0]:
                            posMatch.append([[new.start, newInts[-1]]])
                        else:
                            posMatch.append([[new.start, newInts[0]]])
                    else:
                        posMatch.append([[new.start, newInts[0]],[new.end, newInts[-1]]])
                else:
                    newInts = sorted(newInts)
                    if len(newInts) < 2:
                        newInts.append(nLine.projection(nBase))
                        if len(newInts) < 2:
                            newInts.append(cPoint[1])
                    posMatch.append([[new.start, newInts[0]],[new.end, newInts[-1]]])
            
            return posMatch
        
    def solvePointPointDistance(self, tfSet, bounds):
        distance = self.evaluateBounds(bounds, tfSet)
        
        base = bounds[0].ind
        new = bounds[1].ind
        graphTransf = self.getTransform(tfSet)
        
        nCline = None
        for i in self.nodeList.lines:
            if i.nodeElement.containsPoint(new.value):
                cline = i.nodeElement
                nCline = cline.edit(tfSet)
                break
        if nCline is None:
            for i in self.rhsLines:
                if graphTransf.contains(i.start) and graphTransf.contains(i.end) and i.containsPoint(new.value):
                    cline = i
                    nCline = cline.edit(tfSet)
                    break            
            if nCline is None:
                return []
            
        """
        proj = cline.projection(base.value)
        if proj > new.value:
            distance = -distance
        if nCline.reverse:
            distance = -distance
            """
        
        nBase = graphTransf.apply(base.value)
        nProj = nCline.projection(nBase)
        nProsBaseDist = nProj.subtract(nBase)
        try:
            dProjP = (distance**2-(nProsBaseDist.dotproduct(nProsBaseDist))).sqrt()
        except ValueError:
            return []
            
        """
        if distance < 0:
            dProjP = -dProjP
        """ 
        
        nPoint = nProj.add(nCline.unitDir.scale(dProjP))
        nPoint1 = nProj.subtract(nCline.unitDir.scale(dProjP))
        
        return [[[new.value, nPoint]],[[new.value, nPoint1]]]
        
        
        
        
    def solveDoubleLinePointDistance(self, tfSet, d1, d2, line1, line2, pointNew, pl):
        nl1 = line1.edit(tfSet)
        nl2 = line2.edit(tfSet)
        if nl1.parallel(nl2):
            return []
        n1 = vector3D(nl1.unitDir.y, -nl1.unitDir.x)
        r1_1 = nl1.root.add(n1.scale(d1))
        r1_2 = nl1.root.subtract(n1.scale(d1))
        nl1_1 = nl1.__class__(nl1.sort, r1_1, r1_1.add(nl1.unitDir))
        nl1_2 = nl1.__class__(nl1.sort, r1_2, r1_2.add(nl1.unitDir))
        n2 = vector3D(nl2.unitDir.y, -nl2.unitDir.x)
        r2_1 = nl2.root.add(n2.scale(d2))
        r2_2 = nl2.root.subtract(n2.scale(d2))
        nl2_1 = nl2.__class__(nl2.sort, r2_1, r2_1.add(nl2.unitDir))
        nl2_2 = nl2.__class__(nl2.sort, r2_2, r2_2.add(nl2.unitDir))
        i1 = nl1_1.intersection(nl2_1)
        i2 = nl1_2.intersection(nl2_1)
        i3 = nl1_1.intersection(nl2_2)
        i4 = nl1_2.intersection(nl2_2)
        return [[[pointNew.value, i1]],[[pointNew.value, i2]],[[pointNew.value, i3]],[[pointNew.value, i4]]]        
        
    def solvePointPointLinePointDistance(self, tfSet, d1, d2, point1, line1, pointNew):
        np1 = point1.edit(tfSet)
        nl1 = line1.edit(tfSet)
        nProj = nl1.projection(np1.value)
        d3 = nProj.subtract(np1.value).length()
        if (d1 + d2) < d3:
            return []
        elif (d1 + d2) == d3:
            nPoint = nProj.subtract(np1.value).scale(d1/d3).add(np1.value)
            return [[[pointNew.value, nPoint]]]
        elif abs(d1-d2) == d3:
            if d1 > d2:
                nPoint = nProj.subtract(np1.value).scale(d1/d3).add(np1.value)
            else:
                nPoint = np1.value.subtract(nProj).scale(d2/d3).add(nProj)
            return [[[pointNew.value, nPoint]]]
        posMatch = []
        
        if d1 > (d2 + d3):
            a = (d1*d1 - (d3+d2)**2).sqrt()
            v = nProj.subtract(np1.value)
            nPt1 = nProj.add(nl1.unitDir.scale(a)).add(v.scale(d2/d3))
            nPt2 = nProj.subtract(nl1.unitDir.scale(a)).add(v.scale(d2/d3))
            posMatch.append([[pointNew.value, nPt1]])
            posMatch.append([[pointNew.value, nPt2]])
        a = (d1*d1 - (d3-d2)**2).sqrt()
        v = np1.value.subtract(nProj)
        nPt1 = nProj.add(nl1.unitDir.scale(a)).add(v.scale(d2/d3))
        nPt2 = nProj.subtract(nl1.unitDir.scale(a)).add(v.scale(d2/d3))
        posMatch.append([[pointNew.value, nPt1]])
        posMatch.append([[pointNew.value, nPt2]])
        return posMatch

    def solveDoublePointLineDistance(self, tfSet, d1, d2, point1, point2, lineNew, pl):
        np1 = point1.edit(tfSet)
        np2 = point2.edit(tfSet)
        
        n = vector3D(0,0,1) if pl is None else pl.normal
        u = np2.value.subtract(np1.value)
        d3 = u.length()
        posMatch = []
        if d1+d2 < d3:
            h1 = d3*d1/(d1+d2)
            intPt = u.scale(h1/d3).add(np1.value)
            cosAlpha = d1/h1
            sinSqAlpha = 1 - cosAlpha * cosAlpha
            affine1 = affineTransform3D.rotation(n, sinSqAlpha, cosAlpha)
            affine2 = affineTransform3D.rotation(n.scale(-1), sinSqAlpha, cosAlpha)
            startPoint1 = u.transform(affine1).scale(d1/d3).add(np1.value)
            startPoint2 = u.transform(affine2).scale(d1/d3).add(np1.value)
            endPoint1 = intPt.subtract(startPoint1).scale(d3/d1).add(startPoint1)
            endPoint2 = intPt.subtract(startPoint2).scale(d3/d1).add(startPoint2)
            posMatch.append([[lineNew.start, startPoint1], [lineNew.end, endPoint1]])
            posMatch.append([[lineNew.start, startPoint2], [lineNew.end, endPoint2]])
        elif d1+d2 == d3:
            h1 = d3*d1/(d1+d2)
            intPt = u.scale(h1/d3).add(np1.value)
            v = vector3D(u.y, -u.x)
            pt1 = intPt.add(v)
            pt2 = intPt.subtract(v)
            posMatch.append([[lineNew.start, pt1], [lineNew.end, pt2]])
        if d1 == d2:
            v = vector3D(u.y, -u.x).scale(d1/d3)
            spt = np1.value.add(v)
            ept = np2.value.add(v)
            posMatch.append([[lineNew.start, spt], [lineNew.end, ept]])
        elif abs(d1-d2) == d3:
            v = vector3D(u.y, -u.x)
            if d1>d2:
                intPt = u.scale(d1/d3).add(np1.value)
            else:
                intPt = np1.value.subtract(u.scale(d1/d3))
            pt1 = intPt.add(v)
            pt2 = intPt.subtract(v)
            posMatch.append([[lineNew.start, pt1], [lineNew.end, pt2]])
        elif abs(d1-d2) < d3:
            cosAlpha = (d1-d2)/d3
            sinSqAlpha = 1 - cosAlpha * cosAlpha
            affine1 = affineTransform3D.rotation(n, sinSqAlpha, cosAlpha)
            v1 = u.transform(affine1).scale(1/d3)
            pt1 = np1.value.add(v1.scale(d1))
            pt2 = np2.value.add(v1.scale(d2))
            posMatch.append([[lineNew.start, pt1], [lineNew.end, pt2]])
            affine2 = affineTransform3D.rotation(n.scale(-1), sinSqAlpha, cosAlpha)
            v2 = u.transform(affine2).scale(1/d3)
            pt1 = np1.value.add(v2.scale(d1))
            pt2 = np2.value.add(v2.scale(d2))
            posMatch.append([[lineNew.start, pt1], [lineNew.end, pt2]])
        return posMatch

    def solveDoublePointPointDistance(self, tfSet, d1, d2, point1, point2, pointNew, pl):
        np1 = point1.edit(tfSet)
        np2 = point2.edit(tfSet)
        d3 = np2.value.subtract(np1.value).length()
        if (d1 + d2) < d3 or abs(d1 - d2) > d3:
            return []
        elif (d1 + d2) == d3:
            nPoint = np2.value.subtract(np1.value).scale(d1/d3).add(np1.value)
            return [[[pointNew.value, nPoint]]]
        elif abs(d1-d2) == d3:
            if d1 > d2:
                nPoint = np2.value.subtract(np1.value).scale(d1/d3).add(np1.value)
            else:
                nPoint = np1.value.subtract(np2.value).scale(d2/d3).add(np2.value)
            return [[[pointNew.value, nPoint]]]
        u = np2.value.subtract(np1.value)

        cosAlpha = (d1*d1 + d3*d3 - d2*d2) / (2*d1*d3)
        sinSqAlpha = 1 - cosAlpha * cosAlpha
        n = vector3D(0,0,1) if pl is None else pl.normal
        affine1 = affineTransform3D.rotation(n, sinSqAlpha, cosAlpha)
        affine2 = affineTransform3D.rotation(n.scale(-1), sinSqAlpha, cosAlpha)
        
        nPoint1 = u.transform(affine1).scale(d1/d3).add(np1.value)
        nPoint2 = u.transform(affine2).scale(d1/d3).add(np1.value)
        return [[[pointNew.value, nPoint1]],[[pointNew.value, nPoint2]]]
        
        
        
    def readjustLength(self, tfSet, bounds, lhs, rhs):
        if len(bounds) == 3:
            rand = randint(0, 100000)
            length = (bounds[-1]-bounds[-2])*rand/100000 + bounds[-2]
        elif len(bounds) == 2:
            length = bounds[-1]
        else:
            raise ValueError()
        line = bounds[0].ind
        graphTransf = self.getTransform(tfSet)
        
        rhsCheckE = [False]
        rhsCheckS = [False]
        
        # check if there are any point/line containing line.end or line.start
        rhs.accept(self, direction = ['containsCheck', line.end, rhsCheckE, line])
        rhs.accept(self, direction = ['containsCheck', line.start, rhsCheckS, line])
        
        if rhsCheckE[0] == rhsCheckS[0] == False: # if no other line/point containing both line.start and line.end:
            adjust = []
            for i in graphTransf.vects:
                if i[0] == line.start:
                    adjust.insert(0,line.start)
                elif i[0] == line.end:
                    adjust.insert(0,line.end)
        elif rhsCheckE[0] is False: # if there are no lines/points containing line.end then pass --> adjust is None
            adjust = [line.end, line.start]
        elif rhsCheckS[0] is False: # if 
            adjust = [line.start, line.end]
        else:
            lhsCheckE = [False]
            lhs.accept(self, direction = ['containsCheck', line.end, lhsCheckE, line])
            lhsCheckS = [False]
            lhs.accept(self, direction = ['containsCheck', line.start, lhsCheckS, line])
            if lhsCheckE[0] is False and lhsCheckS[0] is False:
                for i in self.directives:
                    lhsCheckE = False
                    lhsCheckS = False
                    if i == 'length' or i == 'point_on_line':
                        continue
                    for j in i[1]:
                        ind = j[1]
                        if 'lineSegment' in ind.type:
                            if line.end == ind.start or line.end == ind.end:
                                lhsCheckE = True
                            if line.start == ind.start or line.start == ind.end:
                                lhsCheckS = True
                        elif 'point' in ind.type:
                            if line.end == ind.value:
                                lhsCheckE = True
                            if line.start == ind.value:
                                lhsCheckS = True
                if lhsCheckE and lhsCheckS:
                    raise ValueError('cannot apply directive "Length" onto line '+str(line))
                elif lhsCheckE is False:
                    adjust = [line.end, line.start]
                else:
                    adjust = [line.start, line.end]
            elif lhsCheckE[0] is False:
                adjust = [line.end, line.start]
            elif lhsCheckS[0] is False:
                adjust = [line.start, line.end]
            else:
                raise ValueError('cannot apply directive "Length" onto line '+str(line))
        
        new0 =  graphTransf.apply(adjust[0])
        new1 =  graphTransf.apply(adjust[1])
        d0 = new0.subtract(new1)
        newPoint = new1.add(d0.scale(length/d0.length()))
        #print(adjust[0], new0, new1)
        return [[[adjust[0], newPoint]]]
            