# -*- coding: utf-8 -*-
"""
Created on Fri Aug  5 11:27:36 2016

@author: akibdpt
"""
from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division
from __future__ import absolute_import
from builtins import super
"""
from future import standard_library
standard_library.install_aliases()
"""
from .multiplyForm import multiplyForm


"""
 * A discrete-form specifies a form with a discrete operational
 * behavior, corresponding a mathematical set: any two individuals in the
 * form are either identical or disjoint. A discrete form is maximal
 * if no two individuals are identical, and the individuals' attribute forms,
 * if any, are also maximal.
 * 
 * The DiscreteForm class extends on the {MultiplyForm} class and
 * specifies the discrete operational behavior of forms.
"""
class discrete(multiplyForm):
    
    type = 'form-discrete'
    
    def __init__(self, sort, *args, **kwargs):
        super(discrete, self).__init__(sort)
    
    """
    print out the form for easier assessment
    structure:
    |sort name: n elements
     _ individual 1
     _ individual 2
         ...
     _ individual n
      |attribute1 sort name: m elements
       _ attribute1 1
      |attribute2 sort name: p elements
       _ attribute1 2
    """
    def printE(self, string = ''):
        print(string + '|' + self.sort.name+' ('+str(self.sort.baseSort().characteristic.type)+'):',len(self.elements),'elements')
        for ind in self.elements:
            print(string, end = ' ')
            print(ind.printVal())
            if ind.hasAttribute():
                ind.attribute.printE(string+'  ')
        return
        
        """
    def printRhino(self, dX=0, dY = 0):
        if self.sort.baseSort().characteristic.type == 'ind-label':
            s = [0,0,0]            
            for lbl in self.elements:
                if lbl.value == "0":
                    s[0] = 255
                if lbl.value == "1":
                    s[1] = 255
                if lbl.value == "2":
                    s[2] = 255
            s = ' SelLast -Properties o c o '+str(s[0])+','+str(s[1])+','+str(s[2])+' EnterEnd SelNone'
            print(s)
            return
        for ind in self.elements:
            st = ind.printRhino(dX, dY)
            if st is not None:
                print(st)
                if ind.hasAttribute():
                    ind.attribute.printRhino(dX, dY)
        return
        """
        
    def rhinoString(self, dX=0, dY = 0):
        if self.sort.baseSort().characteristic.type == 'ind-label':
            s = [0,0,0]            
            for lbl in self.elements:
                if lbl.value == "0":
                    s[0] = 255
                if lbl.value == "1":
                    s[1] = 255
                if lbl.value == "2":
                    s[2] = 255
            s = ' SelLast -Properties o c o '+str(s[0])+','+str(s[1])+','+str(s[2])+' EnterEnd SelNone\n'
            return s
        s = ''
        for ind in self.elements:
            st = ind.printRhino(dX, dY)
            if st is not None:
                s += st + '\n'
                if ind.hasAttribute():
                    s += ind.attribute.rhinoString(dX, dY)
        return s
        
    """
    inserting an element into the form
    """
    def insert(self, ind):
        if not self.sort == ind.sort:
            raise ValueError('Inserting an object of sort '+ind.sort.name+' into a form of sort '+self.sort.name)
            
        if ind.type == self.type:
            for i in ind:
                self.insert(i)
            return
        self.maximal = False
        thisInd = self.first()
        index = 0
        while thisInd is not None:
            if thisInd.compareValue(ind) == 0:
                if thisInd.hasAttribute() or ind.hasAttribute():
                    thisInd.attribute.sum(ind.attribute)
                return
            elif thisInd.compareValue(ind) == 1:
                self.add(index, ind)
                return
            else:
                index = index+1
                try:
                    thisInd = self.elements[index]
                except:
                    thisInd = None
        self.add(index, ind)
        return
        
    """
    maximalize the form:
    _remove Nil elements
    _reorder the elements in the form in ascending order
    """
    def maximalize(self):
        if self.maximal:
            return
        
        if len(self.elements) == 0:
            return
            
        
        ind = self.first()
        index = 1
        nextInd = self.get(index)
        disorganized = []
        while nextInd is not None:
            #if nextInd is Nil, remove it and assign nextInd as the next one
            if nextInd.isNil():
                self.remove(nextInd)
                nextInd = self.get(index)
                continue
            #if ind is Nil(should only happen for the first element), move 
            #   ind and nextInd one step forward
            elif ind.isNil():
                self.remove(ind)
                ind = nextInd
                nextInd = self.get(index)
                continue
            #if ind is equals to nextInd, sum up their attributes and remove 
            #   nextInd, assign nextInd one step forward 
            cmp = ind.compareValue(nextInd)
            if cmp == 0:
                if ind.hasAttribute() or nextInd.hasAttribute():
                    ind.attribute.sum(nextInd.attribute)                
                self.remove(nextInd)
                nextInd = self.get(index)
            #if ind is greater than nextInd, remove nextInd, add it to 
            #   disorganized to be added back in later in the correct place, 
            #   assign nextInd one step forward
            elif cmp == 1:
                self.remove(nextInd)
                disorganized.append(nextInd)
                nextInd = self.get(index)
            #if ind is lesser than nextInd, nothing happens, more ind and 
            #   nextInd one step forward
            else:
                ind = nextInd
                index += 1
                nextInd = self.get(index)
        
        #add all the disorganized element back in the form again in the correct place
        for ins in disorganized:
            self.insert(ins)
            
        #maximalize the attributes of the individuals in the form
        for ind in self.elements:
            if ind.hasAttribute():
                ind.attribute.maximalize()       
                
        self.maximal = True
        return
    
    """
    sum two forms (OR)
    e.g. formA: a, b, c
         formB:    b, c, d
        _formA.sum(formB)
      -->formA: a, b, c, d
    """
    def sum(self, other):
        if self is other:
            raise ValueError('Operation on two instances of the same form')
        if not self.sort.__eq__(other.sort):
            raise ValueError('Operation on two forms of different sorts '+str(self.sort)+'  and '+str(other.sort))
        
        self.maximalize()
        other.maximalize()
        
        thisIndex = 0
        otherIndex = 0
        thisInd = self.first()
        otherInd = other.first()
        while thisInd is not None and otherInd is not None:
            if thisInd.compareValue(otherInd) == 0:
                #if has attribute, sum
                if thisInd.hasAttribute() or otherInd.hasAttribute():
                    thisInd.attribute.sum(otherInd.attribute)
                
                thisIndex += 1                 
                otherIndex += 1
                thisInd = self.get(thisIndex)           
                otherInd = other.get(otherIndex)
            elif thisInd.compareValue(otherInd) == 1:
                self.add(thisIndex, otherInd)  
                thisIndex += 1                          
                otherIndex += 1
                otherInd = other.get(otherIndex)
            else:                          
                thisIndex += 1                 
                thisInd = self.get(thisIndex)           
        while otherInd is not None:
            self.elements.append(otherInd)
            otherIndex += 1
            otherInd = other.get(otherIndex)
        return

    """
    difference of two forms
    e.g. formA: a, b, c
         formB:    b, c, d
        _formA.difference(formB)
      -->formA: a
    """
    def difference(self, other):
        if self is other:
            raise ValueError('Operation on two instances of the same form')
        if not self.sort.__eq__(other.sort):
            raise ValueError('Operation on two forms of different sorts '+str(self.sort)+'  and '+str(other.sort))
        
        self.maximalize()
        other.maximalize()
            
        thisIndex = 0
        otherIndex = 0
        thisInd = self.first()
        otherInd = other.first()
        while thisInd is not None and otherInd is not None:
            if thisInd.compareValue(otherInd) == 0:
                #if has attribute, calculate difference
                if thisInd.hasAttribute() or otherInd.hasAttribute():
                    thisInd.attribute.difference(otherInd.attribute)
                if not thisInd.hasAttribute():
                    self.pop(thisIndex)
                    thisInd = self.get(thisIndex)
                otherIndex += 1
                otherInd = other.get(otherIndex)
            elif thisInd.compareValue(otherInd) == 1:                            
                otherIndex += 1
                otherInd = other.get(otherIndex)
            else:      
                thisIndex += 1                       
                thisInd = self.get(thisIndex)
        return

    """
    product of two forms (AND)
    e.g. formA: a, b, c
         formB:    b, c, d
        _formA.product(formB)
      -->formA:    b, c
    """
    def product(self, other):
        if self is other:
            raise ValueError('Operation on two instances of the same form')
        if not self.sort.__eq__(other.sort):
            raise ValueError('Operation on two forms of different sorts '+str(self.sort)+'  and '+str(other.sort))
            
        self.maximalize()
        other.maximalize()
        
        thisIndex = 0
        otherIndex = 0
        thisInd = self.first()
        otherInd = other.first()
        while thisInd is not None and otherInd is not None:
            if thisInd.compareValue(otherInd) == 0:
                #if has attribute, product
                if thisInd.hasAttribute():
                    thisInd.attribute.product(otherInd.attribute)
                thisIndex += 1
                otherIndex += 1
                thisInd = self.get(thisIndex)                            
                otherInd = other.get(otherIndex)
            elif thisInd.compareValue(otherInd) == 1:                
                otherIndex += 1
                otherInd = other.get(otherIndex)
            else:                             
                self.pop(thisIndex)
                thisInd = self.get(thisIndex)                            
        while thisIndex < len(self):   
            self.pop(thisIndex)
        return

        
    """
    check if this form is a part of other form
    """
    def partOf(self, other):
        if self is other:
            raise ValueError('Operation on two instances of the same form')
        if not self.sort == other.sort:
            if other.sort.contains(self.sort):
                for i in other:
                    if i.sort == self.sort or i.sort.contains(self.sort) and self.partOf(i):
                        return True
                return False
            raise ValueError('Operation on two forms of different sorts '+str(self.sort)+'  and '+str(other.sort))
        self.maximalize()
        other.maximalize()
                       
        thisIndex = 1
        otherIndex = 1
        thisInd = self.first()
        otherInd = other.first()
        while thisInd is not None and otherInd is not None:
            if thisInd.compareValue(otherInd) == 0:
                if thisInd.hasAttribute() and not thisInd.attribute.partOf(otherInd.attribute):
                    return False
                thisInd = self.get(thisIndex)
                otherInd = other.get(otherIndex)
                thisIndex += 1
                otherIndex += 1
                
            elif thisInd.compareValue(otherInd) == 1:
                otherInd = other.get(otherIndex)
                otherIndex += 1
            else:                             
                return False
        return thisInd is None
        
    def checkPartOf(self, other, flagDesc = True):
        if self.sort.baseSort().characteristic.type == 'ind-description':
            if len(self) > len(other):
                return False
            if other.type == 'form-metaForm':
                check = False
                for i in other:
                    if i.sort == self.sort:
                        other = i
                        check = True
                        break
                if not check:
                    return False
            for i in self:
                if not i.checkPartOf(other):
                    return False
            if not flagDesc:
                return True
            for i in self.elements:
                i.registerDelay(other)
            return True
        else:
            return self.partOf(other)