#-*- coding: utf-8 -*-
"""
@author: akidribm, Dy Raelene Ina Bianchi Louise Mendez
"""

# API support code imports
from ..sortal_api.convert.visualAgnostic import visualA
from ..sortal_api.convert.convert_agnostic_to_sortal import convertA
from ..sortal_api.rhino_specific_methods import rhino_methods as rm

# Rhino import/s
import Rhino as r
import rhinoscriptsyntax as rs
import scriptcontext as sc

# Conversion code initialization
convertA = convertA()

"""
=============
FUNCTION INFO
=============

vis_result: draws a single rule application selected from the list of rule applications by its index; the default index if none is provided is 0.

vis_results: draws all rule applications from the list of rule applications.

PROCESS
Deletes previous main shape with function 'clearShape'  
Redraws new main shape with the class visualA's method, 'plotAgnostic'
The specific function for drawing a form, visuaA(); it can be found in the file 'visualAgnostic.py'

INPUT 
agnostic form of main shape after rule application
layerNameMain - layer where main shape is drawn
refPointName - insertion point of frame instance of main shape
shape

OUTPUT
Rhino GUIDs of new shape drawing
"""

def vis_result(choice, ruleAppns, refPointMain = r.Geometry.Point3d(0,0,0), layerNameMain = 'Default', shapeIds = None):
    if len(ruleAppns) != 0:
        # CLEARING SHAPE
        # Clears previous shape IDs inside frame; input is a list of GUIDs
        if type(shapeIds) != type(None):
            rm.clearShape(shapeIds)                                            
        
        # PLOTTING SHAPE
        # Assigns which agnostic tuple (in this case, the main shape after rule application) to plot
        shapeNew = ruleAppns[choice][-1]    

        descriptions = None
        # DESCRIPTION TEXT STRING
        # Extracts descriptions
        if 'points' not in shapeNew and 'line segments' not in shapeNew and 'plane segments' not in shapeNew:
            for shapeName in shapeNew:
                shapes = shapeNew[shapeName]
                for subshape in shapes:
                    if 'descriptions' in subshape:
                        description_text = subshape['descriptions']
                        descriptions = construct_description(description_text)
        else:
            if 'descriptions' in shapeNew:
                description_text = shapeNew['descriptions']
                descriptions = construct_description(description_text)
                
        # Specific function for plotting is 'visualA', housed inside 'visualAgnostic'
        # To plot a form, the function visualA() takes the following inputs:
        #      (<agnostic form to be plotted>, <layer name where agnostic form is to be plotted>, <reference point)
        # The method visualA.plotAgnostic() is called to plot the form
        vis1 = visualA(shapeNew, refPointMain, layerNameMain)           # Input: element of ruleAppns; creates an instance of visualA
        shapeNewAgnosticIds = vis1.plotAgnostic()                       # Redraws shape after rule application based on agnostic tuple input 'shapeNew'  

        # Display text
        if len(descriptions) > 0:
            print ('\nApplication description(s) text string')
            print (descriptions + ' \n')
        else:
            print ('\nNo external descriptions\n')
        convertA.toSortal(shapeNew).printE()
        print('\n')
            
        return shapeNewAgnosticIds
    else:
        print ('No rule applications can be found. Ending run')
        return

# default - origin  as refPoint for plotting new geometry
def vis_results(ruleAppns, refPointMain = None, layerNameMain = 'Default', shapeIds = None, prnt = False):           
    if len(ruleAppns) != 0:
        # CLEARING SHAPE
        # Clears previous shape IDs inside frame; input is a list of GUIDs
        if type(shapeIds) != type(None) and type(shapeIds) == type([]):
            rm.clearShape(shapeIds)            
        
        newShape = []
        
        # Starting coordinates for movement vector T
        start_distance_z = 0.0
        start_distance_y = 0.0
        start_distance_x = 0.0

        for row_count, ruleAppn in enumerate(ruleAppns):
            # PLOTTING SHAPE
            # Assigns which agnostic tuple (in this case, the main shape after rule application) to plot
            
            # Extracts agnostic tuple, regardless of parametric or non-parametric
            shapeNew = ruleAppn[-1]
            descriptions = None
            # DESCRIPTION TEXT STRING
            # Extracts descriptions
            if 'points' not in shapeNew and 'line segments' not in shapeNew and 'plane segments' not in shapeNew:
                for shapeName in shapeNew:
                    shapes = shapeNew[shapeName]
                    for subshape in shapes:
                        if 'descriptions' in subshape:
                            description_text = subshape['descriptions']
                            descriptions = construct_description(description_text)
            else:
                if 'descriptions' in shapeNew:
                    description_text = shapeNew['descriptions']
                    descriptions = construct_description(description_text)
                    
            # Specific function for plotting is 'visualA', housed inside 'visualAgnostic'
            # To plot a form, the function visualA() takes the following inputs:
            #      (<agnostic form to be plotted>, <layer name where agnostic form is to be plotted>, <reference point)
            # The method visualA.plotAgnostic() is called to plot the form
            #vis1 = visualA(shapeNew, refPointMain, layerNameMain)           # Input: element of ruleAppns; creates an instance of visualA
            vis1 = visualA(shapeNew, refPointMain, layerNameMain)
            shapeNewAgnosticIds = vis1.plotAgnostic()                       # Redraws shape after rule application based on agnostic tuple input 'shapeNew'  
            
            if prnt:
                # Display text
                print ('Rule application ' + str(row_count+1) + ':')
                if descriptions != None:
                    print ('\nApplication description(s) text string')
                    print (descriptions + ' \n')
                else:
                    print ('\nNo external descriptions')
    
                convertA.toSortal(shapeNew).printE()
                print('\n')
            
            newShape.append(shapeNewAgnosticIds)

        return newShape
    else:
        print ('No rule applications can be found. Ending run')
        return

def construct_description(descriptions):
    description_text = ''

    for desc_name in descriptions:
        desc_text = descriptions[desc_name]
        
        if description_text == '':
            description_text = desc_name + ':'
        else:
            description_text = description_text+ '|' + desc_name + ':'
        
        for count, desc in enumerate(desc_text):
            if count == 0:
                description_text = description_text + desc
            else:
                description_text = description_text +';' + desc
                
    return description_text
