#-*- coding: utf-8 -*-
"""
@author: akidribm, Dy Raelene Ina Bianchi Louise Mendez
"""

# RHINO SPECIFIC IMPORTS | FROM PYTHON 3.5 to 2.7
from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals
from __future__ import absolute_import

# Rhino Scripting
import rhinoscriptsyntax as rs
import System
import Rhino as r
import scriptcontext as sc

# Python built-ins
from builtins import range
import os
import warnings

# Sortal library imports
from sortal import sort, form, rule

# API support code imports
from .convert.convert_sortal_to_agnostic import convertS
from .convert.visualAgnostic import visualA

convertS = convertS()

"""
=============
FUNCTION INFO
=============
The three functions in this file return a specific aspect of a rule object, as per the name of the method.
get_rule_description: Get the description of the rule object
get_rule_lhs: Draws the LHS as its sortal data structure; if rhino is set to True, returns a drawing in the Rhino viewport 
              and returns the corresponding list of Rhino GUIDs
get_rule_rhs: Draws the RHS as its sortal data structure; if rhino is set to True, returns a drawing in the Rhino viewport 
              and returns the corresponding list of Rhino GUIDs
         
INPUT
name: Name of target rule object (get_rule_description, get_lhs, get_rhs)
rhino (get_lhs, get_rhs): Boolean value, indicates if side of rule should be drawn and returned as list of Rhino GUIDs; 
                          True (default) - returns Rhino GUIDs (hidden from viewport), False - returns sortal form corresponding to target rule side

OUTPUT
Rule object description (text string); LHS or RHS (as sortal shape or as hidden Rhino GUIDs generated in original location of rule), if successful
None, if not successful

WARNINGS / ERRORS
TypeError: Rule name input is not a string
KeyError: Rule name input is not present in the rule register 
          OR if 'rhino_shapes_c' is not present in the sort register, i.e. sort types in sortal library has not yet been set up
    
Data Flow: rule name -> text string or storal shape
"""

def get_rule_description(name):
    if 'rhino_shapes_c' in sort.register:
        if type(name) == type('a'):
            if name in rule.register:
                return rule.register[name].description
            else:
                raise KeyError('Rule named ' + name + ' does not exist in the rule register')
        else:
            raise TypeError('Input for rule name is not a text string')
        
        return None
    else:
        raise KeyError('Compound sort type "rhino_shapes_c" not present in the sort register. Please set the highest tier sort in your SDL file to a compound sort named "rhino_shapes_c" to ensure compatibility with the API.', 0, '"rhino_shapes_c" not present"')
                  
def get_rule_lhs(name, rhino = True):
    if 'rhino_shapes_c' in sort.register:
        if type(name) == type('a'):
            if name in rule.register:
                lhs = convertS.toAgnostic(rule.register[name].lhs)
                if rhino:
                    draw_lhs = visualA(lhs)
                    lhsRhino = draw_lhs.plotAgnostic()
                    rs.HideObjects(lhsRhino)
                    return lhsRhino
                else:
                    return lhs
            else:
                raise KeyError('Rule named ' + ruleObject + ' does not exist in the rule register')
        else:
            raise TypeError('Input for rule name is not a text string')
        
        return None
    else:
        raise KeyError('Compound sort type "rhino_shapes_c" not present in the sort register. Please set the highest tier sort in your SDL file to a compound sort named "rhino_shapes_c" to ensure compatibility with the API.', 0, '"rhino_shapes_c" not present"')
                  
def get_rule_rhs(name, rhino = True):
    if 'rhino_shapes_c' in sort.register:
        if type(name) == type('a'):
            if name in rule.register:
                rhs = convertS.toAgnostic(rule.register[name].rhs)
                if rhino:
                    draw_rhs = visualA(rhs)
                    rhsRhino = draw_rhs.plotAgnostic()
                    rs.HideObjects(rhsRhino)
                    return rhsRhino
                else:
                    return rhs
            else:
                raise KeyError('Rule named ' + name + ' does not exist in the rule register')
        else:
            raise TypeError('Input for rule name is not a text string')
        
        return None
        
    else:
        raise KeyError('Compound sort type "rhino_shapes_c" not present in the sort register. Please set the highest tier sort in your SDL file to a compound sort named "rhino_shapes_c" to ensure compatibility with the API.', 0, '"rhino_shapes_c" not present"')
          